/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef NO_X11

#include <stdlib.h>
#include <cstddef>
#include <cmath>
#include <cstdio>
#include <error.h>

#include "X.h"
#include <psd_projection.h>

int xwindow::colormap(double x, double min = 0, double max = 1, bool drawcontours=0) {
	static int cnt = 0;
	cnt++;
	x=(x-min)/(max-min);
	if(x==0.0) {
		return 0xffffff;
	};
	if(x<0.0) {
		if(drawmask & DRAWNEGATIVE) return (cnt/1000)%2 * cnt%2 * 0xff0000;
		x=abs(x);
	}
	#if 1
	int r = 0xff*sqrt(x);
	int g = 0xff*x*x*x;
	int b = 0xff*(sin(x*2*M_PI)>0?sin(x*2*M_PI):0);
	#else
	int r = 0xff*(1-x);
	int g = 0xff*(1-x);
	int b = 0xff*(1-x);
	#endif
	int res = (r<<16)+(g<<8)+(b<<0);
	//if(int(x*255)%32==0) res = res^0xffffff;
	if(drawcontours && (int(x*255)%32)<5) res = 0;
	return res;
}

static char x_name[] = "SelaV";
static char x_class[] = "SelaV";
static char x_font[] = "-misc-fixed-medium-r-*-*-15-*-*-*-*-*-iso8859-1";

xwindow::xwindow() {
		dpy = XOpenDisplay(NULL);
		if(NULL == dpy) 
			error(1,0,"X11 error: Could not open display");
		screen = DefaultScreen(dpy);
		visual = DefaultVisual(dpy,screen);
		depth  = DefaultDepth(dpy,screen);
		attributes.background_pixel = XWhitePixel(dpy,screen);
		attributes.event_mask |= StructureNotifyMask;

		window = XCreateWindow( dpy,XRootWindow(dpy,screen),
				       500, 500, 500, 500, 0, depth,  InputOutput,
				       visual ,CWBackPixel, &attributes);
		XSelectInput(dpy,window,ExposureMask | KeyPressMask | ResizeRedirectMask | StructureNotifyMask | ButtonPressMask | ButtonReleaseMask) ;
		fontinfo = XLoadQueryFont(dpy,x_font);
		if(NULL == fontinfo) {
			//int nfonts;
			//char** fontnames = XListFonts(dpy,"*",10,&nfonts);
			//for(int i=0; i<nfonts; i++) {
			//	printf("%s\n",fontnames[i]);
			//}
			//XFreeFontNames(fontnames);
			error(1,0,"X11 error: Could not open font");
		}

		// Set Class hints
		hints = XAllocClassHint();
		hints->res_name = x_name;
		hints->res_class = x_class;
		XSetClassHint(dpy,window,hints);

		XAllocNamedColor(dpy, DefaultColormap(dpy, screen),"red",&color[0],&dummy);
		XAllocNamedColor(dpy, DefaultColormap(dpy, screen),"blue",&color[1],&dummy);
		XAllocNamedColor(dpy, DefaultColormap(dpy, screen),"green",&color[2],&dummy);
		XAllocNamedColor(dpy, DefaultColormap(dpy, screen),"gray",&color[3],&dummy);
		XAllocNamedColor(dpy, DefaultColormap(dpy, screen),"black",&color[4],&dummy);

		gr_values.font = fontinfo->fid;
		gr_values.foreground = color[0].pixel;
		gc =XCreateGC(dpy,window,GCFont+GCForeground, &gr_values);

		gc_grid =XCreateGC(dpy,window,GCFont+GCForeground, &gr_values);
		XSetLineAttributes(dpy,gc_grid,1,LineOnOffDash,CapRound,JoinBevel);
		XSetForeground(dpy,gc_grid,color[3].pixel);

		gc_text =XCreateGC(dpy,window,GCFont+GCForeground, &gr_values);
		XSetForeground(dpy,gc_text,color[1].pixel);
		XSetBackground(dpy,gc_text,color[3].pixel);

		gc_black =XCreateGC(dpy,window,GCFont+GCForeground, &gr_values);
		XSetForeground(dpy,gc_black,color[4].pixel);
		XSetBackground(dpy,gc_black,color[4].pixel);

		gc_red =XCreateGC(dpy,window,GCFont+GCForeground, &gr_values);
		XSetForeground(dpy,gc_red,color[0].pixel);
		XSetBackground(dpy,gc_red,color[0].pixel);

		gc_blue =XCreateGC(dpy,window,GCFont+GCForeground, &gr_values);
		XSetLineAttributes(dpy,gc_blue,2,0,0,0);
		XSetForeground(dpy,gc_blue,color[1].pixel);
		XSetBackground(dpy,gc_blue,color[1].pixel);

		XFlush(dpy);
		XMapWindow(dpy,window);
		XFlush(dpy);
		alloc();
		resize(width,height);
		XResizeWindow(dpy,window,width,height);

	}

xwindow::~xwindow() {
		//fprintf(stderr,"Window closed\n");
		XUnmapWindow(dpy,window);
		XDestroyWindow(dpy,window);
		XFree(hints);
		XCloseDisplay(dpy);
		dealloc();
}

void xwindow::alloc() {
	image32 = (char*)   malloc(width*height*4);
	data =    (double*) malloc(width*height*sizeof(double));
	coords =  (double*) malloc(width*height*2*sizeof(double));
	img = XCreateImage(dpy, CopyFromParent, depth, ZPixmap, 0, image32, width, height, bitmap_pad, bytes_per_line);
}

void xwindow::dealloc() {
	free(image32);
	free(data);
	free(coords);
}

void xwindow::resize(int w, int h) {
	dealloc();
	width=w;
	height=h;
	alloc();
		XFlush(dpy);
		XMapWindow(dpy,window);
		XFlush(dpy);
}

void xwindow::set_name(const char* name) {
	XStoreName(dpy, window, name);
}

void xwindow::update() {
	double* z = coords;
	double* out = data;
	//fprintf(stderr,"Calc Coords\n");
	for(size_t i=0; i<width; i++) {
		for(size_t j=0; j<height; j++) {
			z[0]=limits[0] + i*(limits[2]-limits[0])/width  ;
			z[1]=limits[3] - j*(limits[3]-limits[1])/height ;
			z+=2;
			out+=1;
		}
	}
	//fprintf(stderr,"Eval\n");
	if(fun!=NULL) {
		fun(coords,data,width*height,fargs);
	}
	rescale_colormap();
	//fprintf(stderr,"Draw\n");
	for(size_t i=0,n=0; i<width; i++) {
		for(size_t j=0; j<height; j++,n++) {
			XPutPixel(img,i,j,colormap(data[n],min,max,drawmask&DRAWCONTOURS));
		}
	}
	XClearArea(dpy,window,0,0,width,height,0);
	XPutImage(dpy, window, gc, img, 0, 0, 0, 0, width, height);
	XSetForeground(dpy,gc,color[0].pixel);
	if(root && (drawmask & DRAWCELLS)) {
		draw_cells(root,32);
	}
	XSetForeground(dpy,gc,color[1].pixel);
	XDrawLine(dpy,window,gc,x(0)-5,y(0)-5,x(0)+5,y(0)+5);
	XDrawLine(dpy,window,gc,x(0)+5,y(0)-5,x(0)-5,y(0)+5);
	if(root && (drawmask & DRAWGRID)) {
		draw_grid(5);
	}

	if(root && (drawmask & DRAWPROJECTION)) {
		draw_projection();
	}

	if(root && (drawmask & DRAWRECTANGLE)) {
		draw_rectangle(rectangle);
	}


	XSetForeground(dpy,gc,color[0].pixel);
	XDrawImageString(dpy,window,gc,10,10,"SelaV",5);
	XDrawImageString(dpy,window,gc,37,25,"1D",2);
	//XDrawRectangle(dpy, window, gc_black,
	//	0,0,width,height);
	//XDrawArc(dpy,window,gc,100,100,50,50,1,10);
	//fprintf(stderr,"Done\n");
}

void xwindow::rescale_colormap() {
	min = INFINITY;
	max = -INFINITY;
	for(size_t i=0; i<width*height; i++) {
			double d = data[i]>0?data[i]:0.0;
			max=d>max?d:max;
			min=d<min?d:min;
	}
}

void xwindow::draw_rectangle(double* r) const {
		double q = (r[0]-limits[0])/(limits[2]-limits[0]);
		double p = (r[3]-limits[3])/(limits[1]-limits[3]);
		double wq = (r[2]-r[0])/(limits[2]-limits[0]);
		double wp = (r[3]-r[1])/(limits[3]-limits[1]);
		// Don't draw if rectangle is not visible
		if(q>=1 or q+wq<=0 or p>=1 or p+wp<=0) return;
		XDrawRectangle(dpy, window, gc_red,
			width*q, height*p, width*wq, height*wp
		);
}

void xwindow::draw_grid(int n) {
	double hy = limits[3]-limits[1];
	double wy = pow(n,floor(log(hy/2)/log(n)));
	double hx = limits[2]-limits[0];
	double wx = pow(n,floor(log(hx/2)/log(n)));
	char buf[64];
	for(int i=0; i<2*n; i++) {
		double y0 = ceil(limits[1]/wy)*wy;
		double x0 = ceil(limits[0]/wx)*wx;
		XDrawLine(dpy, window, gc_grid, x(limits[0]), y(y0+i*wy),   x(limits[2]), y(y0+i*wy));
		XDrawLine(dpy, window, gc_grid, x(x0+i*wx),   y(limits[1]), x(x0+i*wx),   y(limits[3]));
		sprintf(buf, "%0.2e", x0+i*wx);
		XDrawImageString(dpy, window, gc_text, x(x0+i*wx),   y(limits[1])-5, buf, strlen(buf));
		sprintf(buf, "%0.2e", y0+i*wy);
		XDrawImageString(dpy, window, gc_text, x(limits[0]), y(y0+i*wy)+5,   buf, strlen(buf));
	}
	//XFillRectangle(dpy,window,gc_grid,0,height-40,64,20);
	//sprintf(buf,"%0.1e",wx);
	//XDrawImageString(dpy,window,gc_text,10,height-30,buf,strlen(buf));
	//sprintf(buf,"%0.1e",wy);
	//XDrawImageString(dpy,window,gc_text,10,height-20,buf,strlen(buf));
}

void xwindow::draw_line(double q1, double p1, double q2, double p2) const {
	double x1 = (q1-limits[0])/(limits[2]-limits[0]);
	double y1 = (p1-limits[3])/(limits[1]-limits[3]);
	double x2 = (q2-limits[0])/(limits[2]-limits[0]);
	double y2 = (p2-limits[3])/(limits[1]-limits[3]);
	XDrawLine(dpy, window, gc,
		width*x1, height*y1, width*x2, height*y2
	);
}

void xwindow::draw_plot(const double* q, const double* p, int npts, int h) const {
	short points[2*npts];
	double qvis[npts];
	double pvis[npts];
	int nvis=0;
	double pmin = INFINITY;
	double pmax = -INFINITY;

	for(int i=0; i<npts; i++) {
		if(q[i]<=limits[2] && q[i]>limits[0]) {
			qvis[nvis] = q[i];
			pvis[nvis] = p[i];
			nvis++;
		}
	}

	for(int i=0; i<nvis; i++) {
		pmin = pvis[i]<pmin?pvis[i]:pmin;
		pmax = pvis[i]>pmax?pvis[i]:pmax;
	}

	for(int i=0; i<nvis; i++) {
		double x = (qvis[i]-limits[0])/(limits[2]-limits[0]);
		double y = (pvis[i]-pmin)/(pmax-pmin);
		points[i*2+0] = width*x;
		points[i*2+1] = height-y*h;
	}
	XDrawLines(dpy, window, gc,
		(XPoint*) points, nvis, CoordModeOrigin
	);
	//fprintf(stderr,"----\n%0.16e\n%0.16e\n",pmin,pmax);
}

void xwindow::draw_projection() const {
	struct SLV::grid_t p = SLV::psd_projection(psd,2);
	size_t npts = 1u<<p.nexp;
	double qdata[npts];
	for(size_t i=0; i<npts; i++) {
		qdata[i] = p.limits[0] + i*(p.limits[1]-p.limits[0])/npts;
	}
	draw_plot(qdata,p.data,npts,height/4);
}

int xwindow::x(double d) {
	return (int) width * ((d-limits[0])/(limits[2]-limits[0]));
}
int xwindow::y(double d) {
	return (int) height * ((d-limits[3])/(limits[1]-limits[3]));
}

void xwindow::center() {
	if(root==NULL) return;
	limits[0] = root->cpoint[0]-root->cwidth[0]*1.03;
	limits[1] = root->cpoint[1]-root->cwidth[1]*1.03;
	limits[2] = root->cpoint[0]+root->cwidth[0]*1.03;
	limits[3] = root->cpoint[1]+root->cwidth[1]*1.03;
}

void xwindow::screenshot(FILE* fd) {
	fprintf(fd,"P3\n# SELA-V\n");
	fprintf(fd,"%ld %ld\n",width,height);
	fprintf(fd,"%d\n",255);
	XImage* image = XGetImage(dpy, window, 0, 0, width, height, AllPlanes, ZPixmap);
	unsigned char r,g,b;
	for(size_t i=0; i<width; i++) {
		for(size_t j=0; j<height; j++) {
			//XColor c;
			//c.pixel = XGetPixel(image, j, i);
			//XQueryColor (dpy, XDefaultColormap(dpy, XDefaultScreen(dpy)), &c);
			//r = c.red/256;
			//g = c.green/256;
			//b = c.blue/256;
			char *ptr = image->data + (i * img->bytes_per_line + 4 * j);
			b=ptr[0]; g=ptr[1]; r=ptr[2];
			fprintf(fd,"%d %d %d  ",r%256,g%256,b%256);
		}
		fprintf(fd,"\n");
	}
}

void xwindow::write_pgm(FILE* fd) {
	fprintf(fd,"P2\n# SELA-V\n");
	fprintf(fd,"%ld %ld\n",width,height);
	fprintf(fd,"%d\n",255);
	double min = INFINITY;
	double max = -INFINITY;
	for(size_t i=0; i<width*height; i++) {
			double d = data[i]>0?data[i]:0.0;
			max=d>max?d:max;
			min=d<min?d:min;
	}
	for(size_t i=0; i<width; i++) {
		for(size_t j=0; j<height; j++) {
			fprintf(fd,"% 4d",int(255*data[i*height+j]/(max-min)));
		}
		fprintf(fd,"\n");
	}
}

void xwindow::write_data(FILE* fd) {
	for(size_t i=0; i<width; i++) {
		for(size_t j=0; j<height; j++) {
			fprintf(fd,"%+e ",data[i*height+j]);
		}
		fprintf(fd,"\n");
	}
}

static char help_message[] = "\
arrow keys: panning\n\
o: zoom out\n\
p: zoom in\n\
d: write data to test.pgm\n\
f: write data to test.dat\n\
s: write screenshot to test.ppm\n\
h: print this help message\n\
u: unzoom\n\
k: toggle sample point drawing\n\
l: toggle cell drawing\n\
c: toggle contour lines\n\
n: toggle negative value highlighting\n\
g: toggle grid\n\
w: toggle projection\n\
r: rescale colormap\n\
i: print current limits\n\
q: exit\n\
mouse button 1 click: highlight cell\n\
mouse button 3 drag&release: zoom into region\n\
";

void xwindow::loop() {
	//center(); // Effectively nullifies limits set before loop call
	while(1){
		XNextEvent(dpy,&event);
		double wq = limits[2]-limits[0];
		double wp = limits[3]-limits[1];
		switch(event.type){
		case Expose:
			update();
			XSync(dpy,true); // clears pending events
			break;
		case KeyPress:
			switch(XLookupKeysym(&event.xkey,0)) {
			case XK_Left:
				limits[0] -= wq*0.03;
				limits[2] -= wq*0.03;
				break;
			case XK_Right:
				limits[0] += wq*0.03;
				limits[2] += wq*0.03;
				break;
			case XK_Up:
				limits[1] += wp*0.03;
				limits[3] += wp*0.03;
				break;
			case XK_Down:
				limits[1] -= wp*0.03;
				limits[3] -= wp*0.03;
				break;
			case XK_o:
				limits[0] -= wq*0.03;
				limits[1] -= wp*0.03;
				limits[2] += wq*0.03;
				limits[3] += wp*0.03;
				break;
			case XK_p:
				limits[0] += wq*0.03;
				limits[1] += wp*0.03;
				limits[2] -= wq*0.03;
				limits[3] -= wp*0.03;
				break;
			case XK_d: {
				FILE* fd = fopen("test.pgm","w");
				write_pgm(fd);
				fclose(fd);
				break;
				}
			case XK_f: {
				FILE* fd = fopen("test.dat","w");
				write_data(fd);
				fclose(fd);
				break;
				}
			case XK_s: {
				FILE* fd = fopen("test.ppm","w");
				screenshot(fd);
				fclose(fd);
				break;
				}
			case XK_h: {
				printf("%s",help_message);
				break;
				}
			case XK_u: /* Unzoom */
				center();
				break;
			case XK_k:
				drawmask ^= DRAWPOINTS;
				break;
			case XK_l:
				drawmask ^= DRAWCELLS;
				break;
			case XK_c:
				drawmask ^= DRAWCONTOURS;
				break;
			case XK_n:
				drawmask ^= DRAWNEGATIVE;
				break;
			case XK_g:
				drawmask ^= DRAWGRID;
				break;
			case XK_w:
				drawmask ^= DRAWPROJECTION;
				break;
			case XK_r:
				rescale_colormap();
				break;
			case XK_i:
				fprintf(stdout,"limits=[%e,%e,%e,%e]\n",limits[0],limits[1],limits[2],limits[3]);
				break;
			case XK_q:
				goto RETURN;
				break;
			}
			update();
			break;
		case ButtonPress: {
			XButtonEvent ev = event.xbutton;
			switch(ev.button) {
			case 1: {
				int i = ev.y + ev.x * height;
				double* z = coords+2*i;
				SLV::cell<2>* leaf = root->get_leaf(z);
				draw_cell(leaf);
				draw_samplepoints(leaf);
				leaf->print();
				for(int n=0; n<4; n++) {
					if(leaf->nbors[n]) {
						draw_cell(leaf->nbors[n]);
					}
				}
				//for(int n=0; n<1<<(2*leaf->nexp); n++) {
				//	fprintf(stdout,"%e\n", leaf->data[n]);
				//}
				double d = data[i];
				fprintf(stdout,"%e\n", d);
				char buf[64];
				sprintf(buf,"%0.2e",d);
				XDrawImageString(dpy,window,gc_text,ev.x,ev.y,buf,strlen(buf));

				break;
				}
			case 2: {
				if(NULL==psd) break;
				int i = ev.y + ev.x * height;
				double* z = &coords[2*i];
				double out[1];
				SLV::psd_eval<2>(z,out,1,psd);
				printf("%e %e %e\n",z[0],z[1],out[0]);
				break;
				}
			case 3: {
				// initialize mouse zoom
				b3pressed = true;
				int i = ev.y + ev.x * height;
				limits[0] = coords[2*i];
				limits[1] = coords[2*i+1];
				//printf("%e %e\n",limits[0],limits[1]);
				break;
				}
			case 4: {
				limits[0] -= wq*0.03;
				limits[1] -= wp*0.03;
				limits[2] += wq*0.03;
				limits[3] += wp*0.03;
				update();
				break;
				}
			case 5: {
				limits[0] += wq*0.03;
				limits[1] += wp*0.03;
				limits[2] -= wq*0.03;
				limits[3] -= wp*0.03;
				update();
				break;
				}
			}
			break;
		}
		case ButtonRelease: {
			XButtonEvent ev = event.xbutton;
			switch(ev.button) {
			case 3: 
				// execute mouse zoom
				b3pressed = false;
				int i = ev.y + ev.x * height;
				limits[2] = coords[2*i];
				limits[3] = coords[2*i+1];
				//printf("%e %e\n",limits[2],limits[3]);
				// reorder limits: lower-left, upper-right
				double xmin = limits[0]<=limits[2] ? limits[0] : limits[2];
				double xmax = limits[0] >limits[2] ? limits[0] : limits[2];
				double ymin = limits[1]<=limits[3] ? limits[1] : limits[3];
				double ymax = limits[1] >limits[3] ? limits[1] : limits[3];
				limits[0] = xmin;
				limits[1] = ymin;
				limits[2] = xmax;
				limits[3] = ymax;
				update();
			}
			break;
		}
		case DestroyNotify:
			goto RETURN;
			break;
		case ResizeRequest:
			resize(
				((XResizeRequestEvent*)&event)->width,
				((XResizeRequestEvent*)&event)->height
			);
			break;
		}
	}
	RETURN:;
}

void xwindow::draw_cell(const SLV::cell<2>* cell) {
	auto cp = cell->cpoint;
	auto cw = cell->cwidth;
	//if(not cell->is_leaf()) return;
	double q = (cp[0]-cw[0]-limits[0])/(limits[2]-limits[0]);
	double p = (cp[1]+cw[1]-limits[3])/(limits[1]-limits[3]);
	double wq = (2*cw[0])/(limits[2]-limits[0]);
	double wp = (2*cw[1])/(limits[3]-limits[1]);
	// Don't draw if rectangle is not visible
	if(q>=1 or q+wq<=0 or p>=1 or p+wp<=0) return;
	//printf("%e %e %e %e\n",q,p,wq,wp);
	XDrawRectangle(dpy, window, gc,
		width*q, height*p, width*wq, height*wp
	);
	if(drawmask & DRAWPOINTS) draw_samplepoints(cell);
}

void xwindow::draw_cells(const SLV::cell<2>* cell, int depth) {
	if(depth<=0) return;
	draw_cell(cell);
	for(int i=0; i<1<<2; i++) {
		if(cell->cells[i]) draw_cells(cell->cells[i],depth-1);
	}
}

void xwindow::draw_samplepoints(const SLV::cell<2>* cell) const {
	if(NULL == cell->data) return;
	size_t npts = cell->npoints();
	double coords[2*npts];
	cell->grid_coords(coords);
	short points[6*npts];
	for(size_t i=0; i<npts; i++) {
		double x = (coords[i*2+0]-limits[0])/(limits[2]-limits[0]);
		double y = (coords[i*2+1]-limits[1])/(limits[3]-limits[1]);
		points[i*6+0] = width*x;
		points[i*6+1] = height*(1-y);
		points[i*6+2] = 1;
		points[i*6+3] = 1;
		points[i*6+4] = 0;
		points[i*6+5] = 23040;
	}
	XDrawArcs(dpy, window, gc_blue,
		(XArc*) points, npts
	);
}

#endif
