/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/


#include "string.h"
#include "math.h"
#include "complex.h"

/* dispersion */

static void D(double* out, double d) {
	double M[] = {
		1, d, 0, 0,
		0, 1, 0, 0,
		0, 0, 1, 0,
		0, 0, 0, 1
	};
	memcpy(out,M,sizeof(M));
}

static void F(double* out, double a, double k0) {
	double T = tan(a);
	double M[] = {
		1,         0, 0, 0,
		abs(T*k0), 1, 0, 0,
		0,         0, 1, 0,
		0,         0, 0, 1
	};
	memcpy(out,M,sizeof(M));
}

static void B(double* out, double d, double k0) {
	double a = k0 * d;
	double R = 1/k0;
	double C = cos(a);
	double S = sin(a);
	double M[] = {
		C,    R*S,      0, R*(1-C)  ,
		-S/R, C,        0, S        ,
		-S,   -R*(1-C), 1, -R*(a-S) ,
		0,    0,        0, 1
	};
	memcpy(out,M,sizeof(M));
}

static void Q(double* out, double d, double k1) {
	_Complex sk1 = csqrt(k1);
	double ask1  = cabs(sk1);
	double C     = creal(ccos(sk1*d));
	double S     = creal(csin(sk1*d));
	double M[] = {
		C,       S/ask1, 0, 0,
		-ask1*S, C,      0, 0,
		0,       0,      1, 0,
		0,       0,      0, 1
	};
	memcpy(out,M,sizeof(M));
}

static void matrixmul(double* out, const double* A, const double* B, size_t n) {
	for(size_t i=0; i<n; i++) {
		for(size_t j=0; j<n; j++) {
			out[i*n+j] = 0;
			for(size_t k=0; k<n; k++) {
				out[i*n+j] += A[i*n+k]*B[k*n+j];
			}
		}
	}
}

void transfer4d(double* R1, const double* R0, double l, double fangle, double k0, double k1) {
	double M[16] = {0};
	if( fangle!=0 ) {
		F(M, fangle, k0);
	} else if( k0!=0 ) {
		B(M, l, k0);
	} else if( k1!=0 ) {
		Q(M, l, k1);
	} else if( l!=0 ) {
		D(M, l);
	} else {
		return;
	}
	matrixmul(R1, M, R0, 4);
}

/* chicane coefficents */

static inline double chicaneF(int n, double a, double b, double lmb) {
	if(n==0) return pow(lmb*lmb-1,-a) * pow(lmb,b);
	return -2*a*chicaneF(n-1,a+1,b+1,lmb) + b*chicaneF(n-1,a,b-1,lmb);
}

double chicanecoef(int n, double angle, double lB, double lD, char type=0) {
	double lmb = 1./sin(angle);

	double bendcoef = 0.0;
	switch(n) {
	case 0:
		bendcoef = lB * lmb * asin(1./lmb);
		break;
	case 1:
		bendcoef = lB * lmb * ( asin(1./lmb) - 1./sqrt(pow(lmb,2)-1));
		break;
	default:
		bendcoef = lB * pow(lmb,n) * chicaneF(n-2, 3./2., -1., lmb);
		break;
	}

	double driftcoef = lD * pow(lmb,n) * chicaneF(n, 1./2., 1, lmb);

	return 4*bendcoef + 2*driftcoef;
}


/* average */
// TODO move to different file; is here for historical reasons

static inline double interp(double x, double x0, double x1, double y0, double y1) {
	return y0*(x-x1)/(+x0-x1) + y1*(x-x0)/(-x0+x1);
}

static double trapz(double x0, double x1, double y0, double y1) {
	return (x1-x0)*(y0+y1)/2;
}

double average(double z0, double z1, const double* zp, const double* yp, size_t n) {
	//  zp[i0]  <= z0  <= zp[i0+1]  <= zp[i1] <= z1 <= zp[i1+1]
	z0 = z0<zp[0]   ? zp[0]   : z0;
	z1 = z1>zp[n-1] ? zp[n-1] : z1;

	// find indices containing the interval
	size_t i0=0;
	size_t i1=0;
	for(size_t i=0; i<n; i++) {
		i0 = zp[i]<=z0 ? i : i0;
		i1 = z1<=zp[i] ? i1 : i;
	}

	// interpolate yp at interval edges
	double yp0 = interp(z0, zp[i0], zp[i0+1], yp[i0], yp[i0+1]);
	double yp1 = interp(z1, zp[i1], zp[i1+1], yp[i1], yp[i1+1]);

	double avg = 0;
	avg += trapz(z0, zp[i0], yp0, yp[i0]);
	for(size_t i=i0; i<i1; i++) {
		avg += trapz(zp[i], zp[i+1], yp[i], yp[i+1]);
	}
	avg += trapz(zp[i1], z1, yp[i1], yp1);
	return avg/(z1-z0);
}

/* stuff for betaeffective */
//double as = 0; // average sqrt
//as += avgsqrt(z0, zp[i0], beta0, beta[i0]);
//for(size_t i=i0; i<i1; i++) {
//	as += avgsqrt(zp[i], zp[i+1], beta[i], beta[i+1]);
//}
//as += avgsqrt(zp[i1], z1, beta[i1], beta1);
//return as*as;
//static double avgsqrt(double x0, double x1, double y0, double y1) {
//    // \int_x0^x1 sqrt(b+m*x) dx / (x1-x0)
//    //double b = y0
//    double m = (y1-y0)/(x1-x0);
//    return 2.0/3.0/m * (y1*sqrt(y1) - y0*sqrt(y0)) / (x1-x0);
//}
