/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef SLV_CELL_H
#define SLV_CELL_H

#include "debug.h"

#include <array>
#include <cmath>
#include <cstring>
#include <memory>
#include <vector>
#include <cassert>


namespace SLV {

typedef void (*fun_t)(const double*, double*, int, const void*);
typedef void (*map_t)(const double*, double*, int, const void*, double);
typedef void (*ffun_t)(const double*, const double*, double*, int, const void*);


/// The cell<DIM> class describes a (hyper)-rectangular region in a
/// DIM-dimensional cartesian space. Additionally it supports refinement
/// of this hypercube into 2^DIM child-cubes. By storing pointers to these
/// childs, effectively a tree structure is generated.
template <int DIM>
class cell {
public:
	typedef double                   value_type;
	typedef int                      index_type;
	typedef int                      dim_type;
	typedef std::array<double,DIM>   coord_type;
	typedef cell<DIM>*               cell_ptr;

	coord_type     cpoint;                 ///< Center point of the HC.
	coord_type     cwidth;                 ///< Extensions in each dimension from center.
	cell_ptr    cells[1<<DIM]  = { NULL }; ///< Pointers to the child-trees.
	cell_ptr    nbors[2*DIM ]  = { NULL }; ///< Pointers to neighbor cells.
	cell_ptr                       parent = NULL; ///< Pointer to parent cell.
	index_type idx =               -42;    ///< General purpose integer.
	double weight  =               0.0;    ///< General purpose double.

	double* data = NULL; ///< Pointer to data
	size_t  nexp = 0; ///< Points per dim == 1<<nexp

	int loc_code = 0;

	cell() {
	}

	/// Initialize from center point and widths.
	///
	/// \param cpoint Center-point of the HC.
	/// \param cwidth Extensions from center of the HC.
	cell(coord_type cpoint, coord_type cwidth, int level = 0) : cpoint(cpoint), cwidth(cwidth) {
		DBG_CNSTR_MSG_3;
		//parent = cell_ptr(this);
		parent = NULL;
	}

	cell(const cell& a, cell* parent=NULL) : parent(parent) {
		cpoint = a.cpoint;
		cwidth = a.cwidth;
		idx = a.idx;
		nexp = a.nexp;
		loc_code = a.loc_code;
		if(nexp>0) {
			alloc(nexp);
			memcpy(data,a.data,npoints()*sizeof(double));
		}
		set_neighbors();
		for(int i=0; i<1<<DIM; i++) {
			if( NULL == a.cells[i] ) continue;
			cells[i] = new cell(*a.cells[i],this);
		}
	}



	// For future reference:
	//
	// ``loc_code''s describe the position of a cell within its parent:
	//  n-th bit set <-> cell lies in positive direction on n-th axis from parents cpoint
	//
	//   ---------
	//  | 10 | 11 |
	//  |---------|
	//  | 00 | 01 |
	//   ---------
	//
	//
	// ``nbors'' describe the position of neighbors of a cell in the following way:
        // b_n b_(n-1) ... b_1 : axis
	//                 b_0 : 0/1 => negative/positive direction along axis
	//
	// e.g: 00 = left, 01 = right, 10 = down, 11 = up
	// (NB: 100 = front, 101 = back)


	/// Initialize from center point and widths and parent
	///
	/// \param cpoint Center-point of the HC.
	/// \param cwidth Extensions from center of the HC.
	cell(coord_type cpoint, coord_type cwidth, cell_ptr parent, int loc_code = 0) : cpoint(cpoint), cwidth(cwidth), parent(parent), loc_code(loc_code) {
		DBG_CNSTR_MSG_3;
		set_neighbors();
	}



	/// Delete all childs on deconstruction
	~cell() {
		dealloc();
		for(int i=0; i<1<<DIM; i++) {
			if(cells[i]) {
				delete cells[i];
			}
		}
	}

	uint64_t oct_id() {
		if(parent) {
			return loc_code + (parent->oct_id()<<3);
		} else {
			return 7;
		}

	}

	size_t npoints() const {
		return 1<<(nexp*DIM);
	}

	void alloc(int n) {
		nexp = n;
		data = (double*) malloc( npoints() * sizeof(double));
	}

	void dealloc() {
		if(data!=NULL) free(data);
		data = NULL;
		nexp = 0;
	}

	void grid_coords(double* coords) const {
		int Nd = 1;
		int npts = npoints();
		for(int d=0; d<DIM; d++) {
			for(int n=0; n < npts; n++) {
				int i = n*DIM+d;
				//coords[i] = (n/Nd) % (1<<nexp);
				//coords[i] = cpoint[d] - cwidth[d] + 2 * cwidth[d] * coords[i] / (1<<nexp);
				double x = ((n/Nd)%(1<<nexp)) / (double)(1<<nexp);
				coords[i] = cpoint[d] - cwidth[d] + 2 * cwidth[d] * x;
			}
			Nd *= 1<<nexp;
		}
	};


	void set_neighbors() {
		/// Find all the neighbors
		if(parent==NULL) return;
		int nbor = 0;
		for(int ax=0; ax<DIM; ax++) { // Loop over axis
			for(int up=0; up<2; up++) { // Loop over positive (1) and negative (0) direction

				cell_ptr nbor_parent = parent; // Parent of the cell we are looking for

				// Is our neighbor in a neighbor of our parent?
				if( (bool)(loc_code & (1<<ax)) == (bool)up) {
					if(parent->nbors[nbor]) { // Does the neighbor exist?
						nbor_parent = parent->nbors[nbor];
					}
					else { // If not -> bad luck. continue
						nbor++;
						continue;
					}
				}

				int new_loc_code = loc_code ^ (1<<ax); // Location of the neigbor is our location with the bit on current axis flipped
				if( nbor_parent->cells[new_loc_code] ) { // If the parent has a child at this location do a handshake
					nbors[nbor] = nbor_parent->cells[new_loc_code];
					// We are the neibors neighbor on the same axis, but inverted direction i.e. nbor^1
					nbor_parent->cells[new_loc_code]->nbors[nbor ^ 1] = cell_ptr(this);
				}
				nbor++;
			}
		}
	}

	void set_links() {
		set_neighbors();
		for(int i=0; i<1<<DIM; i++) {
			if(cells[i]!=NULL) {
				cells[i]->parent = this;
			}
		}
		for(int i=0; i<1<<DIM; i++) {
			if(cells[i]!=NULL) {
				cells[i]->set_links();
			}
		}
	}

	/// Multiply by a factor
	void multiply(value_type factor) {
		if( data ) {
			for(size_t i=0; i<npoints(); i++) {
				data[i] *= factor;
			}
			return;
		}
		for(int i=0; i<1<<DIM; i++) {
			if( cells[i] ) {
				cells[i]->multiply(factor);
			}
		}
	}

	/// Modify values
	void modify(ffun_t f, void* farg) {
		if( data ) {
			int npts = npoints();
			double* coords = (double*) malloc(npts * DIM * sizeof(double));
			grid_coords(coords);
			f(data,coords,data,npts,farg);
			free(coords);
		}
		for(int i=0; i<1<<DIM; i++) {
			if( NULL == cells[i] )  continue;
			cells[i]->modify(f,farg);
		}
	}

	/// Return level below root
	int level() {
		if(parent) {
			return parent->level() + 1;
		}
		return 0;
	}

	/// True if cell contains data samples
	int filled() {
		return nexp>0;
	}

	/// Return integral of psd on cell
	double integral() const {
		if(NULL==data) return 0;
		double  sum = 0;
		int npts = npoints();
		for(int n=0; n < npts; n++) {
			sum += data[n];
		}
		double vol = 1;
		for(int d=0; d<DIM; d++) vol *= 2.*cwidth[d];
		vol /= 1<<(DIM*nexp);

		return sum*vol;
	}

	/// Return maximum datum of psd on cell
	double max() const {
		if(NULL==data) return 0;
		double m = data[0];
		int npts = npoints();
		for(int n=0; n < npts; n++) {
			m = data[n]>m ? data[n] : m;
		}
		return m;
	}

	/// Return minimum datum of psd on cell
	double min() const {
		if(NULL==data) return 0;
		double m = data[0];
		int npts = npoints();
		for(int n=0; n < npts; n++) {
			m = data[n]<m ? data[n] : m;
		}
		return m;
	}

	double integral(ffun_t f, void* args) const {
		if(NULL==data) return 0;
		const int npts = npoints();
		double coords[npts*DIM];
		grid_coords(coords);
		double vals[npts];
		f(data,coords,vals,npts,args);
		double sum = 0;
		for(int n=0; n < npts; n++) {
			sum += vals[n];
		}
		double vol = 1;
		for(int d=0; d<DIM; d++) vol *= 2.*cwidth[d];
		vol /= 1<<(DIM*nexp);

		return sum*vol;
	}


	/// Check wheter \p pt lays within the hypercube
	bool contains(const double* pt) const {
		for(int n = 0; n < DIM; n++) {
			if( (pt[n] >= cpoint[n]+cwidth[n]) | (pt[n] < cpoint[n]-cwidth[n]) ) {
				return false;
			}
		}
		return true;
	}


	/// Generates child-cells and stores pointers to them in #cells.
	// TODO Clean up when re-refining!
	void refine() {
		coord_type new_cpoint  = cpoint; // Buffer for cpoint of the child.
		coord_type half_cwidth = cwidth; // cwidth/2, which is the cwidth of the childs.
		for(auto & width : half_cwidth) width /= 2;

		for(int n = 0; n < 1<<DIM; n++) {
			for(int i = 0; i < DIM; i++) {
				new_cpoint[i] = cpoint[i] + ((1 & n>>i) ? half_cwidth[i] : (-half_cwidth[i]));
			}

			cells[n] = cell_ptr( new cell<DIM>(new_cpoint, half_cwidth, cell_ptr(this), n) ) ;
		}
	}


	int refined() {
		for(int i = 0; i < 1<<DIM; i++) {
			if(cells[i]) return 1;
		}
		return 0;
	}

	/// Sets idx to the number of leafs
	double set_weight_integral() {
		weight = integral();
		for(int i=0; i<(1<<DIM); i++) {
			if(NULL == cells[i]) continue;
			weight += cells[i]->set_weight_integral();
		}
		return weight;
	}


	/// Sets idx to the number of leafs
	int set_idx_leaf_count() {
		idx = 0;
		for(int i=0; i<(1<<DIM); i++) {
			if( cells[i] ) idx += cells[i]->set_idx_leaf_count(); 
		}

		if( idx == 0) idx = 1;
		return idx;
	}


	/// Return central cell
	/// ATTENTION: Abuses idx
	cell_ptr central_cell() {
		set_idx_leaf_count();
		return central_cell_worker(0);
	}

	cell_ptr central_cell_worker(int offset = 0) {
		// The whole offset shizzle is to make sure,
		// we are "spiraling" into the middle
		// after we hit the point where all cells
		// have the same cell count
		index_type max = 0;
		int i_off = 0;
		cell_ptr max_cell = this;
		for(int i=0; i<(1<<DIM); i++) {
			i_off = (i+offset) % (1<<DIM);
			if( cells[i_off] && cells[i_off]->idx > max ) {
				max_cell = cells[i_off];
				max = max_cell->idx;
			}
		}
		if( max_cell != this ) {
			return max_cell->central_cell_worker(offset+1);
		}
		else return this;
	}


	/// Refines tree at \pt depth levels
	/// \return pointer to deepest cell
	cell_ptr refine_at(const double* pt, int depth) {

			// If depth is 0/negative we are done and return this cell.
			if( depth<=0 ) return cell_ptr(this);

			#if 1
			if( !contains(pt) ) return NULL;
			#endif

			int cell_id = 0;
			for(dim_type i = 0; i < DIM; i++) { // Calculate child, that contains the pt
				if(pt[i] >= cpoint[i]) cell_id += 1<<i;
			}

			// If the child does not exist yet, we refine this cell
			if( !cells[cell_id] ) refine();

			// Refine child
			return cells[cell_id]->refine_at(pt,depth-1);
	}




	/// Generates a neighbor
	void make_neighbor(size_t nbor) {
		if( nbors[nbor] ) return; // Neighbor already exists. Nothing to do.

		// In order to generate a neighbor for us, we have to refine the neighbor
		// of our parent in the same direction.

		// If our parent has no suitable neighbor, we first have to create a neighbor
		// for our parent.
		if( parent && !(parent->nbors[nbor]) ) {
			parent->make_neighbor(nbor);
		}

		// If we didnt hit the outer boundaries, we now should be able
		// to refine our parents neighbor.
		if( parent && parent->nbors[nbor] ) {
			parent->nbors[nbor]->refine();
		}

		return;
	}


	/// Pushes pointers to all leafs of the tree into \p leafvec.
	void leafs(std::vector<cell_ptr> & leafvec) const {
		for(auto cell : cells) {
			if(cell->refined() == 0 ) {
				leafvec.push_back(cell);
			}
			else cell->leafs(leafvec);
		}
	}

	cell_ptr next_nbor(int i, int j) {
		int mask = (1<<DIM)-1;
		int diff = i^j^mask;
		int nbor;
		cell_ptr cell = this;
		for(int d=0;d<DIM;d++) {
			if(((diff>>d) & 1) == 0) continue;
			nbor = (d<<1) + ((i>>d)&1);
			cell = cell->nbors[nbor];
			if(cell == NULL) return NULL;
		}
		return cell;
	}

	inline double grid(const int* idxs) const {
		if(data==NULL) return 0.0;
		size_t offset = 0;
		for(int d=0; d<DIM; d++) {
			//assert(idxs[d]>=0 && idxs[d] < (1<<nexp));
			offset += (1<<(d*nexp)) * idxs[d];
		}
		return data[offset];
	}

	inline double grid_virtual(const int* idxs_in) const {
		const cell<DIM>* cell = this;
		int idxs[DIM];
		for(int d=0; d<DIM; d++) {
			idxs[d] = ( idxs_in[d] + (1<<nexp) ) % (1<<nexp);
			if(idxs_in[d]<0)               cell = cell->nbors[(d<<1)+0];
			else if(idxs_in[d]>=(1<<nexp)) cell = cell->nbors[(d<<1)+1];
			if(cell==NULL) return 0;
		}
		return cell->grid(idxs);
	}

	void get_grid(const double* x, double* out, double* grid, int N) const {
		int NDIM = 1;
		int Nd = 1;
		//for(int n=0; n < NDIM; n++) grid[n] = 0;
		//if(data==NULL) return;//

		for(int d=0; d<DIM; d++) {
			NDIM*=N;
		}
		int indexes[NDIM*DIM];

		int i = 0;
		double y;
		for(int d=0; d < DIM; d++) {
			//assert(x[d]>=0 && x[d]<1);
			y = x[d] * (1<<nexp);
			//out[d]  = fmod(y,1);
			out[d] = y - (int) y;
			for(int n=0; n < NDIM; n++) {
				i = n*DIM+d;
				indexes[i] = (n/Nd % N) - (N-1)/2;
				indexes[i] += (int) y;

			}
			Nd *= N;
		}

		for(int n=0; n < NDIM; n++) {
			grid[n] = grid_virtual(indexes+n*DIM);
		}

	}

	inline bool is_leaf() const {
		return data!=NULL;
	}

	size_t count_leafs() {
		if(is_leaf()) return 1;
		int sum=0;
		for(int i=0; i<1<<DIM; i++) {
			if(cells[i]!=NULL) sum+=cells[i]->count_leafs();
		}
		return sum;
	}

	cell<DIM>** list_leafs(size_t* ncells) {
		*ncells=count_leafs();
		cell<DIM>** cell_list = (cell<DIM>**) malloc(*ncells * sizeof(cell<DIM>*));
		list_leafs(cell_list);
		return cell_list;
	}

	cell<DIM>** list_leafs(cell<DIM>** dest) { /* Worker function */
		if(is_leaf()) {
			*dest=this;
			dest++;
		} else {
			for(int i=0; i<1<<DIM; i++) {
				if(cells[i]!=NULL) {
					dest = cells[i]->list_leafs(dest);
				}
			}
		}
		return dest;
	}



	cell<DIM>** list_cells_at_level(size_t* ncells, int level) {
		*ncells=count_cells_at_level(level);
		cell<DIM>** cell_list = (cell<DIM>**) malloc(*ncells * sizeof(cell<DIM>*));
		list_cells_at_level(cell_list,level);
		return cell_list;
	}

	cell<DIM>** list_cells_at_level(cell<DIM>** dest, int level) { /* Worker function */
		if(level==0) {
			*dest=this;
			dest++;
		} else {
			for(int i=0; i<1<<DIM; i++) {
				if(cells[i]!=NULL) {
					dest = cells[i]->list_cells_at_level(dest,level-1);
				}
			}
		}
		return dest;
	}

	size_t count_cells_at_level(int level) {
		if(level==0) return 1;
		int sum=0;
		for(int i=0; i<1<<DIM; i++) {
			if(cells[i]!=NULL) sum+=cells[i]->count_cells_at_level(level-1);
		}
		return sum;
	}



	void translate(const double* delta) {
		for(int d=0; d<DIM; d++) cpoint[d] += delta[d];
		for(int i=0; i<1<<DIM; i++) {
			if(cells[i]!=NULL) cells[i]->translate(delta);
		}
	}


	cell<DIM>** list_cells(size_t* ncells, int depth) {
		*ncells=count_cells(depth);
		cell<DIM>** cell_list = (cell<DIM>**) malloc(*ncells * sizeof(cell<DIM>*));
		list_cells(cell_list,depth);
		return cell_list;
	}

	cell<DIM>** list_cells(cell<DIM>** dest, int depth) { /* Worker function */
		if(depth<=0) {
			*dest=this;
			dest++;
		} else {
			for(int i=0; i<1<<DIM; i++) {
				if(cells[i]!=NULL) {
					dest = cells[i]->list_cells(dest,depth-1);
				}
			}
		}
		return dest;
	}


	size_t count_cells(int depth) {
		if(depth<=0) return 1;
		int sum=0;
		for(int i=0; i<1<<DIM; i++) {
			if(cells[i]!=NULL) sum+=cells[i]->count_cells(depth-1);
		}
		return sum;
	}
	

	void sample(fun_t psi, const void* fargs, map_t map, const void* margs, double t, int nexp_) {
		dealloc();
		nexp = nexp_;
			
		// Allocate
		int npts = npoints();
		double* coords = (double*) malloc(npts * DIM * sizeof(double));
		alloc(nexp);

		// Get coordinates of points inside cell
		grid_coords(coords);

		// Track coordinates back
		if(map != NULL) {
			map(coords, coords, npts, margs, t);
		}
		
		// Evaluate psi at tracked coordinates
		psi(coords, data, npts, fargs);

		free(coords);
	}

	#if 0
	struct grid_t vgrid(double* data, double* limits, int corner) {
		// Null data for missing neighbors
		double null_data[1<<(DIM*nexp)] = {0};

		// Get pointers to neighbor data
		double* datas[1<<DIM];
		cell_ptr cell;
		for(int i=0; i<1<<DIM; i++) {
			cell = next_nbor(corner,i);
			//if(cell) cell->eps_highlightbox(stdout,0.5);
			datas[i] = (cell!=NULL && cell->data!=NULL) ? cell->data : null_data;
		}

		// Merge datas to output data
		merge(datas,data,DIM,nexp);

		// Calculate limit of grid
		for(int d=0; d < DIM; d++) {
			limits[d] = cpoint[d] - 3*cwidth[d];
			limits[d] += ((corner>>d)&1) * 2 * cwidth[d];
			limits[d+DIM] = limits[d] + 4 * cwidth[d];
		}

		struct grid_t grid = {data,limits,DIM,nexp+1};
		return grid;
	}
	#endif

	/// Returns a vector containing pointers to all leafs of the tree.
	std::vector<cell_ptr> leafs() const {
		std::vector<cell_ptr> leafvec;
		leafs(leafvec);
		return leafvec;
	}


	/// Returns pointer to the leaf which contains the point \p pos.
	/// ATTENTION: If \pos is not within the boundary, the corresponing edge-leaf is returned!!!
	///
	/// \return #idx of the leaf which contains the point \p pos.
	cell_ptr get_leaf(const double* z) {
		unsigned int cell_id = 0;
		for(int i = 0; i < DIM; i++) { // Calculate child, that contains the pt
			if(z[i] >= cpoint[i]) cell_id += 1<<i;
		}

		if( cells[cell_id] ) { // If the child exists
			return cells[cell_id]->get_leaf(z); // Call the child cell
		}

		return this; // if there is no child, we are returned
	}


	/// Pushes all cells \p lvl below this one into \p cont
	/// \p CONT type must implement push_back()
	template <typename CONT>
	void get_cells_at_lvl(CONT* container, int lvl) {
		if( lvl <= 0 ) {
			container->push_back(this);
			return;
		}

		else {
			for(int i=0; i<(1<<DIM); i++) {
				if( cells[i] ) get_cells_at_lvl(container, lvl-1);
			}
		}
	}

	/// Returns true if we are on the boundary
	int is_boundary() {
		for(int nbor=0; nbor<(2*DIM); nbor++) {
			if( !nbors[nbor] ) { // A nbor is missing: This is part of the boundary
				return 1;
			}
		}
		return 0;
	}

	/// Pushes all cells belonging to the boundary into \p cont
	/// \p CONT type must implement push_back()
	/// ATTENTION: ABUSES idx
	template <typename CONT>
	void get_boundary(CONT* container) {

		if( is_boundary() ) { // We are on the boundary
			get_boundary_worker(container); // Call routine hangling along the boundary
			return;
		}

		// We are not yet on the boundary -> go one nbor in the neg first dim and try again
		nbors[0]->get_boundary(container);
	}

	template <typename CONT>
	void get_boundary_worker(CONT* container) {
		if( idx == -1 ) return;

		if( is_boundary() ) { // We indeed are on the boundary and have not yet been visited
			container->push_back(this); // Store us
			idx = -1; // We have been visited
			for(int nbor=0; nbor<(2*DIM); nbor++) {
				if( nbors[nbor] ) nbors[nbor]->get_boundary_worker(container); // Recursion
			}
			return;
		}
		for(int nbor=0; nbor<(2*DIM); nbor++) {
			if( nbors[nbor] && nbors[nbor]->is_boundary() ) nbors[nbor]->get_boundary_worker(container); // Recursion
		}
	}

	void print() {
		printf("cpoint:");
		for(int i=0; i<DIM; i++) printf(" %e",cpoint[i]);
		printf("\n");
		printf("integral: %e\n", integral());
		printf("nexp: %ld\n", nexp);
		printf("data: %p\n", data);
	}


}; // End of class cell<DIM>


} // End of namespace SLV

#endif
