/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "csr.h"
#include <cstdio>
#include <cmath>
#include <cstring>
#include <fftw3.h>
#include <complex.h>

//
// Interpolator functions
//
static double interpolate0(const double* data, double lo, double hi, size_t n, double x) {
	double d = (x-lo)/(hi-lo);
	size_t i = d*n;
	d = d*n - i;
	if(i<0 || i>=n) return 0.0;
	double f0 = data[i];
	return f0;
}
static double interpolate1(const double* data, double lo, double hi, size_t n, double x) {
	double d = (x-lo)/(hi-lo);
	size_t i = d*n;
	d = d*n - i;
	if(i<0 || i>=n) return 0.0;
	double f0 = data[i+0];
	double f1 = data[i+1];
	double res = f0 + d * (f1-f0);
	return res;
}
static double interpolate2(const double* data, double lo, double hi, size_t n, double x) {
	double d = (x-lo)/(hi-lo);
	size_t i = d*n;
	d = d*n - i;
	if(i<0 || i>=n) return 0.0;
	double f0 = data[((i+0)%n+n)%n];
	double f1 = data[((i+1)%n+n)%n];
	double f2 = data[((i+2)%n+n)%n];
	return
	  f0 * (d-1) * (d-2) / (0-1) / (0-2)
	+ f1 * (d-0) * (d-2) / (1-0) / (1-2)
	+ f2 * (d-0) * (d-1) / (2-0) / (2-1);
}
static double interpolate3(const double* data, double lo, double hi, size_t n, double x) {
	double d = (x-lo)/(hi-lo);
	size_t i = d*n;
	d = d*n - i;
	if(i<0 || i>=n) return 0.0;
	double fm1 = data[((i-1)%n+n)%n];
	double f0  = data[((i+0)%n+n)%n];
	double f1  = data[((i+1)%n+n)%n];
	double f2  = data[((i+2)%n+n)%n];
	return
	  fm1 * (d-0) * (d-1) * (d-2) / (-1-0) / (-1-1) / (-1-2)
	+ f0  * (d+1) * (d-1) * (d-2) / ( 0+1) / ( 0-1) / ( 0-2)
	+ f1  * (d+1) * (d-0) * (d-2) / ( 1+1) / ( 1-0) / ( 1-2)
	+ f2  * (d+1) * (d-0) * (d-1) / ( 2+1) / ( 2-0) / ( 2-1);
}

//
// Kernel functions
//
static double intxn13(int n, double a, double b) {
	/* = int_a^b x^(n-1/3) dx */
	return 1./(n+2./3.) * ( powl(b,n+2./3.) - powl(a,n+2./3.) );
}
static double F0(double xi, double a, double b) {
	/* F0 = int 1 * x^(-1/3) dx */
	return intxn13(0,a,b);
}
static void kernel0(double _Complex* out, size_t n, double h, double sl) {
	memset(out, 0, n * sizeof(double _Complex));
	for(size_t i=0; i<n/2; i++) {
		double x0 = (i+0) * h;
		double a = x0;
		double b = (i+1) * h;
		b = b<sl ? b : sl;

		out[i+0] += F0(x0,a,b);
		if(b==sl) return;
	}
}
static double F1(double xi, double xj, double a, double b) {
	/* F1 = int (x-xj)/(xi-xj) * x^(-1/3) dx */
	double a0 = -xj;
	double a1 = 1.0;
	return 1./(xi-xj) * (
		  a1 * intxn13(1,a,b)
		+ a0 * intxn13(0,a,b)
	);
}
static void kernel1(double _Complex* out, size_t n, double h, double sl) {
	memset(out, 0, n * sizeof(double _Complex));
	for(size_t i=0; i<n/2-1; i++) {
		double x0 = (i+0) * h;
		double x1 = (i+1) * h;
		double a = x0;
		double b = x1;
		b = b<sl ? b : sl;

		out[i+0] += F1(x0,x1,a,b);
		out[i+1] += F1(x0,x1,a,b);
		if(b==sl) return;
	}
}
static double F2(double xi, double xj, double xk, double a, double b) {
	/* F2 = int (x-xj)(x-xk)/(xi-xj)/(xi-xk) * x^(-1/3) dx */
	double a0 = xj*xk;
	double a1 = -(xj+xk);
	double a2 = 1;
	return 1./(xi-xj)/(xi-xk) * (
		  a2 * intxn13(2,a,b)
		+ a1 * intxn13(1,a,b)
		+ a0 * intxn13(0,a,b)
	);
}
static void kernel2(double _Complex* out, size_t n, double h, double sl) {
	memset(out, 0, n * sizeof(double _Complex));
	for(size_t i=0; i<n/2-2; i++) {
		double x0 = (i+0) * h;
		double x1 = (i+1) * h;
		double x2 = (i+2) * h;
		double a = x0;
		double b = x1;
		b = b<sl ? b : sl;

		out[i+0] += F2(x0,x1,x2,a,b) ;
		out[i+1] += F2(x1,x0,x2,a,b) ;
		out[i+2] += F2(x2,x0,x1,a,b) ;
		if(b==sl) return;
	}
}
static double F3(double xi, double xj, double xk, double xl,  double a, double b) {
	/* F3 = int (x-xj)(x-xk)(x-xl)/(xi-xj)/(xi-xk)/(xi-xl) * x^(-1/3) dx */
	double a0 = -xj*xk*xl;
	double a1 = xj*xk + xj*xl + xk*xl;
	double a2 = -(xj + xk + xl);
	double a3 = 1;
	return 1./(xi-xj)/(xi-xk)/(xi-xl) * (
		  a3 * intxn13(3,a,b)
		+ a2 * intxn13(2,a,b)
		+ a1 * intxn13(1,a,b)
		+ a0 * intxn13(0,a,b)
	);
}
static void kernel3(double _Complex* out, size_t n, double h, double sl) {
	memset(out, 0, n * sizeof(double _Complex));
	for(size_t i=0; i<n/2-3; i++) {
		double xm1 = (i-1) * h;
		double x0  = (i+0) * h;
		double x1  = (i+1) * h;
		double x2  = (i+2) * h;
		double a = x0;
		double b = x1;
		b = b<sl ? b : sl;

		out[(i-1)>=0 ? i-1 : n-i]
		         += F3(xm1,x0,x1,x2,a,b) ;
		out[i+0] += F3(x0,xm1,x1,x2,a,b) ;
		out[i+1] += F3(x1,xm1,x0,x2,a,b) ;
		out[i+2] += F3(x2,xm1,x0,x1,a,b) ;
		if(b==sl) return;
	}
}

// kernel array
static void (*kernel[])(double _Complex*, size_t, double, double)
	= { &kernel0, &kernel1, &kernel2, &kernel3 };

// interpolator array
double (*interpolate[])(const double*, double, double, size_t, double)
	= { &interpolate0, &interpolate1, &interpolate2, &interpolate3 };

static double cardano(double p, double q) {
	double y =  p/3.;
	double x = -q/2.;
	double r = 0;
	r += cbrt(x + sqrt(x*x + y*y*y));
	r += cbrt(x - sqrt(x*x + y*y*y));
	return r;
}
#if 0
static void kernel_phi(double _Complex* out, size_t n, double h, double sl) {
	memset(out, 0, n * sizeof(double _Complex));
	for(size_t i=0; i<n/2; i++) {
		double sh = h*i; // \hat{s} - \hat{s}'
		if(0==i) sh = 1e-6;
		double uh = cardano(12., -24*sh);
		double uh2 = uh*uh;
		double x = (uh2/4.+1);
		double y = (uh2/12.+1);
		double phi = -(uh2-4) / 2 / (x*x*x)
		             -4/uh2 * ( (3*uh2/4+1)/(x*x*x) - 1/(y*y));
		out[i] = phi;
		if(sh>sl) return;
	}
}
#endif


static inline double W(double x){
	return x<=0.0 ? 0.0 : (cosh(5/3.*asinh(x))-cosh(asinh(x)))/sinh(2*asinh(x)) * 9./4.;
}
static void kernel_murphy(double _Complex* out, size_t n, double h, double sl) {
	memset(out, 0, n * sizeof(double _Complex));
	for(size_t i=0; i<n/2; i++) {
		out[i] = W((i+1)*h) - W((i+0)*h);
		if(h*i>sl) break;
	}
}

static double theta(double x) {
	double y = x*x+2;
	return 2*x*y/(y*y-1);
}
static void kernel_theta(double _Complex* out, size_t n, double h, double sl, double gamma, double radius) {
	memset(out, 0, n * sizeof(double _Complex));
	for(size_t i=0; i<n/2; i++) {
		double a = (i+0)*h;
		double b = (i+1)*h;
		double muhat_a = cardano(3, 3*gamma*gamma*gamma/radius * a);
		double muhat_b = cardano(3, 3*gamma*gamma*gamma/radius * b);
		out[i] = theta(muhat_b) - theta(muhat_a);
		out[i] *= -1; // TODO: factor not yet in theory
		if(h*i>sl) break;
	}
}




void solve_csr(const double* data, double* out, size_t npts, const void* args_, char* debug) {
	solve_csr_args* args = (solve_csr_args*) args_;

	const double T = args->T;
	const double length = args->length;
	const double angle = args->angle;
	const double s = args->s;
	const double gamma = args->gamma;
	const size_t npad = args->npad>0?args->npad:1;
	const int    mode = args->mode;
	const int    filter = args->filter;
	const double filterwidth = args->filterwidth;
	const double large_dist_cutoff = args->large_dist_cutoff;
	const int    transient = args->transient;

	const double radius = fabs(length / sin(angle));
	const double phi = s/radius;
	const double sl = radius * phi * phi * phi / 24; // Slippage length
	const size_t N = npad*npts;
	const double h = T/npts;

	double _Complex rho_s[N] = {0};
	double _Complex rho_f[N] = {0};
	double _Complex sol_s[N] = {0};
	double _Complex sol_f[N] = {0};
	double _Complex grn_s[N] = {0};
	double _Complex grn_f[N] = {0};
	double filter_f[N] = {0};
	double T1[npts] = {0};
	double T2[npts] = {0};

	// copy rho data into 0-padded array
	for(size_t i=0; i < npts; i++) {
		rho_s[i] = data[i];
	}
	
	// calculate kernel
	if(mode<=3) {
		kernel[mode](grn_s, N, h, sl);
		double slarge = radius/(gamma*gamma*gamma) * large_dist_cutoff;
		size_t nlarge = floor(slarge/h);
		nlarge = nlarge > N ? N : nlarge;
		for(size_t i=0; i<nlarge; i++) {
			grn_s[i] = 0.0;
		}
	} else if(mode==4) { // murphy
		double norm = gamma*gamma*gamma/radius *3./2.;
		kernel_murphy(grn_s, N, h*norm, sl*norm);
	} else if(mode==5) { // theta
		kernel_theta(grn_s, N, h, sl, gamma, radius);
	}

	// calculate fourier transforms
	fftw_plan p;
	p = fftw_plan_dft_1d(N, (fftw_complex*) rho_s, (fftw_complex*) rho_f, FFTW_FORWARD, FFTW_ESTIMATE);
	fftw_execute(p); fftw_destroy_plan(p);

	p = fftw_plan_dft_1d(N, (fftw_complex*) grn_s, (fftw_complex*) grn_f, FFTW_FORWARD, FFTW_ESTIMATE);
	fftw_execute(p); fftw_destroy_plan(p);

	// calculate filter
	switch(filter) {
	case 0: // no filter
		for(size_t n=0; n<N; n++) {
			filter_f[n] = 1;
		}
		break;
	case 1: // low-pass filter
		for(size_t n=0; n<N; n++) {
			double k = 2.0 * M_PI * ( (n <= N/2) ? 1.0*n : -1.0*(1.0*N-1.0*n) ) / N;
			filter_f[n] = fabs(k/M_PI)<=fabs(filterwidth) ? 1 : 0;
		}
		break;
	case 2: // Gaussian filter
		for(size_t n=0; n<N; n++) {
			double k = 2.0 * M_PI * ( (n <= N/2) ? 1.0*n : -1.0*(1.0*N-1.0*n) ) / N;
			filter_f[n] = exp(-0.5*pow((k/M_PI/filterwidth),2));
		}
		break;
	}

	// calculate convolution in frequency space
	if(mode <= 3) {
		for(size_t n=0; n<N; n++) {
			double k = 2.0 * M_PI * ( (n <= N/2) ? 1.0*n : -1.0*(1.0*N-1.0*n) ) / N;
			sol_f[n] =  rho_f[n] * grn_f[n] * filter_f[n] * I * k / h;
			// 1/h comes from derivative;
			// NO h in sol_f because we integrated analytically
		}
	} else if(mode == 4) { // murphy
		for(size_t n=0; n<N; n++) {
			sol_f[n] =  rho_f[n] * grn_f[n] * filter_f[n] 
			 // * pow(gamma,4)/radius/radius * (4./3.)
			 // / (gamma*gamma*gamma/radius *3./2.)
			 * gamma / radius * 8./9.;
		}
	} else if(mode == 5) { // theta
		for(size_t n=0; n<N; n++) {
			sol_f[n] =  rho_f[n] * grn_f[n] * filter_f[n]
			* gamma / radius;
		}
	}

	// fourier transform back
	p = fftw_plan_dft_1d(N, (fftw_complex*) sol_f, (fftw_complex*) sol_s, FFTW_BACKWARD, FFTW_ESTIMATE);
	fftw_execute(p); fftw_destroy_plan(p);

	// write real part of solution into T1
	const double K = -2*pow(3*radius*radius,-1./3.); // As defined in Borland 2001
	for(size_t n=0; n<npts; n++) {
		T1[n] = creal(sol_s[n])/N * K;
	}

	// calculate transient term T2
	if(transient) {
		for(size_t n=0; n<npts; n++) {
			double s1 = n*h - sl;
			double s4 = n*h - 4*sl;
			if(npad==1) { // prediodic boundary
				s1 = fmod(fmod(s1,T)+T,T);
				s4 = fmod(fmod(s4,T)+T,T);
			}
			T2[n] = 0;
			if(transient == 1) {
				T2[n] += interpolate[mode > 3 ? 3 : mode](data, 0, T, npts, s1);
				T2[n] -= interpolate[mode > 3 ? 3 : mode](data, 0, T, npts, s4);
				T2[n] *= K/pow(sl,1./3.);
			}
			if(transient == 2) {
				T2[n] += interpolate[mode > 3 ? 3 : mode](data, 0, T, npts, s1);
				T2[n] *= K/pow(sl,1./3.);
			}
		}
	}


	// write output array
	for(size_t n=0; n<npts; n++) {
		out[n] = T1[n] + T2[n];
	}

	// write debug file
	if(debug) {
		FILE* fd;
		fd = fopen(debug,"w");
		fprintf(fd,"q k nh ");
		fprintf(fd,"rho_s_real rho_s_imag rho_f_real rho_f_imag ");
		fprintf(fd,"grn_s_real grn_s_imag grn_f_real grn_f_imag ");
		fprintf(fd,"sol_s_real sol_s_imag sol_f_real sol_f_imag ");
		fprintf(fd,"filter_f ");
		fprintf(fd,"T1 T2");
		fprintf(fd,"\n");
		for(size_t n=0; n<N; n++) {
			double q = n*T/npts-T/2.;
			double k = 2.0 * M_PI * ( (n <= N/2) ? 1.0*n : -1.0*(1.0*N-1.0*n) ) / N / h;
			fprintf(fd,"%+e %+e %+e ", q, k, n*h);
			fprintf(fd,"%+e %+e %+e %+e ",  creal(rho_s[n]), cimag(rho_s[n]), creal(rho_f[n]), cimag(rho_f[n]));
			fprintf(fd,"%+e %+e %+e %+e ",  creal(grn_s[n]), cimag(grn_s[n]), creal(grn_f[n]), cimag(grn_f[n]));
			fprintf(fd,"%+e %+e %+e %+e ",  creal(sol_s[n]), cimag(sol_s[n]), creal(sol_f[n]), cimag(sol_f[n]));
			fprintf(fd,"%+e ", filter_f[n]);
			if(n<npts) fprintf(fd,"%+e %+e", T1[n], T2[n]);
			fprintf(fd,"\n");
		}
		fclose(fd);

		//sprintf(buf,"%s-csr-rho.dat",prefix);
		//fd = fopen(buf,"w");
		//for(size_t n=0; n<N; n++) {
		//	fprintf(fd,"%+e ",  creal(rho_s[n]));
		//	fprintf(fd,"%+e ",  cimag(rho_s[n]));
		//	fprintf(fd,"%+e ",  creal(rho_f[n]));
		//	fprintf(fd,"%+e ",  cimag(rho_f[n]));
		//	fprintf(fd,"\n");
		//}
		//fclose(fd);

		//sprintf(buf,"%s-csr-grn.dat",prefix);
		//fd = fopen(buf,"w");
		//for(size_t n=0; n<N; n++) {
		//	fprintf(fd,"%+e ",  creal(grn_s[n]));
		//	fprintf(fd,"%+e ",  cimag(grn_s[n]));
		//	fprintf(fd,"%+e ",  creal(grn_f[n]));
		//	fprintf(fd,"%+e ",  cimag(grn_f[n]));
		//	fprintf(fd,"\n");
		//}
		//fclose(fd);

		//sprintf(buf,"%s-csr-sol.dat",prefix);
		//fd = fopen(buf,"w");
		//for(size_t n=0; n<N; n++) {
		//	fprintf(fd,"%+e ",  creal(sol_s[n]));
		//	fprintf(fd,"%+e ",  cimag(sol_s[n]));
		//	fprintf(fd,"%+e ",  creal(sol_f[n]));
		//	fprintf(fd,"%+e ",  cimag(sol_f[n]));
		//	fprintf(fd,"\n");
		//}
		//fclose(fd);

		//sprintf(buf,"%s-csr-T1T2.dat",prefix);
		//fd = fopen(buf,"w");
		//for(size_t n=0; n<npts; n++) {
		//	fprintf(fd,"%+e ",  n*T/npts-T/2.);
		//	fprintf(fd,"%+e ",  T1[n]);
		//	fprintf(fd,"%+e ",  T2[n]);
		//	fprintf(fd,"\n");
		//}
		//fclose(fd);
	}
}
