/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "csr.h"
#include <cstdio>
#include <cmath>
#include <fftw3.h>
#include <complex.h>

void solve_csr_old(const double* data, double* out, size_t npts, const void* args_, char* prefix) {
	solve_csr_args* args = (solve_csr_args*) args_;

	const double T = args->T;
	const double length = args->length;
	const double angle = args->angle;
	const double s = args->s;
	const size_t npad = args->npad>0?args->npad:1;

	const double radius = fabs(length / sin(angle));
	const double phi = s/radius;
	const double sl = radius * phi * phi * phi / 24; // Slippage length
	const size_t N = npad*npts;
	const double h = T/npts;

	double _Complex rho_s[N] = {0};
	double _Complex rho_f[N] = {0};
	double _Complex sol_s[N] = {0};
	double _Complex sol_f[N] = {0};
	double _Complex grn_s[N] = {0};
	double _Complex grn_f[N] = {0};
	double T1[npts] = {0};
	double T2[npts] = {0};

	// copy rho data into 0-padded array
	for(size_t i=0; i < npts; i++) {
		rho_s[i] = data[i];
	}

	// calculate greens function
	double svals[N] = {0};
	for(size_t i=0; i<N; i++) {
		double s = h*((double) i - (i<=N/2 ? 0 : N));
		svals[i] = s;
		grn_s[i] = s>0.0 && s<=sl ? pow(fabs(s),-1.0/3.0) : 0 ; // TODO: Why s>0.0??
		grn_s[i] *= h;
		// 2021-10-27 investigated the s>0.0 matter; this is indeed the
		// correct way to place the kernel
	}

	// calculate fourier transforms
	fftw_plan p;
	p = fftw_plan_dft_1d(N, (fftw_complex*) rho_s, (fftw_complex*) rho_f, FFTW_FORWARD, FFTW_ESTIMATE);
	fftw_execute(p); fftw_destroy_plan(p);

	p = fftw_plan_dft_1d(N, (fftw_complex*) grn_s, (fftw_complex*) grn_f, FFTW_FORWARD, FFTW_ESTIMATE);
	fftw_execute(p); fftw_destroy_plan(p);

	// calculate convolution in frequency space
	for(size_t n=0; n<N; n++) {
		double k = 2.0 * M_PI * ( (n <= N/2) ? 1.0*n : -1.0*(1.0*N-1.0*n) ) / N / h;
		sol_f[n] = I * k * rho_f[n] * grn_f[n];
		// amstutz 2021-11-01: moved *h into grn_s, to make comparable w/ new csr routine
		// 1/h in k comes from derivative
	}

	// fourier transform back
	p = fftw_plan_dft_1d(N, (fftw_complex*) sol_f, (fftw_complex*) sol_s, FFTW_BACKWARD, FFTW_ESTIMATE);
	fftw_execute(p); fftw_destroy_plan(p);

	// write real part of solution into T1
	const double K = -2*pow(3*radius*radius,-1./3.); // As defined in Borland 2001
	for(size_t n=0; n<npts; n++) {
		T1[n] = creal(sol_s[n])/N * K;
	}


	// calculate T2
	for(size_t n=0; n<npts; n++) {
		const int nsl = sl/h;
		int i;
		i = n - nsl;	
		i = npad==1 ? (i%npts+npts)%npts : i; // periodic boundary
		T2[n]  = i>=0 ? data[i] : 0;
		i = n - 4*nsl;	
		i = npad==1 ? (i%npts+npts)%npts : i; // periodic boundary
		T2[n] -= i>=0 ? data[i] : 0;
		T2[n] *= K/pow(sl,1./3.);
	}


	// write output array
	for(size_t n=0; n<npts; n++) {
		out[n] = T1[n] + T2[n];
	}

	// write debug file
	if(prefix) {
		FILE* fd;
		char buf[1024];

		sprintf(buf,"%s-csr-rho.dat",prefix);
		fd = fopen(buf,"w");
		for(size_t n=0; n<N; n++) {
			fprintf(fd,"%+e ",  creal(rho_s[n]));
			fprintf(fd,"%+e ",  cimag(rho_s[n]));
			fprintf(fd,"%+e ",  creal(rho_f[n]));
			fprintf(fd,"%+e ",  cimag(rho_f[n]));
			fprintf(fd,"\n");
		}
		fclose(fd);

		sprintf(buf,"%s-csr-grn.dat",prefix);
		fd = fopen(buf,"w");
		for(size_t n=0; n<N; n++) {
			fprintf(fd,"%+e ",  creal(grn_s[n]));
			fprintf(fd,"%+e ",  cimag(grn_s[n]));
			fprintf(fd,"%+e ",  creal(grn_f[n]));
			fprintf(fd,"%+e ",  cimag(grn_f[n]));
			fprintf(fd,"\n");
		}
		fclose(fd);

		sprintf(buf,"%s-csr-sol.dat",prefix);
		fd = fopen(buf,"w");
		for(size_t n=0; n<N; n++) {
			fprintf(fd,"%+e ",  creal(sol_s[n]));
			fprintf(fd,"%+e ",  cimag(sol_s[n]));
			fprintf(fd,"%+e ",  creal(sol_f[n]));
			fprintf(fd,"%+e ",  cimag(sol_f[n]));
			fprintf(fd,"\n");
		}
		fclose(fd);

		sprintf(buf,"%s-csr-T1T2.dat",prefix);
		fd = fopen(buf,"w");
		for(size_t n=0; n<npts; n++) {
			fprintf(fd,"%+e ",  n*T/npts-T/2.);
			fprintf(fd,"%+e ",  T1[n]);
			fprintf(fd,"%+e ",  T2[n]);
			fprintf(fd,"\n");
		}
		fclose(fd);
	}
}
