/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "eps.h"
#include "psd_eval.h"


namespace SLV {

void eps_startbox(const cell<2>* cell, FILE* file, double* bbox = NULL) {
	double xlo = cell->cpoint[0] - cell->cwidth[0];
	double ylo = cell->cpoint[1] - cell->cwidth[1];
	double xhi = cell->cpoint[0] + cell->cwidth[0];
	double yhi = cell->cpoint[1] + cell->cwidth[1];
	if(bbox!=NULL) {
		if(xlo>bbox[2] || xhi<bbox[0] || ylo>bbox[3] || yhi<bbox[1]) return;
	}
	fprintf(file,"gsave\n");
	fprintf(file,"%+e %+e %+e %+e box\n",xlo,ylo,xhi,yhi);
}


void eps_endbox(FILE* file) {
	fprintf(file,"grestore\n");
}


void eps_drawbox(const cell<2>* cell, FILE* file) {
	eps_startbox(cell, file);
	fprintf(file,"0.0 setgray\n");
	fprintf(file,"drawbox\n");
	eps_endbox(file);
}


void eps_highlightbox(const cell<2>* cell, FILE* file, double h) {
	eps_startbox(cell,file);
	fprintf(file,"%f setgray\n",h);
	fprintf(file,"highlightbox\n");
	eps_endbox(file);
}


void eps_highlightpoints(FILE* file, const double* z, int npts) {
	for(int i=0; i<npts; i++) {
		fprintf(file,"%e %e highlightpoint\n",z[0],z[1]);
		z+=2;
	}
}


void eps_drawtree(const cell<2>* cell, FILE* file, double* bbox, int depth) {
	if(depth==1 || cell == NULL) return;
	eps_drawbox(cell, file);
	for(int i=0; i<(1<<2); i++) {
		eps_drawtree(cell->cells[i], file, bbox, depth-1);
	}
}

void eps_header(FILE* file, double scale) {
	fprintf(file,
		"%%!PS-Adobe-3.0 EPSF-3.0\n"
		"%%%%BoundingBox: %e %e %e %e\n",-0.2*scale,-0.1*scale,1.05*scale,1.05*scale
	);
	fprintf(file,
		"\n\n"
		"/show_box true def\n"
		"/show_density true def\n"
		"\n\n"
	);
	fprintf(file,
		"/box {\n"
		"/pmax exch def /qmax exch def /pmin exch def /qmin exch def\n"
		"qmin pmin translate\n"
		"qmax qmin sub pmax pmin sub scale\n"
		"} def\n\n"

		"/drawbox {\n"
		"newpath\n"
		"0.0001 setlinewidth\n"
		"0 0 moveto 1 0 lineto 1 1 lineto 0 1 lineto closepath stroke\n"
		"} def\n"

		"show_box not {/drawbox {} def } if\n\n"
		"/drawdensity {data N N 8 [N 0 0 N 0 0] {cmap} false 3 colorimage} def\n"
		"show_density not {/drawdensity {} def } if\n\n"

		"/highlightbox {\n"
		"newpath\n"
		"0.0001 setlinewidth\n"
		"0.5 setgray\n"
		"0 0 moveto 1 0 lineto 1 1 lineto 0 1 lineto closepath fill\n"
		"} def\n"

		"/highlightpoint {\n"
		"gsave\n"
		"newpath\n"
		"0.0001 setlinewidth\n"
		"1 0 0 setrgbcolor\n"
		"0.002 0 360 arc\n"
		"closepath fill\n"
		"grestore\n"
		"} def\n"

		"\n\n"
	);
	fprintf(file,
		"/width  { %e mul } def\n"
		"/height { %e mul } def\n\n",scale, scale
	);
}


void eps_axis(FILE* file, double* bbox) {
	double dx=(bbox[2]-bbox[0])/4;
	fprintf(file,
		"gsave\n"
		"0 width -0.02 height translate\n"
		"newpath\n"
		"0 0 moveto 1 width 0 height lineto\n"
		"0 0.25 1 { width -0.01 height moveto 0 width 0.02 height rlineto} for\n"
		"stroke\n"
		"-0.11 width -0.06 height translate\n"
		"/Courier findfont 0.03 height scalefont setfont\n"
		"newpath\n"
		"0.00 width 0.0 height moveto (% 0.2e) show\n"
		"0.25 width 0.0 height moveto (% 0.2e) show\n"
		"0.50 width 0.0 height moveto (% 0.2e) show\n"
		"0.75 width 0.0 height moveto (% 0.2e) show\n"
		"1.00 width 0.0 height moveto (% 0.2e) show\n"
		"closepath\n"
		"grestore\n\n", bbox[0]+0*dx, bbox[0]+1*dx, bbox[0]+2*dx, bbox[0]+3*dx, bbox[0]+4*dx
	);

	double dy=(bbox[3]-bbox[1])/4;
	fprintf(file,
		"gsave\n"
		"-0.02 width 0 height translate\n"
		"newpath\n"
		"0 0 moveto 0 width 1 height lineto\n"
		"0 0.25 1 { -0.01 width exch height moveto 0.02 width 0 height rlineto} for\n"
		"stroke\n"
		"-0.18 width 0 height translate\n"
		"/Courier findfont 0.03 height scalefont setfont\n"
		"newpath\n"
		"0.0 width 0.00 height moveto (% 0.2e) show\n"
		"0.0 width 0.25 height moveto (% 0.2e) show\n"
		"0.0 width 0.50 height moveto (% 0.2e) show\n"
		"0.0 width 0.75 height moveto (% 0.2e) show\n"
		"0.0 width 1.00 height moveto (% 0.2e) show\n"
		"closepath\n"
		"grestore\n\n", bbox[1]+0*dy, bbox[1]+1*dy, bbox[1]+2*dy, bbox[1]+3*dy, bbox[1]+4*dy
	);
}


void eps_draw(const psd<2>* psi, FILE* file, double* bbox = NULL, int depth = 0, int ppd = 0) {
	const cell<2>* root= psi->root;
	double scale = 72*6;
	double treebbox[] = {
		root->cpoint[0]-root->cwidth[0],
		root->cpoint[1]-root->cwidth[1],
		root->cpoint[0]+root->cwidth[0],
		root->cpoint[1]+root->cwidth[1]
	};
	if(bbox==NULL) bbox=&treebbox[0];
	//printf("bbox: %e %e %e %e\n",bbox[0],bbox[1],bbox[2],bbox[3]);

	eps_header(file,scale);
	eps_axis(file,bbox);

	fprintf(file,
		"gsave\n\n"
		"newpath\n"
		"0 0 moveto\n"
		"1 width 0 height lineto\n"
		"1 width 1 height lineto\n"
		"0 width 1 height lineto\n"
		"closepath clip\n\n"
	);
	#if 1
	if(ppd>0) {
		double fvals[ppd*ppd];
		double zin[ppd*ppd*2];

		double* z = zin;
		for(int j=0; j<ppd; j++) {
			for(int i=0; i<ppd; i++,z+=2) {
				z[0] = bbox[0]+i*(bbox[2]-bbox[0])/(ppd);
				z[1] = bbox[1]+j*(bbox[3]-bbox[1])/(ppd);
			}
		}

		psd_eval<2>(zin,fvals,ppd*ppd,psi);

		double max=-INFINITY;
		double min=INFINITY;
		for(int c=0; c<ppd*ppd; c++) {
			max=fvals[c]>max?fvals[c]:max;
			min=fvals[c]<min?fvals[c]:max;
		}

		fprintf(file,"gsave \n1 width 1 height scale\n");
		fprintf(file,"/N %d def\n",ppd);
		fprintf(file,
			"/crange {\n"
				"/min %e def\n"
				"/max %e def\n"
				"min sub max min sub div\n"
			"} def\n",
		0.,max);
		fprintf(file,
			"/cmap {\n"
				"crange\n"
				"/val exch def\n"
				"/color 3 string def\n"
				"val 0 eq {<ffffff>} {\n"
				"color 0 val sqrt 255 mul cvi put\n"
				"color 1 val dup dup mul mul 255 mul cvi put\n"
				"color 2 val 360 mul sin dup 0 lt {pop 0} if 255 mul cvi put\n"
				"color\n"
				"} ifelse\n"
			"} def\n"
		);
		fprintf(file,"/data {");
		for(int i=ppd*ppd-1; i>=0; i--) {
			fprintf(file,"%e ",fvals[i]);
			if(0==i%16) fprintf(file,"\n");
		}
		fprintf(file,"} def\n");
		fprintf(file,"drawdensity\n");
		fprintf(file,"grestore\n\n");
	}
	#endif
	fprintf(file,
		"%e %e scale\n",
			1./(bbox[2]-bbox[0])*scale,
			1./(bbox[3]-bbox[1])*scale
	);
	fprintf(file, "%e %e translate\n", -bbox[0], -bbox[1]);
	eps_drawtree(root,file,bbox,depth);
}



}// Namespace
