/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <cmath>
#include <functions.h>

void disk2D(const double* in, double* out, int N, const void* arg) {
	const double* z = in;
	double x;
	double y;
	double c = *((double*) arg);
	for(int i=0; i<N; i++) {
		x=z[0]/c;
		y=z[1]/c;
		*out = x*x+y*y>1?0:1;
		z+=2; out++;
	}
}

void smiley2D(const double* in, double* out, int N, const void* arg) {
	const double* z = in;
	double x;
	double y;
	double r;
	double c = *((double*) arg);
	for(int i=0; i<N; i++) {
		// Head
		x=z[0]/c*2;
		y=z[1]/c*2;
		r=sqrt(x*x+y*y);
		*out = r>1?0:exp(-0.5*r/0.33/0.33);

		// Right eye
		x=(z[0]/c-0.2)/0.1;
		y=(z[1]/c-0.2)/0.1;
		r=sqrt(x*x+y*y);
		*out = r<1?0:*out;

		// Left eye
		//x=(z[0]/c+0.2)/0.2;
		//y=(z[1]/c-0.2)/0.2;
		//r=sqrt(x*x+y*y);
		//*out += r<1?exp(-0.5*r/0.33/0.33):0;

		// Mouth
		x=(z[0])/c;
		y=(z[1]+0.2)/c;
		double w = 0.3;
		double t = fabs((x-w)/2/w);
		if(fabs(x)<w&&fabs(y)<0.1) {
			*out=t;
			if(t!=1 or false) {
				*out *= 1+0.5*sin(1/(1-t)*2*M_PI);
			}
		}

		z+=2; out++;
	}
}

void gauss2D(const double* in, double* out, int N, const void* __arg) {
	const double* z = in;
	gauss2D_args* arg = (gauss2D_args*) __arg;
	double S[2][2] = {
		{arg->sig_q*arg->sig_q,0},
		{0,arg->sig_p*arg->sig_p}
	};
	S[0][1] = S[1][0] = arg->correlation * arg->sig_q * arg->sig_p;
	double det = S[0][0]*S[1][1]-S[0][1]*S[1][0];
	double SI[2][2] = { /* Inverse covariance matrix */
		{ S[1][1]/det,-S[0][1]/det},
		{-S[1][0]/det, S[0][0]/det}
	};
	for(int i=0; i<N; i++) {
		double x = z[0];
		double y = z[1];
		double r = x*(x*SI[0][0]+y*SI[0][1]) +
                           y*(x*SI[1][0]+y*SI[1][1]);
		if(sqrt(r)>arg->cutoff) {
			*out = 0.0;
		} else {
			*out = exp(-0.5*(r))/(2*M_PI)/sqrt(det);
		}
		z+=2; out++;
	}
}

void rectangle(const double* in, double* out, int N, const void* arg) {
	const double* z = in;
	const double* limits = (const double*) arg;
	const double wx = (limits[2]-limits[0]);
	const double wy = (limits[3]-limits[1]);
	for(int i=0; i<N; i++) {
		double dx = z[0]-limits[0];
		double dy = z[1]-limits[1];
		if( dx<=wx && dx>=0 && dy<=wy && dy>=0 ) {
			*out = 1.0/wx/wy;
		} else { 
			*out = 0.0;
		}
		z+=2; out++;
	}
}

void monomial(const double* val, const double* in, double* out, int N, const void* args_) {
	const double* z = in;
	struct monomial_args* args = (struct monomial_args*) args_;
	for(int i=0; i<N; i++) {
		double v = val[i];
		for(int d=0; d<args->dim; d++) v*=pow(z[d],args->exps[d]);
		*out = v;
		z+=2; out++;
	}
}


static double bump__(double x, double a, double d) {
	/*
	*     ____________
	*    /      |     \
	* __/       |      \___
	*           0       a
	*                 |d|
	*/
	double I=2*(a-d/2);
	x=fabs(x);
	if(x>=a)   return 0.0;
	if(x<=a-d) return 1/I;
	x=(x-(a-d))/(d);
	return exp(1.-1./(1.-x*x))/I;

}
static double gauss__(double x, double sigma, double cutoff) {
	if(fabs(x)>cutoff*sigma) return 0;
	double y = x/sigma;
	return 1/sqrt(2*M_PI)/sigma * exp(-0.5*y*y);
}
void bump(const double* in, double* out, int N, const void* args) {
	// Args is a pointer to 4 doubles
	// sigma_p,cutoff, a, d
	const double* z = in;
	double sigma  = ((double*) args)[0];
	double cutoff = ((double*) args)[1];
	double a      = ((double*) args)[2];
	double d      = ((double*) args)[3];
	double chirp  = ((double*) args)[4];
	for(int i=0; i<N; i++) {
		*out = bump__(z[0],a,d) * gauss__(z[1]-chirp*z[0],sigma,cutoff);
		z+=2; out++;
	}
}
