/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

static double linear(double f0, double f1, double x) {
		return f0 + x * (f1-f0);
}
double interpolate_linear(const double* grid, const double* x, int dim) {
	dim-=1;
	const double* f0 = grid;
	const double* f1 = grid+(1<<dim);
	if(dim==0) return linear(*f0,*f1,x[0]);
	return linear(
		interpolate_linear(f0,x,dim),
		interpolate_linear(f1,x,dim),
		x[dim]);
}



static double cubic(double f0, double f1, double f2, double f3, double x) {
	double res =
	f0 * (1+x-1) * (1+x-2) * (1+x-3)
	   / (-1 * -2 * -3)
	+
	f1 * (1+x-0) * (1+x-2) * (1+x-3)
	   / (1 * -1 * -2)
	+
	f2 * (1+x-0) * (1+x-1) * (1+x-3)
	   / (2 * 1 * -1)
	+
	f3 * (1+x-0) * (1+x-1) * (1+x-2)
	   / (3 * 2 * 1);
	//return res>0.0?res:0.0;
	return res;
}
double interpolate_cubic(const double* grid, const double* x, int dim) {
	dim-=1;
	int offset = (1<<(2*dim));
	const double* f0 = grid+0*offset;
	const double* f1 = grid+1*offset;
	const double* f2 = grid+2*offset;
	const double* f3 = grid+3*offset;
	if(dim==0) return cubic(*f0,*f1,*f2,*f3,x[0]);
	return cubic(
		interpolate_cubic(f0,x,dim),
		interpolate_cubic(f1,x,dim),
		interpolate_cubic(f2,x,dim),
		interpolate_cubic(f3,x,dim),
		x[dim]
	);
}



static double p10(double x) {
	return (-2.0*x+3)*x*x;
}
static double p11(double x) {
	return (x-1)*x*x;
}
static double N4C1(double f0, double f1, double f2, double f3, double x) {
	const double y0p0 = f1;
	const double y1p0 = f2;
	const double y0p1 = (-f0+f2)/2;
	const double y1p1 = (-f1+f3)/2;

	return
		y1p0 *  p10(x) +
		y1p1 *  p11(x) +
		y0p0 *  p10(-x+1) +
		y0p1 * -p11(-x+1);
}
double interpolate_N4C1(const double* grid, const double* x, int dim) {
	dim-=1;
	int offset = (1<<(2*dim));
	const double* f0 = grid+0*offset;
	const double* f1 = grid+1*offset;
	const double* f2 = grid+2*offset;
	const double* f3 = grid+3*offset;
	if(dim==0) return N4C1(*f0,*f1,*f2,*f3,x[0]);
	return N4C1(
		interpolate_N4C1(f0,x,dim),
		interpolate_N4C1(f1,x,dim),
		interpolate_N4C1(f2,x,dim),
		interpolate_N4C1(f3,x,dim),
		x[dim]
	);
}

static double p20(double x) { return ((6*x-15)*x+10)*x*x*x; }
static double p21(double x) { return ((-3*x+7)*x-4)*x*x*x; }
static double p22(double x) { return ((0.5*x-1)*x+0.5)*x*x*x; }
static double N4C2(double f0, double f1, double f2, double f3, double x) {
	const double y0p0 = f1;
	const double y1p0 = f2;
	const double y0p1 = (-f0+f2)/2;
	const double y1p1 = (-f1+f3)/2;
	const double y0p2 = f0-2*f1+f2;
	const double y1p2 = f1-2*f2+f3;
	return
		y1p0 *  p20(x) +
		y1p1 *  p21(x) +
		y1p2 *  p22(x) +
		y0p0 *  p20(-x+1) +
		y0p1 * -p21(-x+1) +
		y0p2 *  p22(-x+1);
}
double interpolate_N4C2(const double* grid, const double* x, int dim) {
	dim-=1;
	int offset = (1<<(2*dim));
	const double* f0 = grid+0*offset;
	const double* f1 = grid+1*offset;
	const double* f2 = grid+2*offset;
	const double* f3 = grid+3*offset;
	if(dim==0) return N4C2(*f0,*f1,*f2,*f3,x[0]);
	return N4C2(
		interpolate_N4C2(f0,x,dim),
		interpolate_N4C2(f1,x,dim),
		interpolate_N4C2(f2,x,dim),
		interpolate_N4C2(f3,x,dim),
		x[dim]
	);
}
