/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "maps.h"

#include <math.h>


namespace SLV {


void map_compose(const double* z, double* out, int npts, const void* args, double t) {
	map_compose_args arg = *(map_compose_args*) args;
	size_t n = arg.n;
	if(t>=0) {
		arg.maps[0](z,out,npts,arg.args[0],t);
		for(size_t i=1; i<n; i++) {
			arg.maps[i](out,out,npts,arg.args[i],t);
		}
	} else {
		arg.maps[n-1](z,out,npts,arg.args[n-1],t);
		for(size_t i=1; i<n; i++) {
			arg.maps[n-1-i](out,out,npts,arg.args[n-1-i],t);
		}
	}
}


/* ------------------------------------------ */

void map_identity(const double* z, double* out, int npts, const void* args, double t) {
	for(int i=0; i<npts; i++) {
		out[0] = z[0];
		out[1] = z[1];
		z+=2; out+=2;
	}
}

/* ------------------------------------------ */

void map_translate(const double* z, double* out, int npts, const void* args, double t) {
	map_translate_args arg = *(map_translate_args*) args;
	for(int i=0; i<npts*arg.dim; i++) {
		out[i] = z[i] + t * arg.delta[i%arg.dim];
	}
}

/* ------------------------------------------ */

void map_rotate(const double* z, double* out, int npts, const void* args, double t) {
	double angle = t * (*(double*) args);
	for(int i=0; i<npts; i++) {
		double q = z[0];
		double p = z[1];
		out[0] = cos(angle) * q - sin(angle) * p;
		out[1] = cos(angle) * p + sin(angle) * q;

		z+=2; out+=2;
	}
}

/* ------------------------------------------ */

void map_hyperbolic(const double* z, double* out, int npts, const void* args, double t) {
	double alpha = (*(double*) args);
	double alpha_pt = pow(alpha,+t);
	double alpha_mt = pow(alpha,-t);
	for(int i=0; i<npts; i++) {
		out[0] = z[0]*alpha_pt;
		out[1] = z[1]*alpha_mt;

		z+=2; out+=2;
	}
}

/* ------------------------------------------ */

void map_kickl(const double* z, double* out, int npts, const void* args, double t) {
	double kick = *(double*) args;
	for(int i=0; i<npts; i++) {
		out[0] = z[0];
		out[1] = z[1] + t * kick * z[0];
		z+=2; out+=2;
	}
}

/* ------------------------------------------ */

void map_driftl(const double* z, double* out, int npts, const void* args, double t) {
	double drift = *(double*) args;
	for(int i=0; i<npts; i++) {
		out[0] = z[0] + t * drift * z[1];
		out[1] = z[1];
		z+=2; out+=2;
	}
}

/* ------------------------------------------ */

static double poly(double x, int n, double* factors) {
	double r = factors[n-1];
	for(int i=n-2; i>=0; i--){
		//r = r*x-factors[i];
		r = r*x+factors[i]; // 2021-09-09 corrected from above
	}
	return r;
}

void map_kickpoly(const double* z, double* out, int npts, const void* _args, double t) {
	struct map_kickpoly_args* args = (map_kickpoly_args*) _args;
	for(int i=0; i<npts; i++) {
		out[0] = z[0];
		out[1] = z[1] + t * poly(z[0],args->n,args->factors);
		z+=2; out+=2;
	}
}

/* ------------------------------------------ */

void map_driftpoly(const double* z, double* out, int npts, const void* _args, double t) {
	struct map_driftpoly_args* args = (map_driftpoly_args*) _args;
	for(int i=0; i<npts; i++) {
		out[0] = z[0] + t * poly(z[1],args->n,args->factors);
		out[1] = z[1];
		z+=2; out+=2;
	}
}

/* ------------------------------------------ */

void map_driftsine(const double* z, double* out, int npts, const void* _args, double t) {
	double* args = (double*) _args;
	for(int i=0; i<npts; i++) {
		out[0] = z[0] + t * args[0]*sin(args[1]*z[1] + args[2]);
		out[1] = z[1];
		z+=2; out+=2;
	}
}

/* ------------------------------------------ */

void map_kicksine(const double* z, double* out, int npts, const void* _args, double t) {
	double* args = (double*) _args;
	for(int i=0; i<npts; i++) {
		out[0] = z[0];
		out[1] = z[1] + t * args[0]*sin(args[1]*z[0] + args[2]);
		z+=2; out+=2;
	}
}

/* ------------------------------------------ */
void map_cavity(const double* z, double* out, int npts, const void* args_, double t) {
	const struct map_cavity_args* args = (const struct map_cavity_args*) args_;
	double k = 2.0*M_PI * args->freq / 299792458.0;
	double kick;
	double kick0 = args->ampl *  cos( args->phase );
	switch(args->mode) {
	case(0):
		for(int i=0; i<npts; i++) {
			kick = args->ampl *  cos( k * z[0] + args->phase );
			out[0] = z[0];
			out[1] = z[1] + t * (kick-kick0);
			z+=2; out+=2;
		}
		break;
	case(1):
		for(int i=0; i<npts; i++) {
			kick = args->ampl * k * -sin(args->phase) * z[0];
			out[0] = z[0];
			out[1] = z[1] + t * kick;
			z+=2; out+=2;
		}
		break;
	}
}

/* ------------------------------------------ */

static inline double map_chicanec_F(int n, double a, double b, double lmb) {
	if(n==0) return pow(lmb*lmb-1,-a) * pow(lmb,b);
	return -2*a*map_chicanec_F(n-1,a+1,b+1,lmb) + b*map_chicanec_F(n-1,a,b-1,lmb);
}

static inline double map_chicanec_s(double p, double alpha, double lD, double lB, double P0) {
	double R0 = lB / sin(alpha);
	double R = R0/P0 * p;
	double sB = R * asin(lB/R);
	double sD = lD / cos(asin(lB/R));
	return 4*sB + 2*sD;
}

double chicanec_R56(double alpha, double lB, double lD) {
	double lmb  = 1.0/sin(alpha);
	double R56B = lB * lmb * ( asin(1/lmb) - 1./sqrt(lmb*lmb -1) );
	double R56D = lD * lmb * map_chicanec_F(1,0.5,1,lmb);
	double R56  = 4*R56B + 2*R56D;
	return -R56;
}

void map_chicanec(const double* z, double* out, int npts, const void* args_, double t) {
	const struct map_chicanec_args* args = (const struct map_chicanec_args*) args_;

	if(abs(args->alpha)<1e-6) {
		for(int i=0; i<npts; i++) {
			out[0] = z[0];
			out[1] = z[1];
			z+=2; out+=2;
		}
		return;
	}

	double P0  = args->energy;
	double R56 = chicanec_R56(args->alpha,args->lB,args->lD);


	switch(args->mode) {
	case(0): {
		double s0 = map_chicanec_s(P0,args->alpha,args->lD,args->lB,P0);
		for(int i=0; i<npts; i++) {
			double s = map_chicanec_s(P0+z[1],args->alpha,args->lD,args->lB,P0); 
			out[0] = z[0] + t * -(s-s0);
			out[1] = z[1];
			z+=2; out+=2;
		}
		break;
	}
	case(1):
		for(int i=0; i<npts; i++) {
			double d = R56/P0 * z[1];
			out[0] = z[0] + t * d;
			out[1] = z[1];
			z+=2; out+=2;
		}
		break;
	}
}

/* ------------------------------------------ */


static inline double ipol_lin(const double* data, const double* limits, size_t npts, double x) {
	double lo = limits[0];
	double hi = limits[1];
	x=(x-lo)/(hi-lo);
	if( x<0|| x>=1 ) return  0.0;

	size_t i = x*npts;
	x = x*npts-i;
	double f0=data[i];
	double f1=i<npts-1?data[i+1]:0.0;
	return f0*(1-x)+f1*x;
}

static inline double ipol_cub(const double* data, const double* limits, size_t npts, double x) {
	const double lo = limits[0];
	const double hi = limits[1];
	x=(x-lo)/(hi-lo);
	//if( x<0|| x>=1 ) return  0.0;
	x -= floor(x);

	const size_t i = x*npts;
	x = x*npts-i;

	size_t ii;
	ii=(i-1+npts)%npts; double f0=data[ii];
	ii=(i+0+npts)%npts; double f1=data[ii];
	ii=(i+1+npts)%npts; double f2=data[ii];
	ii=(i+2+npts)%npts; double f3=data[ii];

	return 
	f0 * (1+x-1) * (1+x-2) * (1+x-3)
	   / (-1 * -2 * -3)
	+
	f1 * (1+x-0) * (1+x-2) * (1+x-3)
	   / (1 * -1 * -2)
	+
	f2 * (1+x-0) * (1+x-1) * (1+x-3)
	   / (2 * 1 * -1)
	+
	f3 * (1+x-0) * (1+x-1) * (1+x-2)
	   / (3 * 2 * 1);
}

void map_datakick(const double* zin, double* out, int npts, const void* args_, double t) {
	const struct map_datakick_args* args = (const struct map_datakick_args*) args_;
	const double* z = zin;

	double q,kick;
	double kick0=ipol_cub(args->data,args->limits,args->npts,0);
	for(int i=0; i<npts; i++) {
		q = z[args->dim_q];
		//kick = ipol_lin(args->data,args->limits,args->npts,q);
		kick = ipol_cub(args->data,args->limits,args->npts,q);
		kick -= kick0;
		out[args->dim_q] = q;
		out[args->dim_p] = z[args->dim_p] + args->factor * t * kick;
		z+=args->dim; out+=args->dim;
	}
}

/* ------------------------------------------ */

void free_map_args(map_t map, void* args_) {
		if(map == SLV::map_datakick) {
			map_datakick_args* args = (map_datakick_args*) args_;
			free(args->data);
			free(args->limits);
		}
		if(map == SLV::map_translate) {
			map_translate_args* args = (map_translate_args*) args_;
			free(args->delta);
		}
}


}
