/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef SLV_MAPS_H
#define SLV_MAPS_H

#include <cstddef>
#include <cstdlib>

typedef void (*map_t)(const double*, double*, int, const void*, double);

namespace SLV {

template<int DIM>
void map_identity(const double* zin, double* out, int npts, const void* args, double t) {
	for(int i=0; i<DIM*npts; i++) {
		out[i] = zin[i];
	}
}

struct map_compose_args {
	map_t maps[64];
	void* args[64];
	size_t n;
};
void map_compose(const double* z, double* out, int npts, const void* args, double t);

void map_identity(const double* z, double* out, int npts, const void* args, double t);

struct map_translate_args {
	double* delta;
	int dim;
};
void map_translate(const double* z, double* out, int npts, const void* args, double t);

void map_rotate(const double* z, double* out, int npts, const void* args, double t);
void map_hyperbolic(const double* z, double* out, int npts, const void* args, double t);

void map_kickl(const double* zin, double* out, int npts, const void* args, double t);
void map_driftl(const double* zin, double* out, int npts, const void* args, double t);

struct map_kickpoly_args {
	int n;
	double factors[64];
};
void map_kickpoly(const double* zin, double* out, int npts, const void* args, double t);
struct map_driftpoly_args {
	int n;
	double factors[64];
};
void map_driftpoly(const double* zin, double* out, int npts, const void* args, double t);

void map_kicksine(const double* zin, double* out, int npts, const void* args, double t);
void map_driftsine(const double* zin, double* out, int npts, const void* args, double t);

void map_cavity(const double* z, double* out, int npts, const void* args, double t);
struct map_cavity_args {
	double freq, phase, ampl;
	int mode;
};

double chicanec_R56(double alpha, double lB, double lD);
void map_chicanec(const double* z, double* out, int npts, const void* args, double t);
struct map_chicanec_args {
	double alpha, lB, lD, energy;
	int mode;
};

void map_datakick(const double* zin, double* out, int npts, const void* args, double t);
struct map_datakick_args {
	double* data;
	double* limits;
	size_t dim;
	size_t dim_q;
	size_t dim_p;
	size_t npts;
	double factor;
};

void free_map_args(map_t map, void* args);


}

#endif
