/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <pthread.h>
#include <error.h>

#include "math_eval.h"

static void remove_newline(char* s) {
	for(size_t i=0; i<strlen(s); i++) {
		if(s[i] == '\n') s[i]=' ';
	}
}

void* p_evaluator_create(char* fstr) {
	static pthread_mutex_t mutex;
	pthread_mutex_lock(&mutex);
	remove_newline(fstr);
	void* ev = evaluator_create(fstr);
	if(NULL==ev) error(1,0,"Invalid math expression: %s",fstr);
	pthread_mutex_unlock(&mutex);
	return ev;
}

void map_analytic(const double* z, double* out, int npts, const void* _args, double t) {
	map_analytic_args* args = (map_analytic_args*) _args;
	void* evs[args->dim];
	static pthread_mutex_t mutex;
	pthread_mutex_lock(&mutex);
	for(int d=0; d<args->dim; d++) {
		evs[d] = p_evaluator_create(args->fstrs[d]);
	}
	pthread_mutex_unlock(&mutex);
	for(int i=0; i<npts; i++,out+=args->dim,z+=args->dim) {
		double values[1+args->dim];
		memcpy(values+1,z,args->dim*sizeof(double));
		values[0] = t;
		for(int d=0; d<args->dim; d++) {
			out[d] = evaluator_evaluate(evs[d], 1+args->dim, args->names, values);
		}
	}
	for(int d=0; d<args->dim; d++) {
		evaluator_destroy(evs[d]);
	}
}
