/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef SLV_NOISE_H
#define SLV_NOISE_H

#include <cell.h>
#include <random>

namespace SLV {

template<int DIM>
void noise_uniform(SLV::cell<DIM>* cell, double a) {
	if( cell->data ) {
		const int npts = cell->npoints();
		for(int i=0; i<npts; i++) {
			double r = drand48();
			cell->data[i] *= 1+a*2*(r-0.5);
		}
	}
	for(int i=0; i<1<<DIM; i++) {
		if( NULL == cell->cells[i] )  continue;
		noise_uniform(cell->cells[i],a);
	}
}

template<int DIM>
void noise_poisson(SLV::cell<DIM>* cell, double N, int seed=0) {
	static std::mt19937 gen(seed);
	noise_poisson_worker(cell, N, &gen);
}

template<int DIM>
void noise_poisson_worker(SLV::cell<DIM>* cell, double N, std::mt19937* gen) {
	// N: total number of particles
	if( cell->data ) {

		// calculate phase-space volume covered by a single point
		double V = 1;
		for(int d=0; d<DIM; d++) V *= 2.*cell->cwidth[d];
		V /= 1<<(DIM*cell->nexp);

		const int npts = cell->npoints();
		for(int i=0; i<npts; i++) {
			if(cell->data[i]<=0) continue;
			double L = cell->data[i] * V * N;
			std::poisson_distribution<int> dist(L);
			double r = (double) dist(*gen) / L;
			//cell->data[i] *= r!=0.0 ? r : 1;
			cell->data[i] *= r;
		}
	}
	for(int i=0; i<1<<DIM; i++) {
		if( NULL == cell->cells[i] )  continue;
		noise_poisson_worker(cell->cells[i], N, gen);
	}
}

}
#endif
