/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <fftw3.h>
#include <cstdio>
#include <stdlib.h>

#include "poisson.h"
#include "bessel.h"



//void green_f_adisk(const double* in, _Complex* out, int npts, const void* args) {
//	double a = *(double*) args;
//	double k,signk;
//	for(int i=0; i<npts; i++,in++,out++) {
//		k = *in;
//		signk = (k>= 0 ? 1.0 : -1.0);
//		*out =  -1*I * 2./a/a * (k!=0 ? 1./k - signk * a * bessk1( a * fabs(k)) : 0.0 );
//	}
//}

void green_f_adisk(const double* in, double _Complex* out, int N, const void* args) {
	double a = *(double*) args; // beam radius
	double k,xi,xiabs;
	double B;
	for(int i=0; i<N; i++,in++,out++) {
		k = *in;
		xi = k*a;
		xiabs = fabs(xi);
		//if( xiabs<1e-3) {
		//	B = -xi/2.0*(log(xiabs)-0.615931515658412448810720031375774);
		//} else {
			B = (1.0/xi - xiabs * bessk1(xiabs) / xi);
		//}
		//B += a*k*log(a)/2.0 * exp(-xiabs);
		//B = 1.0/xi; // FIXME
		//B=1/(xi*xi*xi);
		//B=1/xi;
		double _Complex res = (-1.0*I * 2.0 / a) * B;
		*out = (xiabs==0 ? 0.0 : res);
	}
}

//void green_f_csr(const double* in, double _Complex* out, int N, const void* args) {
//	// see Heifets-2002-CSRInstaBC-PhysRevSTAB.5.064401.pdf
//	// A = 3**(-1./3.)*gamma(2./3.)*(sqrt(3)*i - 1)
//	// differs by 1/2 from  Murphy-1995-LongWakeSynRad-TPQ14.PDF
//	double Areal = -0.9388929401017445663411803800;
//	double Aimag =  1.6262102751239442129162137543;
//	double _Complex A;
//	double radius = *(double*) args;
//	double k,sign;
//	for(int i=0; i<N; i++,in++,out++) {
//		k = *in;
//		sign = k<0?-1:1;
//		k = k * sign;
//		A = Areal - sign * I * Aimag;
//		double _Complex res = - A * pow(k/radius/radius,1./3.) / 2;
//		*out = res;
//	}
//}

void green_f_csr(const double* in, double _Complex* out, int N, const void* args) {
	// see Heifets-2002-CSRInstaBC-PhysRevSTAB.5.064401.pdf
	// A = 3**(-1./3.)*gamma(2./3.)*(sqrt(3)*i - 1)
	// differs by 1/2 from  Murphy-1995-LongWakeSynRad-TPQ14.PDF
	double _Complex A = 0.938892940101744566341180380 - 1.6262102751239442129162137543I;
	//const double G1o3 = 2.6789385347077; // Gamma[(1/3)
	//double _Complex A = sqrt(M_PI/2)/3/G1o3*(sqrt(3)+3*I);
	double radius = *(double*) args;
	double k;
	for(int i=0; i<N; i++,in++,out++) {
		k = *in;
		//*out =  -1i * (k>0?A:conj(A)) * cpow(k/radius/radius,1./3.);
		*out =  -(k>0?I*A:conj(I*A)) * cpow(fabs(k)/radius/radius,1./3.);
	}
	fprintf(stderr,">>>CSR: %e %e\n",creal(A),cimag(A));
	fprintf(stderr,">>>CSR: %e %e\n",creal(out[0]),cimag(out[0]));
}

void solve_poisson(const double* data, double* out, size_t npts, const void* args_, char* debug) {
	solve_poisson_args* args = (solve_poisson_args*) args_;

	const double T = args->T;
	const size_t npad = args->npad>0?args->npad:1;
	const size_t N = npad*npts;

	double x[N] = {0};
	double _Complex rho_s[N] = {0};
	double _Complex rho_f[N] = {0};
	double _Complex sol_s[N] = {0};
	double _Complex sol_f[N] = {0};
	double _Complex grn_s[N] = {0};
	double _Complex grn_f[N] = {0};

	for(size_t i=0; i < npts; i++) {
		rho_s[i] = data[i];
	}

	fftw_plan p;

	if( NULL != args->green_s ) {
		// Calculate spatial values
		for(size_t n=0; n<N; n++) {
			x[n] = -(T*npad)/2.0 + (T*npad)/N * n ;
		}
		args->green_s(x,x,N,args->fargs); // Call Greens function in spacial domain
		for(size_t n=0; n<N; n++) {
			grn_s[n] = x[n] / npts;
			//grn_s[n] = x[n];
		}

		p = fftw_plan_dft_1d(N, (fftw_complex*) grn_s, (fftw_complex*) grn_f, FFTW_FORWARD, FFTW_ESTIMATE);
		fftw_execute(p);
		fftw_destroy_plan(p);
	}


	if( NULL != args->green_f ) {
		// Calculate k values
		for(size_t n=0; n<N; n++) {
			x[n] = ( (n < N/2) ? 1.0*n : -1.0*(1.0*N-1.0*n) ) / (T*npad) * 2.0 * M_PI;
		}
		args->green_f(x,grn_f,N,args->fargs); // Call Greens function in freq domain
	}

	p = fftw_plan_dft_1d(N, (fftw_complex*) rho_s, (fftw_complex*) rho_f, FFTW_FORWARD, FFTW_ESTIMATE);
	fftw_execute(p);
	fftw_destroy_plan(p);

	// Calculate convolution in frequency space
	for(size_t n=0; n<N; n++) {
		sol_f[n] = rho_f[n] * grn_f[n];
	}

	//
	//sol_f[N/2] = 0.0; // FIXME DOUBLE CHECK was uncommented originally
	//for(size_t n=0; n<N/4; n++) {
	//	sol_f[N/2+n] = 0;
	//	sol_f[N/2-n] = 0;
	//}

	p = fftw_plan_dft_1d(N, (fftw_complex*) sol_f, (fftw_complex*) sol_s, FFTW_BACKWARD, FFTW_ESTIMATE);
	fftw_execute(p);
	fftw_destroy_plan(p);

	for(size_t n=0; n<npts; n++) {
		out[n] = creal(sol_s[n])/N;
	}

	
	#if 1
	// write debug file
	if(debug) {
		FILE* fd;
		fd = fopen(debug,"w");
		fprintf(fd,"q k nh ");
		fprintf(fd,"rho_s_real rho_s_imag rho_f_real rho_f_imag ");
		fprintf(fd,"grn_s_real grn_s_imag grn_f_real grn_f_imag ");
		fprintf(fd,"sol_s_real sol_s_imag sol_f_real sol_f_imag ");
		fprintf(fd,"\n");
		for(size_t n=0; n<N; n++) {
			double h = T/npts;
			double q = n*T/npts-T/2.;
			double k = 2.0 * M_PI * ( (n <= N/2) ? 1.0*n : -1.0*(1.0*N-1.0*n) ) / N / h;
			fprintf(fd,"%+e %+e %+e ", q, k, n*h);
			fprintf(fd,"%+e %+e %+e %+e ",  creal(rho_s[n]), cimag(rho_s[n]), creal(rho_f[n]), cimag(rho_f[n]));
			fprintf(fd,"%+e %+e %+e %+e ",  creal(grn_s[n]), cimag(grn_s[n]), creal(grn_f[n]), cimag(grn_f[n]));
			fprintf(fd,"%+e %+e %+e %+e ",  creal(sol_s[n]), cimag(sol_s[n]), creal(sol_f[n]), cimag(sol_f[n]));
			fprintf(fd,"\n");
		}
		fclose(fd);
	}
	#else
	if(prefix) {
		FILE* fd;
		char buf[1024];

		sprintf(buf,"%s-poisson-rho.dat",prefix);
		fd = fopen(buf,"w");
		for(size_t n=0; n<N; n++) {
			fprintf(fd,"%+e ",  creal(rho_s[n]));
			fprintf(fd,"%+e ",  cimag(rho_s[n]));
			fprintf(fd,"%+e ",  creal(rho_f[n]));
			fprintf(fd,"%+e ",  cimag(rho_f[n]));
			fprintf(fd,"\n");
		}
		fclose(fd);

		sprintf(buf,"%s-poisson-grn.dat",prefix);
		fd = fopen(buf,"w");
		for(size_t n=0; n<N; n++) {
			fprintf(fd,"%+e ",  creal(grn_s[n]));
			fprintf(fd,"%+e ",  cimag(grn_s[n]));
			fprintf(fd,"%+e ",  creal(grn_f[n]));
			fprintf(fd,"%+e ",  cimag(grn_f[n]));
			fprintf(fd,"\n");
		}
		fclose(fd);

		sprintf(buf,"%s-poisson-sol.dat",prefix);
		fd = fopen(buf,"w");
		for(size_t n=0; n<N; n++) {
			fprintf(fd,"%+e ",  creal(sol_s[n]));
			fprintf(fd,"%+e ",  cimag(sol_s[n]));
			fprintf(fd,"%+e ",  creal(sol_f[n]));
			fprintf(fd,"%+e ",  cimag(sol_f[n]));
			fprintf(fd,"\n");
		}
		fclose(fd);
	}
	#endif

	fftw_cleanup();
}
