/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <complex.h>
#include <fftw3.h>
#include <cstdio>
#include <stdlib.h>

#include "poisson1d.h"



void solve_poisson1d(const double* rho, double* out, size_t npts, const void* args_, char* prefix) {
	solve_poisson1d_args* args = (solve_poisson1d_args*) args_;

	const double T = args->T;
	const size_t npad = args->npad>0?args->npad:1;
	const size_t N = npad*npts;
	double h = T/N;

	double _Complex in[N]; // _Complex array for FFT-business

	for(size_t i=0; i<N; i++) {
		in[i] = 1-rho[i];
	}

	/* Exectute forward FFT */
	fftw_plan p;
	p = fftw_plan_dft_1d(N, (fftw_complex*) in, (fftw_complex*) in, FFTW_FORWARD, FFTW_ESTIMATE);
	fftw_execute(p);

	const double _Complex W = cexp(2.0 * M_PI * I / (double) N);
	double _Complex Wm = 1;

	for(size_t m=0; m<N; m++) {
		//double _Complex denom = Wm + 1.0/Wm - 2.;
		double _Complex fac = -(-1.0/Wm + Wm) / (Wm + 1.0/Wm - 2.) / 2;
		if(m!=0) {
			in[m] *= fac;
		}
		Wm *= W;
	}
	in[0] = 0.0;

	/* Execute backward FFT */
	p = fftw_plan_dft_1d(N, (fftw_complex*) in, (fftw_complex*) in, FFTW_BACKWARD, FFTW_ESTIMATE);
	fftw_execute(p);
	fftw_destroy_plan(p);

	for(size_t i=0; i<N; i++) {
		out[i] = creal(in[i]);
		out[i] *= h * h;
	}

	
	#if 0
	if(prefix) {
		FILE* fd;
		char buf[1024];

		sprintf(buf,"%s-poisson1d-rho.dat",prefix);
		fd = fopen(buf,"w");
		for(size_t n=0; n<N; n++) {
			fprintf(fd,"%+e ",  creal(rho_s[n]));
			fprintf(fd,"%+e ",  cimag(rho_s[n]));
			fprintf(fd,"%+e ",  creal(rho_f[n]));
			fprintf(fd,"%+e ",  cimag(rho_f[n]));
			fprintf(fd,"\n");
		}
		fclose(fd);

		sprintf(buf,"%s-poisson1d-sol.dat",prefix);
		fd = fopen(buf,"w");
		for(size_t n=0; n<N; n++) {
			fprintf(fd,"%+e ",  creal(sol_s[n]));
			fprintf(fd,"%+e ",  cimag(sol_s[n]));
			fprintf(fd,"%+e ",  creal(sol_f[n]));
			fprintf(fd,"%+e ",  cimag(sol_f[n]));
			fprintf(fd,"\n");
		}
		fclose(fd);
	}
	#endif

	fftw_cleanup();
}
