/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef SLV_PSD_H
#define SLV_PSD_H

#include <cell.h>
#include <debug.h>

#include <signal.h>
#include <string.h>
#include <assert.h>

#include <cmath>


namespace SLV {

typedef void (*fun_t)(const double*, double*, int, const void*);
typedef void (*map_t)(const double*, double*, int, const void*, double);
typedef void (*ffun_t)(const double*, const double*, double*, int, const void*);



/// The psd<DIM> class represents a memory efficient approximation
/// of a DIM-dimensional function on a hypercuboid region.
/// This efficiency is achieved by adaptively refining the region of interest into 2^\p DIM sub-hypercubes.
/// For every leaf-region of the resulting 2^DIM-tree a \p Interpolator object is calculated and stored
/// in memory. To evaluate the PSD at a coordinate \p pt, the \p Interpolator object responsible for the
/// region \p pt lays in is called.

template <int DIM>
class psd {

public:

	typedef typename SLV::cell<DIM>::value_type value_type;
	typedef typename SLV::cell<DIM>             cell_type;
	typedef typename SLV::cell<DIM>::cell_ptr   cell_ptr;
	typedef typename SLV::cell<DIM>::coord_type coord_type;
	typedef typename SLV::psd<DIM-1>            proj_type;
	typedef typename proj_type::cell_ptr       proj_cell_ptr;
	typedef typename proj_type::coord_type     proj_coord_type;


	cell_ptr   root  = NULL;   ///< Pointer to the ultimate parent of the underlying tree-structure
	int        depth = 8;      ///< Refinement depth
	int        nexp  = 4;
	double     weight = 1.0;   ///< Charge
	double     epsilon = 0; ///< Cutoff
	coord_type refpoint = {0.0}; ///< Position of {0,0} in absolute phase space

	int topology = 0;
	int interpolation = 2;

	int calls = 0;


	/// Constructor
	///
	/// \param cpoint Center point of the region of interest
	/// \param cwidth Extension of the ROI in each dimension from the center point
	psd(coord_type cpoint, coord_type cwidth) {
		DBG_CNSTR_MSG_1;
		root = cell_ptr( new cell<DIM>(cpoint,cwidth) );
	}

	psd* copy() {
		psd* res = new psd(*this);
		res->root = new cell<DIM>(*root);
		return res;
	}


	/// Returns a pointer to the projection of the PSD along \p axis.
	proj_type* projection(int axis) {
		proj_type* ret_projection = new proj_type(coord_sub(root->cpoint,axis), coord_sub(root->cwidth,axis));
		projection_worker(root, axis, ret_projection);
		ret_projection->depth = depth;
		return ret_projection;
	}

	void set_topology(int t) {
		topology = t;
		for(int i=0; i<DIM; i++) {
			if( topology & (1<<i) ) {
				root->nbors[2*i+0] = root;
				root->nbors[2*i+1] = root;
			}
		}
	}

	void projection_worker(cell_ptr cell, int axis, proj_type* projection) {
		// FIXME
		#if 0
		if( cell->payload ) { // we assume if a cell has payload it is not further refined

			proj_cell_ptr bin = projection->root->refine_at(coord_sub(cell->cpoint,axis),cell->level()); // Find bin
			proj_interp* new_interp = cell->payload->projection(axis); // Delete me!

			if(bin->payload) { // There is already something in the bin
				bin->payload->add(new_interp);
				delete new_interp;
			}
			else { // There is nothing in the bin
				bin->payload = new_interp;
			}

			return;
		}
		#endif

		for(int i=0; i<(1<<DIM); i++) { // Recursion
			if( cell->cells[i] ) {
				projection_worker(cell->cells[i], axis, projection);
			}
		}
	}




	/// For a 1-DIM PSD stores the data points of the PSD on a uniform grid to \p dd.
	/// \p dd has to be of size data_size()
	void data_dump(double* dd) {
		if(DIM!=1) {
			printf("psd::data_dump() is currently not implemented for DIM!=1.");
			raise(SIGSEGV);
		}

		// data_dump_worker returns pointer to last array element, which we manually set to zero
		dd = data_dump_worker(root,dd);
		*dd = 0.0;

		return;
	}

	value_type* data_dump_worker(cell_ptr cell, value_type* dd, int level=0) {
		int ndata;
		if(cell==NULL) {
			ndata = (1<<(depth - level)) * ((1<<nexp) - 1);
			memset(dd, 0 , sizeof(value_type) * ndata);
			return dd + ndata;
		}
		if(cell->filled()) {
			ndata = (1<<cell->nexp) - 1;
			memcpy(dd, cell->data, sizeof(value_type) * ndata); // Copy interp data
			return dd + ndata; // Return advanced pointer
		}
		if(level==depth) {
			ndata = (1<<(depth - level)) * ((1<<cell->nexp) - 1);
			memset(dd, 0 , sizeof(value_type) * ndata);
			return dd + ndata;
		}
		for(int i=0; i<1<<DIM; i++) { // Loop over childs
			dd = data_dump_worker(cell->cells[i],dd,level+1);
		}
		return dd;
	}


	/// For a 1-DIM PSD, returns the number of data points on a uniform grid.
	/// Double counting is accounted for. 
	int data_size() {
		if(DIM!=1) {
			printf("psd::data_size() is currently not implemented for DIM!=1.");
			raise(SIGSEGV);
		}

		return (1<<depth)*((1<<root->size) - 1) + 1;
	}




	/// Calculate the integral of the PSD over the complete region of interest
	value_type integral() const {
		value_type sum = 0;
		integral_worker(root,&sum);
		return sum;
	}

	void integral_worker(cell_ptr cell, value_type* sum) const {
		if(cell->data) {
			*sum += cell->integral();
			return;
		}
		for(int i=0; i<1<<DIM; i++) {
			if( cell->cells[i] ) {
				integral_worker(cell->cells[i],sum);
			}
		}
	}


	/// Calculate the integral of the PSD over the complete region of interest
	double integral(ffun_t f, void* args) const {
		double sum = 0;
		integral_worker(root,&sum,f,args);
		return sum;
	}

	void integral_worker(cell_ptr cell, double* sum, ffun_t f, void* args) const {
		if(cell->data) {
			*sum += cell->integral(f,args);
			return;
		}
		for(int i=0; i<1<<DIM; i++) {
			if( cell->cells[i] ) {
				integral_worker(cell->cells[i],sum,f,args);
			}
		}
	}


	/// Mutliply the PSD by \p factor
	void multiply(double factor) {
		root->multiply(factor);
	}

	/// Modify values
	void modify(ffun_t f, void* farg) {
		root->modify(f,farg);
	}

	/// Normalize so that integral = 1
	void normalize() {
		root->multiply(1./integral());
	}


	#if 0
	/// Multiply the PSD by a function func
	/// Psi(z) -> Psi(z) * f(z)
	template<typename FUNC>
	void multiply(FUNC func) {
		multiply_worker(root,func);
	}

	template<typename FUNC>
	void multiply_worker(cell_ptr cell, FUNC func) {
		if( cell->payload ) {
			int i = 0;
			for(auto pt : cell->grid_sample(1<<cell->size)) {
				cell->payload->data[i] *= func(pt);
				i++;
			}
			return;
		}
		for(int i=0; i<1<<DIM; i++) {
			if( cell->cells[i] ) {
				multiply_worker(cell->cells[i],func);
			}
		}
	}

	/// Normalize PSD so that its integral is 1
	void normalize() {
		multiply(1./integral());
	}
	#endif

	
	#if 0
	/// Returns the expectation value of \p func.
	/// func: R^DIM -> V, where V is a Vector Space
	/// Scalar multiplication and vector addition
	/// must be implemented for the return type of func.
	template<typename FUNC>
	auto expectation(FUNC func) {
		auto sum = 0 * func(root->cpoint); // Generate neutral-element
		expectation_worker(root,func,&sum);
		return sum;
	}

	template<typename FUNC>
	void expectation_worker(cell_ptr cell, FUNC func, auto* sum) {
		if( cell->payload ) {
			// func is evaluated at the center point of the leaf
			// and weighted with the integral.
			#if 1
			*sum = *sum + func(cell->cpoint) * cell->payload->integral();
			#else
			printf("Called New Expectation!\n");
			*sum = *sum + cell->payload->expectation(func);
			#endif
		}
		for(int i=0; i<1<<DIM; i++) {
			if( cell->cells[i] ) { // Recursion
				expectation_worker(cell->cells[i],func,sum);
			}
		}
	}
	#endif




	/// Remove the axis'th element of an coordinate vector
	proj_coord_type coord_sub(const coord_type & pt, const unsigned int & axis) {
			int ret_ax = 0;
			unsigned int ax = 0;
			proj_coord_type ret;
			while(ax < DIM) {
				if(ax != axis) { // We want to keep this element
					ret[ret_ax] = pt[ax];
					ret_ax++;
					ax++;
				}
				else { // Dont want it
					ax++;
				}

			}
			return ret;
	}


	/// Evaluate function at \p pt
	///
	/// \param pt point at which function is evaluated
	/// \return PSD(pt)
	double operator() (coord_type pt) {
		return 0;
	}


	double max() {
		double ans = 0;
		std::vector<cell_ptr> leafs = root->leafs();
		for(auto leaf : leafs) {
			if(leaf->filled()) {
				for(int n=0; n<(1<<leaf->nexp * DIM); n++) {
					ans = (leaf->data[n] > ans ? leaf->data[n] : ans );
				}
			}
		}
		return ans;
	}


	#if 0
	/// Translate the PSD
	void translate(const double* delta) {
		translate_worker(root,delta);
	}
	void translate_worker(cell_ptr cell, const double* delta) {
		// Translate the box
		for(int i=0; i<DIM; i++) {
			cell->cpoint[i] += delta[i];
		}
		// Translate interpolator if it exists
		if(cell->filled()) {
			for(int i=0; i<DIM; i++) {
				cell->payload->low_edge[i] += delta[i];
			}
		}
		// recursively translate childs
		for(int i=0; i<(1<<DIM); i++) {
			if( cell->cells[i]) {
				translate_worker(cell->cells[i],delta);
			}
		}
	}
	#endif

	void print(FILE* fd = stdout) const {
		fprintf(fd,"Center: ");
		for(int d=0; d<DIM; d++) fprintf(fd," %e",root->cpoint[d]);
		fprintf(fd,"\n");
		fprintf(fd,"Width:  ");
		for(int d=0; d<DIM; d++) fprintf(fd," %e",root->cwidth[d]);
		fprintf(fd,"\n");
		fprintf(fd,"Limits: ");
		for(int d=0; d<DIM; d++) fprintf(fd," %e", root->cpoint[d] - root->cwidth[d]);
		for(int d=0; d<DIM; d++) fprintf(fd," %e", root->cpoint[d] + root->cwidth[d]);
		fprintf(fd,"\n");
		fprintf(fd,"Ref.Pt: ");
		for(int d=0; d<DIM; d++) fprintf(fd," %e",refpoint[d]);
		fprintf(fd,"\n");
		fprintf(fd,"Depth:   %d\n", depth);
		fprintf(fd,"nexp:    %d\n", nexp);
		fprintf(fd,"Weight:  %e\n", weight);
		fprintf(fd,"Ipol:    %d\n", interpolation);
		fprintf(fd,"Topo::   %d\n", topology);
		fprintf(fd,"Leafs:   %lu\n", root->count_leafs());
		fprintf(fd,"Integrl: %e\n", integral());
	}





	~psd() {
		DBG_DESTR_MSG_3;
		delete   root;
	}





}; // End of class psd


template<> // Define DIM=0 to void infinite template deduction
class psd<0> {
public:
	typedef void* cell_ptr;
	typedef void  coord_type;
};


} // End of namespace SLV

#endif
