/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef SLV_PSDENSEMBLE_H
#define SLV_PSDENSEMBLE_H

#include <cstdio>
#include <cstdlib>

#include "psd_eval.h"


namespace SLV {

template<int DIM>
cell<DIM>* weighted_descent(cell<DIM>* cell) {
	double a = cell->weight * rand()/(1.0*RAND_MAX);
	for(int i=0; i<1<<DIM; i++) {
		if(NULL == cell->cells[i]) continue;
		a -= cell->cells[i]->weight; 
		if(a<=0) {
			return weighted_descent(cell->cells[i]);
		}
	}
	return cell;
}

// Produces a point sampled from the density inside the cell
template<int DIM>
void psd_get_random_point(psd<DIM>* psd, double* pt) {
	cell<DIM>* cell = weighted_descent(psd->root);
	const double max = cell->max();
	const double min = cell->min();
	for(int i=0;i <16; i++) {
		for(int d=0; d<DIM; d++) { // generate random point in cell
			pt[d] = cell->cpoint[d] + cell->cwidth[d] * (1 - 2.0*rand()/(1.0*RAND_MAX));
		} 
		double u;
		psd_eval<DIM>(pt, &u, 1, psd);
		u = (u-min)/(max-min);
		double v = rand()/(1.0*RAND_MAX);
		if(u>=v) {
			//fprintf(stderr,"%e %e %e %d\n",u,min,max);
			break;
		}
	}
}

template<int DIM>
void psd_ensemble(psd<DIM>* psd, FILE* fd, int n, unsigned int seed, double* reference=NULL) {
	srand(seed);
	psd->root->set_weight_integral();
	double pt[DIM];
	double zero[DIM] = {0.0};
	if(reference==NULL) reference = zero;
	for(int i=0; i<n; i++) {
		psd_get_random_point(psd, pt);
		for(int d=0; d<DIM; d++) {
			fprintf(fd,"%e ",pt[d]+reference[d]);
		}
		fprintf(fd,"\n");
	}
}


//template<int DIM>
void psd_ensemble_astra(psd<2>* psd, FILE* fd, int n, unsigned int seed, double* reference=NULL) {
	double x,y,z,px,py,pz,clock,charge;
	int index,flag;
	double zero[2] = {0.0};
	if(reference==NULL) reference = zero;
	x=0; y=0; z=reference[0];
	px=0; py=0; pz=reference[1];
	clock=0; charge=psd->weight*1e+9/(n+1); index=1; flag=5;
	fprintf(fd,"%+e %+e %+e %+e %+e %+e %+e %+e %d %d\n",
		x, y, z, px, py, pz, clock, charge, index, flag); // reference prtcl
	srand(seed);
	psd->root->set_weight_integral();
	double pt[2];
	for(int i=0; i<n; i++) {
		psd_get_random_point(psd, pt);
		z  = pt[0];
		pz = pt[1];
		fprintf(fd,"%+e %+e %+e %+e %+e %+e %+e %+e %d %d\n",
			x, y, z, px, py, pz, clock, charge, index, flag);
	}
}

}

#endif
