/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef SLV_PSDEVAL_H
#define SLV_PSDEVAL_H

#include <psd.h>
#include <interpolate.h>

#include <cstdint>


namespace SLV {

template<int DIM>
inline size_t to_idx(double* x) {
	size_t res = 0;
	for(int d = 0; d<DIM; d++) {
		x[d] *= 2;
		res += (1<<d) * (int) x[d];
		x[d] -= (int) x[d];
	}
	return res;
}

template<int DIM>
const cell<DIM>* map_to_leaf(const cell<DIM>* cell, const double* z, double* out, int topology ) {

	// Map input to unit-square
	for(int d = 0; d<DIM; d++) {
		out[d] = (z[d]-(cell->cpoint[d]-cell->cwidth[d]))/(2*cell->cwidth[d]);
		if( topology & (1<<d) ) {
			out[d] = fmod(out[d],1);
			out[d] = out[d]>=0?out[d]:1+out[d];
		}
		if(out[d]<0 || out[d]>=1) {
			return NULL;
		}
	}

	while(cell->cells[0] != NULL) {
		size_t idx = to_idx<DIM>(out); // Also transforms out into next cell
		cell = cell->cells[idx];
	}

	return cell;
}


template<int DIM>
void psd_eval(const double* zin, double* out, int npts, const void* arg) {
	const psd<DIM>* psi = (const psd<DIM>*) arg;
	double x[DIM];
	double grid[512] = {0};
	const cell<DIM>* cell;
	const double* z = zin;
	for(int i=0; i<npts; i++) {
		cell = map_to_leaf(psi->root,z,x,psi->topology);
		if( NULL==cell ) {
			out[i] = 0.0;
		} else {
			switch(psi->interpolation) {
			case 0:
				cell->get_grid(x,x,grid,1);
				out[i] = grid[0];
				break;
			case 1:
				cell->get_grid(x,x,grid,2);
				out[i] = interpolate_linear(grid,x,DIM);
				break;
			case 2:
				cell->get_grid(x,x,grid,4);
				out[i] = interpolate_cubic(grid,x,DIM);
				break;
			case 3:
				cell->get_grid(x,x,grid,4);
				out[i] = interpolate_N4C1(grid,x,DIM);
				break;
			case 4:
				cell->get_grid(x,x,grid,4);
				out[i] = interpolate_N4C2(grid,x,DIM);
				break;
			default:
				out[i] = 0.0;
				break;
			}
			out[i] = fabs(out[i])>psi->epsilon?out[i]:0.0;
		}
		z+=DIM;
	}
}

template<int DIM>
void psd_eval_support(const double* zin, double* out, int npts, const void* arg) {
	const psd<DIM>* psi = (const psd<DIM>*) arg;
	double x[DIM];
	const cell<DIM>* cell;
	const double* z = zin;
	for(int i=0; i<npts; i++) {
		cell = map_to_leaf(psi->root,z,x,psi->topology);
		if( cell && cell->data && cell->integral()>0.0 ) {
			for(int j=0; j<npts; j++) out[j]=1;
			return;
		}
		z+=DIM;
	}
	for(int j=0; j<npts; j++) out[j]=0;
}

}

#endif


