/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef SLV_PSDFILL_H
#define SLV_PSDFILL_H


#ifndef SLV_MULTITHREAD
#define SLV_MULTITHREAD 1
#endif

#include <psd.h>
#include <cstdio>
#include <pthread.h>
#include <sys/sysinfo.h>

namespace SLV {

int get_nthreads() {
	const char* s = getenv("SLV_NTHREADS");
	if(s==NULL) {
		return get_nprocs();
	} else {
		return atoi(s);
	}
}


template<int DIM>
struct sample_thread_args {
	SLV::cell<DIM>** cells;
	size_t ncells;
	fun_t fun;
	const void* farg;
	map_t map;
	const void* marg;
	double t;
	int nexp;
	volatile bool* ready;
};

template<int DIM>
void* sample_thread(void* args_) {
	struct sample_thread_args<DIM>* args = (sample_thread_args<DIM>*) args_;
	for(size_t i=0; i<args->ncells; i++) {
		args->cells[i]->sample(args->fun,args->farg,args->map,args->marg,-1.0*args->t,args->nexp);
		// TODO
		// If zero-integral cells are dealloc'd, nexp is set to 0,
		// which causes trouble, when interpolating in these cells.
		//if(args->cells[i]->integral() == 0.0 ) {
		//	args->cells[i]->dealloc();
		//}
	}
	return NULL;
}

/// Initialize the PSD by means of a flood fill algorithm.
/// The tree is refined at \p pt to a depth of \p maxdepth.
/// Starting from the generated cell at \pt it is checked
/// whether there is density on the cells edges. If so,
/// a neighbor on the respective edge is generated and the process
/// gets repeated.
template<int DIM>
void fill(psd<DIM>* psd, fun_t fun, const void* farg, map_t map, const void* marg, double t, int pre_nexp = 2, double* pt = NULL) {
	double seed[DIM] = {0.0};
	if(map!=NULL) map(seed,seed,1,marg,t);
	if(pt==NULL)  pt = seed;
	cell<DIM>* cell = psd->root->refine_at(pt,psd->depth);
	
	#if 1
	// If function is a psd, use characteristic function (i.e. support)
	// during the discovery step
	if(fun == (fun_t) SLV::psd_eval<DIM>) {
		//fill_worker(SLV::psd_eval_support<DIM>,farg,map,marg,t,cell,psd->nexp);
		fill_worker(SLV::psd_eval_support<DIM>,farg,map,marg,t,cell,pre_nexp);
	} else {
		fill_worker(fun,farg,map,marg,t,cell,pre_nexp);
	}
	#endif

	#if 0
	// If function is a psd, temporarily set the interpolation 
	// to zero, during the discovery step
	SLV::psd<DIM>* psdin = NULL;
	int interpolation;
	if(fun == (fun_t) SLV::psd_eval<DIM>) {
		psdin = (SLV::psd<DIM>*) farg;
		interpolation = psdin->interpolation;
		psdin->interpolation = 0;
	}
	fill_worker(fun,farg,map,marg,t,cell,pre_nexp);
	// Reset interpolation
	if(psdin!=NULL) psdin->interpolation=interpolation;
	#endif

	#if 0
	fill_worker(fun,farg,map,marg,t,cell,pre_nexp);
	#endif

	SLV::cell<DIM>** leafs;
	size_t nleafs;
	//leafs = psd->root->list_leafs(&nleafs);
	leafs = psd->root->list_cells_at_level(&nleafs,psd->depth);
	//leafs = psd->root->list_leafs_partition(&nleafs,RANK,NRANKS);
	//fprintf(stderr,"fill: leafs: %lu\n",nleafs);

	#if SLV_MULTITHREAD
	size_t nthreads = get_nthreads();
	pthread_t threads[nthreads];
	sample_thread_args<DIM> args[nthreads];
	for(size_t n=0; n<nthreads; n++) {
		size_t ncells= nleafs/nthreads;
		args[n] = {
			.cells = &leafs[n*ncells],
			.ncells = ncells + (n==nthreads-1 ? nleafs%nthreads : 0),
			.fun = fun,
			.farg = farg,
			.map = map,
			.marg = marg,
			.t = t,
			.nexp = psd->nexp
		};
		pthread_create(&threads[n],NULL,sample_thread<DIM>,(void*) &args[n]);
	}

	for(size_t n=0; n<nthreads; n++) {
		pthread_join(threads[n],NULL);
	}
	//fprintf(stderr,"fill: done\n");
	#endif

	free(leafs);
}



template<int DIM>
void fill_worker(fun_t fun, const void* fargs, map_t map, const void* margs, double t, SLV::cell<DIM>* cell, int nexp)  {

	// If cell is already filled, nothing to do
	if( NULL == cell || cell->filled() ) {
		return;
	}

	// Fill cell
	cell->sample(fun, fargs, map, margs, -1.0*t, nexp);

	// Recursion
	if(cell->integral() > 0.0 ) {
		for(int nbor=0; nbor<2*DIM; nbor++) { // Loop over neighbors
			if( cell->nbors[nbor] == NULL ) { // If neighbor does not exist
				cell->make_neighbor(nbor); // Generate neighbor ...
			}
			fill_worker(fun,fargs,map,margs,t,cell->nbors[nbor],nexp);
		}
	}
}

}


#endif
