/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <cstdio>
#include <cstdlib>
#include <math.h>

#include "psd_frommoments.h"

static int compar(const void * a, const void * b) {
	return ((double*)a)[0] > ((double*)b)[0];
}

static double interpolate(const double* data, int n, double* limits, double x) {
	double f[4];
	double D = (limits[1]-limits[0])/(n-1);
	double d = fmod(x-limits[0],D)/D;
	int l = floor((x-limits[0])/D);
	for(int j=0; j<4; j++) {
		int m = l-1+j;
		f[j] = ( m>=0 && m<(n))?data[m]:0.0;
	}

	double res = 0.0;
	for(int i=0; i<4; i++) {
		double prod = 1.0;
		for(int k=0; k<4; k++) {
			if(i!=k) {
				prod*= (1.+d-k)/(i-k);
			}
		}
		res += f[i] * prod;
	}
	return res;
}

double* calc_moments(double* data, int npts, int nslices) {
	const int order = 4;

	qsort(data,npts,2*sizeof(double),*compar);

	double* moments = (double*) calloc(order * nslices,sizeof(double));
	double dslice = (data[2*(npts-1)]-data[0])/nslices;

	int n = 0;
	for(int s=0; s<nslices; s++) {

		double ubound = data[0] + (s+1)*dslice; // Upper bound of slice
		double x,xn;
		while(n<npts && data[2*n]<=ubound) { // count particles in slice
			x = data[2*n+1];
			xn = 1;
			for(int o=0; o<order; o++) {
				moments[o*nslices+s] += xn;
				xn *= x;
			}
		n++;} //increment total and slice particle counter

		int count = moments[0*nslices+s];
		if(count > 0) {
			moments[0*nslices+s] = moments[0*nslices+s]/npts/dslice;
			moments[1*nslices+s] = moments[1*nslices+s]/count; double mu = moments[1*nslices+s];
			moments[2*nslices+s] = moments[2*nslices+s]/count - mu*mu; double var = moments[2*nslices+s];
			moments[3*nslices+s] = (moments[3*nslices+s]/count  - 3*mu*var - mu*mu*mu)/(var*sqrt(var));
		} else { // if no particles, default to values of previous slice
			moments[0*nslices+s] = moments[0*nslices+s-1];
			moments[1*nslices+s] = moments[1*nslices+s-1];
			moments[2*nslices+s] = moments[2*nslices+s-1];
			moments[3*nslices+s] = moments[3*nslices+s-1];
		}

	}

	// Smooth the moments again
	double newval = 0.0;
	double preced = 0.0;
	double curval = 0.0;
	for(int smooth=0; smooth<8; smooth++) {
		for(int o=0; o<order; o++) {
			moments[o*nslices+0] = 0;
			moments[o*nslices+nslices-1] = 0;
			for(int s=1; s<nslices-1; s++) {
				curval = moments[o*nslices+s];
				if(!isnormal(curval)) curval=0.0;
				newval =
					+curval / 2.
					+preced / 4.
					+moments[o*nslices+s+1] / 4.;
				preced = curval;
				moments[o*nslices+s] = newval;
			}
		}
	}

	return moments;
}

void psd_frommoments(const double* in, double* out, int N, const void* arg_) {
	psd_frommoments_args* args = (psd_frommoments_args*) arg_;

	const double* z = in;
	for(int i=0; i<N; i++) {
		double m[args->order] = {0};

		for(int o=0; o<args->order; o++) {
			m[o] = interpolate(args->moments+(o*args->nslices),args->nslices,args->limits,z[0]);
		}

		double pc = z[1]-m[1];
		double pc2 = pc*pc;
		double var = m[2]+args->spread*args->spread;
		if(var<=0.0) *out =  0.0;
		else if(fabs(pc) > args->cutoff * sqrt(fabs(var))) *out = 0.0;
		else *out = m[0] * 1.0/sqrt(2*M_PI*var) * exp(-0.5*pc2/var);
		z+=2; out++;
	}
}
