/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef SLV_PSDPROJECTION_H
#define SLV_PSDPROJECTION_H

#include <psd.h>

namespace SLV {

template<typename T>
unsigned int nbits_set(T i) {
	int r;
	for(r=0; i; i>>=1) {
		r+=i&1;
	}
	return r;
}



struct grid_t {
	int dim;
	int nexp;
	double* data;
	double* limits;
};

template<int DIM>
struct grid_t psd_projection(const SLV::psd<DIM>* psd, unsigned int dim_bits, char method=0, ffun_t f=NULL, void* args=NULL) {

	grid_t p = {0}; // projection grid
	p.dim = DIM - nbits_set(dim_bits); // Dimension of the projection
	if(p.dim <= 0) return p;
	p.nexp = (psd->nexp+psd->depth);
	int npts = 1<<(p.dim*p.nexp);
	p.data = (double*) calloc(npts,sizeof(double));
	p.limits = (double*) malloc(2*p.dim*sizeof(double));

	//fprintf(stderr,"%p\n",p.data);


	// Calculate which dimensions to project along
	bool keep_dim[DIM];
	for(int d=0; d<DIM; d++) {
		keep_dim[d] = (((dim_bits>>d)&1) == 0);
	}

	// Calculate limits
	for(int d=0,dd=0; d<DIM; d++) {
		if(keep_dim[d]) {
			double cp = psd->root->cpoint[d];
			double cw = psd->root->cwidth[d];
			p.limits[0*p.dim+dd] = cp - cw;
			p.limits[1*p.dim+dd] = cp + cw;
			dd++;
		}
	}
	// Calculate ``volume'' of cell in integrated dimensions
	//double norm=psd->weight;
	double norm = 1;
	for(int d=0; d<DIM; d++) {
		if( not keep_dim[d] ) {
			double cw = psd->root->cwidth[d];
			norm *= (2.*cw)/(1<<p.nexp);
		}
	}

	cell_projection(psd->root,p,keep_dim,0,method,f,args);

	// Normalize
	for(int i=0; i<npts; i++) {
		p.data[i] *= norm;
	}

	//fprintf(stderr,"%p\n",p.data);

	return p;
}

template<int DIM>
void cell_projection(cell<DIM>* const cell, grid_t p, const bool* keep_dim, int depth, int method=0, ffun_t f=NULL, void* args=NULL) {
	const double* src = cell->data;
	double* dest = p.data;

	double NC1[] = {1.0};
	double NC2[] = {2./6., 4./6.};
	double NC4[] = {14./90., 32./90., 12./90., 32./90.};
	double* NC[] = {NC1,NC2,NC4};
	int NCmod[] = {1,2,4};

	if(cell->data!=NULL) {

		double* vals=NULL;
		if( f!=NULL ) {
			const int npts = cell->npoints();
			vals = (double*) malloc(sizeof(double) * npts);
			double coords[npts*DIM];
			cell->grid_coords(coords);
			f(cell->data,coords,vals,npts,args);
			src=vals;
		}

		for(int i=0; i<(1<<(DIM*cell->nexp)); i++) {
			size_t offset = 0;
			double factor = 1;
			for(int d=0,dd=0; d<DIM; d++) {
				if( keep_dim[d] ) {
					int idx = (i/(1<<(d*cell->nexp))) % (1<<cell->nexp);
					offset += idx  * (1<<(dd*p.nexp));
					//fprintf(stderr,"%04d %04d %04d\n",d,idx,offset);
					dd++;
				} else {
					int idx = (i/(1<<(d*cell->nexp)));
					factor *= NCmod[method]*NC[method][idx%NCmod[method]]; /// TODO???
				}
			}
			//fprintf(stderr,"%04d -> %04lu (%e)\n",i,offset,src[i]);
			//printf("%02d %e\n",i,factor);
			dest[offset] += src[i] * factor;
		}
		//printf("\n");
		
		free(vals);
		return;
	}

	depth += 1;
	for(int n=0; n<1<<DIM; n++) {
		size_t offset = 0;
		for(int d=0,dd=0; d<DIM; d++) {
			if( keep_dim[d] ) {
				dd++;
				if( ((n>>d)&1)==1) {
					offset += (1<<(p.nexp*dd - depth));
				}
			}
		}
		p.data = dest+offset;
		if(cell->cells[n] != NULL) {
			cell_projection(cell->cells[n], p, keep_dim, depth, method, f, args);
		}
		
	}
}



}

#endif
