/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef SLV_PSDSAVE_H
#define SLV_PSDSAVE_H

#include <cstdio>
#include <cstdint>
#include <error.h>

const uint64_t SLV_MAGIC = 0xAFFEAFFE;
const uint64_t SLV_BINVERSION = 2;
int SLV_ERR = 0;

namespace SLV {

template <int DIM>
void serialize(cell<DIM>* cell, FILE* f) {
	uint8_t refined = cell->cells[0]==NULL?0:1;
	uint8_t nexp = cell->nexp;
	
	fwrite(&refined,sizeof(refined),1,f);
	fwrite(&nexp,sizeof(nexp),1,f);
	if(nexp>0) {
		fwrite(cell->data,sizeof(cell->data[0]),1<<(nexp*DIM),f);
	}

	// Recursion
	if(refined==1) {
		for(int i=0; i<1<<DIM; i++) {
			serialize<DIM>(cell->cells[i],f);
		}
	}
}

template <int DIM>
void deserialize(cell<DIM>* cell, FILE* f) {
	uint8_t refined;
	uint8_t nexp;
	fread(&refined,sizeof(refined),1,f);
	fread(&nexp,sizeof(nexp),1,f);
	if(nexp>0) {
		cell->alloc(nexp);
		fread(cell->data,sizeof(cell->data[0]),1<<(nexp*DIM),f);
	}
	// Recursion
	if(refined==1) {
		cell->refine();
		for(int i=0; i<1<<DIM; i++) {
			deserialize<DIM>(cell->cells[i],f);
		}
	}
}

template <int DIM>
void save(psd<DIM>* psd, FILE* f) {
	fwrite(&SLV_MAGIC,sizeof(SLV_MAGIC),1,f);
	fwrite(&SLV_BINVERSION,sizeof(SLV_BINVERSION),1,f);

	fwrite(&psd->depth,sizeof(psd->depth),1,f);
	fwrite(&psd->nexp,sizeof(psd->nexp),1,f);
	fwrite(&psd->weight,sizeof(psd->weight),1,f);
	fwrite(&psd->refpoint,sizeof(psd->refpoint),1,f);
	fwrite(&psd->topology,sizeof(psd->topology),1,f);
	fwrite(&psd->interpolation,sizeof(psd->interpolation),1,f);

	fwrite(&psd->root->cpoint,sizeof(psd->root->cpoint),1,f);
	fwrite(&psd->root->cwidth,sizeof(psd->root->cwidth),1,f);

	serialize<DIM>(psd->root,f);
}

template <int DIM>
psd<DIM>* load(FILE* f) {
	uint64_t magic, binversion;
	fread(&magic,sizeof(SLV_MAGIC),1,f);
	fread(&binversion,sizeof(SLV_BINVERSION),1,f);
	if(magic != SLV_MAGIC)           { SLV_ERR = 1; error(1,0,"Wrong file format\n"); }
	if(binversion != SLV_BINVERSION) { SLV_ERR = 2; error(1,0,"Wrong binary version\n"); }

	psd<DIM>* psd = new SLV::psd<DIM>({0},{0});
	
	fread(&psd->depth,sizeof(psd->depth),1,f);
	fread(&psd->nexp,sizeof(psd->nexp),1,f);
	fread(&psd->weight,sizeof(psd->weight),1,f);
	fread(&psd->refpoint,sizeof(psd->refpoint),1,f);
	fread(&psd->topology,sizeof(psd->topology),1,f);
	fread(&psd->interpolation,sizeof(psd->interpolation),1,f);

	fread(&psd->root->cpoint,sizeof(psd->root->cpoint),1,f);
	fread(&psd->root->cwidth,sizeof(psd->root->cwidth),1,f);

	psd->set_topology(psd->topology);
	deserialize<DIM>(psd->root,f);
	
	SLV_ERR = 0;
	return psd;
}



}

#endif
