/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef SLV_PSDSTATS_H
#define SLV_PSDSTATS_H

#include "psd.h"
#include <cstdio>

namespace SLV {

template<int DIM>
using cellfun_t = void (*)(const SLV::cell<DIM>*, void*, void*);


template<int DIM>
void centroid__(const SLV::cell<DIM>* cell, void* arg, void* out_) {
	if(NULL==cell->data) return;
	double* centr = (double*) arg;
	double* out = (double*) out_;

	const int npts = cell->npoints();
	double coords[npts*DIM];
	cell->grid_coords(coords);

	double zeros[DIM] = {0};
	if(NULL==centr) centr = zeros;

	double V = 1; // Volume
	for(int d=0; d<DIM; d++) V *= 2.*cell->cwidth[d];
	V /= 1<<(DIM*cell->nexp);

	const double* data = cell->data;
	const double* z = coords;
	for(int n=0; n < npts; n++) {
		for(int d=0; d < DIM; d++) {
			double x=(z[d]-centr[d]);
			out[d] += *data * x * V;
		}
		data+=1;
		z+=DIM;
	}
}

template<int DIM>
void variance__(const SLV::cell<DIM>* cell, void* arg, void* out_) {
	if(NULL==cell->data) return;
	double* centr = (double*) arg;
	double* out = (double*) out_;

	const int npts = cell->npoints();
	double coords[npts*DIM];
	cell->grid_coords(coords);

	double zeros[DIM] = {0};
	if(NULL==centr) centr = zeros;

	double V = 1; // Volume
	for(int d=0; d<DIM; d++) V *= 2.*cell->cwidth[d];
	V /= 1<<(DIM*cell->nexp);

	const double* data = cell->data;
	const double* z = coords;
	double x[DIM];
	for(int n=0; n < npts; n++) {
		for(int d=0; d < DIM; d++) {
			x[d]=(z[d]-centr[d]);
		}
		for(int i=0; i < DIM; i++) {
			for(int j=0; j < DIM; j++) {
				out[i+j*DIM] += *data * V * x[i]*x[j];
			}
		}
		data+=1;
		z+=DIM;
	}
}

template<int DIM>
void walk(const SLV::cell<DIM>* cell, cellfun_t<DIM> fn, void* arg, void* out) {
	fn(cell,arg,out);
	for(int i=0; i<1<<DIM; i++) {
		if(cell->cells[i]) {
			walk(cell->cells[i],fn,arg,out);
		}
	}
}

template<int DIM>
void centroid(const SLV::psd<DIM>* psd, double* out) {
	SLV::walk(psd->root,centroid__<DIM>,NULL,out);
}

template<int DIM>
void variance(const SLV::psd<DIM>* psd, double* out) {
	double cnt[DIM] = {0};
	SLV::walk(psd->root,centroid__<DIM>,NULL,cnt);
	SLV::walk(psd->root,variance__<DIM>,cnt,out);
}

template<int DIM>
void psd_stats(const SLV::psd<DIM>* psd, FILE* fd = stdout) {

	double cnt[DIM] = {0};
	SLV::walk(psd->root,centroid__<DIM>,NULL,cnt);

	double var[DIM*DIM] = {0};
	SLV::walk(psd->root,variance__<DIM>,cnt,var);

	for(int i=0; i<DIM; i++) {
		fprintf(fd,"%e ",cnt[i]);
	}
	fprintf(fd,"\n");
	for(int i=0; i<DIM*DIM; i++) {
		fprintf(fd,"%e ",var[i]);
	}
	fprintf(fd,"\n");

}


}
#endif
