/*
* libselav - a framework for semi-lagrangian simulation of sparse densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef SLV_PSDSTEP_H
#define SLV_PSDSTEP_H

#include <psd.h>
#include <psd_fill.h>

namespace SLV {

template<int DIM>
static void ensemble_bbox(const double* data, double* limits, size_t N) {
	// find limits
	for(int d=0; d<DIM; d++) {
		limits[d+0*DIM] =  INFINITY;
		limits[d+1*DIM] = -INFINITY;
	}
	double* lim;
	for(size_t i=0; i<N*DIM; i++) {
		int d = i%DIM;
		lim = limits+d;
		*lim = data[i]<*lim ? data[i]: *lim;
		lim = limits+d+DIM;
		*lim = data[i]>*lim ? data[i]: *lim;
	}
}

enum step_strat {KEEP, EQUAL, AUTO, AREA};

struct psd_step_args {
	int nexp;
	int depth;
	enum step_strat center_strat;
	enum step_strat box_strat;
	double* limits;
	int pre;
	double epsilon;
};


template<int DIM>
psd<DIM>* psd_step(const psd<DIM>* psd_in, map_t map, const void* marg, double t, struct psd_step_args args) {
	auto cpoint = psd_in->root->cpoint;
	auto cwidth = psd_in->root->cwidth;

	SLV::psd<DIM>* psd = new SLV::psd<DIM>(*psd_in); 
	psd->root = new SLV::cell<DIM>(cpoint,cwidth);
	psd->set_topology(psd_in->topology);
	psd->epsilon=args.epsilon;
	if(args.nexp > 0) {
		psd->nexp = args.nexp;
	}

	double* limits;
	double limits__[2*DIM];
	if(NULL==args.limits) {
		// Get center points of tracer cells
		size_t ntracer;
		SLV::cell<DIM>** tracer_cells = psd_in->root->list_leafs(&ntracer);
		//SLV::cell<2>** tracer_cells = psd_in->root->list_cells(&ntracer,6);
		//fprintf(stderr,"ntracer: %lu\n",ntracer);
		double ztrace[DIM*ntracer];
		for(size_t i=0; i<ntracer; i++) {
			for(int d=0; d<DIM; d++) {
				ztrace[i*DIM+d] = tracer_cells[i]->cpoint[d];
			}
		}
		free(tracer_cells);


		// Map tracers forward
		map(ztrace,ztrace,ntracer,marg,t);

		// New limits
		ensemble_bbox<DIM>(ztrace,limits__,ntracer);
		limits = limits__;
	} else {
		limits = args.limits;
	}


	// CENTER
	switch(args.center_strat) {
	case KEEP:
		for(int d=0; d<DIM; d++) {
			psd->root->cpoint[d] = psd_in->root->cpoint[d];
		}
		break;
	case AUTO:
		for(int d=0; d<DIM; d++) {
			psd->root->cpoint[d] = (limits[1*DIM+d]+limits[0*DIM+d])/2;
			//psd->root->cwidth[d] = (limits[1*DIM+d]-limits[0*DIM+d])/2;
			/// ^^^ if commented in, weird oscillations show up amstutz/2021-08-10
		}
		//printf("%e %e %e %e\n",limits[0],limits[1],limits[2],limits[3]);
		break;
	default:
		break;
	}


	if(NULL==args.limits) { // If limits are not explicitly set

		// Calculate scaling factors in all dimensions
		double scales[DIM];
		double maxscale = 0;
		for(int d=0; d<DIM; d++) {
			double neg = (psd->root->cpoint[d] - limits[d]) / psd_in->root->cwidth[d];
			double pos = (limits[d+DIM] - psd->root->cpoint[d]) / psd_in->root->cwidth[d];
			scales[d] = pos>neg?pos:neg;
			maxscale = scales[d]>maxscale ? scales[d] : maxscale;
		}

		// BOX
		switch(args.box_strat) {
		case KEEP:
			for(int d=0; d<DIM; d++) {
				psd->root->cwidth[d] = psd_in->root->cwidth[d];
			}
			break;
		case EQUAL: {
				int exp = ceil(log2(maxscale));
				double scale = pow(2,exp);
				psd->depth += exp;
				for(int d=0; d<DIM; d++) {
					psd->root->cwidth[d] *= scale;
				}
			}
			break;
		case AUTO: {
				for(int d=0; d<DIM; d++) {
					int exp = ceil(log2(scales[d]));
					double scale = pow(2,exp);
					psd->depth += exp;
					psd->root->cwidth[d] *= scale;
				}
			}
			break;
		case AREA: {
				int n0 = psd->nexp + psd->depth;
				double scale = 1;
				for(int d=0; d<DIM; d++) {
					double s = (limits[d+DIM]-limits[d+0])/(2*psd->root->cwidth[d]);
					psd->root->cwidth[d] *= s;
					scale *= s;
				}
				double n1 = log2(scale)/DIM + n0;
				int n2 = ceil(n1);
				for(int d=0; d<DIM; d++) {
					psd->root->cwidth[d] *= pow(2,((double)n2)-n1);
				}
				psd->depth = n2 - psd->nexp;
			}
			break;
		}
	}

	if(args.depth>0) {
		psd->depth = args.depth;
	}

	// Fill
	SLV::fill<DIM>(psd,SLV::psd_eval<DIM>,psd_in,map,marg,t,args.pre);

	// Update reference point
	if(args.center_strat != KEEP) {
		double center[DIM] = {0.0};
		map(center,center,1,marg,t); // Map center forward
		for(int d=0; d<DIM; d++) {
			psd->refpoint[d] += center[d];
			center[d] *= -1.0; // for subsequent translation of the tree
		}
		psd->root->translate(center);
	}

	return psd;
}

}

#endif
