/*
* selav1d - semi-lagrangian simulation of sparse phase-space densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

%option noyywrap nounput
%option yylineno
%option nodefault

%{

#include "main.tab.h"

int comment_nest;
#ifndef BBUFSIZE
	#define BBUFSIZE (100*1000)
#endif
char buf[BBUFSIZE];
int block_nest;

#include "include_stack.h"

%}

%x STR STR2 BLCK
%x COMMENT COMMENT2

EXP ([Ee][-+]?[0-9]+)

%%


"''" {yylval.sval = strdup(""); return STRING;} /* Empty string */
"'" BEGIN(STR); /* Reads STRINGs */
<STR>[^']* {yylval.sval = strdup(yytext); return STRING;}
<STR>"'" {BEGIN(INITIAL);}

"\"\"" {yylval.sval = strdup(""); return STRING;} /* Empty string */
"\"" BEGIN(STR2); /* Reads STRINGs */
<STR2>[^\"]* {yylval.sval = strdup(yytext); return STRING;}
<STR2>"\"" {BEGIN(INITIAL);}

"{" { block_nest = 1; buf[0]=0; BEGIN(BLCK);} /* Reads STRINGs */
<BLCK>{
	[^{}]* { strcat(buf,yytext); }
	"{" { block_nest++; strcat(buf,yytext); }
	"}" { block_nest--;
		if(block_nest<=0) {
			yylval.sval=strdup(buf);
			BEGIN(INITIAL);
			return BLOCK;
		} else {
			strcat(buf,yytext);
		}
	}
}

"//" BEGIN(COMMENT);
<COMMENT>{
	. /* ignore */
	\n { BEGIN(INITIAL);}
}

"/*" { comment_nest=1; BEGIN(COMMENT2); };
<COMMENT2>{
	. /* ignore */
	"\n"  /* ignore */
	"/*" { comment_nest++; }
	"*/" { comment_nest--; if(comment_nest<=0) BEGIN(INITIAL); }
}





<<EOF>> {
	buffer_stack_pop();
	if(buffer_stack_ptr<0) yyterminate();
	return ACCEPT;
}
"exit" {
	exit(1);
	buffer_stack_pop();
	if(buffer_stack_ptr<0) yyterminate();
	return ACCEPT;
}

[ \t\n] /* ignore white space */;

 /* floats */
([0-9]*\.[0-9]+|[0-9]+\.){EXP}? { yylval.fval = atof(yytext); return FLOAT; }
[0-9]+{EXP}                     { yylval.fval = atof(yytext); return FLOAT; }
[0-9]+                          { yylval.fval = atof(yytext); return FLOAT; }

"+"   { return PLUS; }
"-"   { return MINUS; }
"**"  { return POW; }
"*"   { return TIMES; }
"/"   { return DIVIDE; }
":="  { return COLONEQUAL; }
"sin" { return SIN; }
"cos" { return COS; }
"asin" { return ASIN; }
"acos" { return ACOS; }
"tan" { return TAN; }
"atan2" { return ATAN2; }
"sqrt" { return SQRT; }
"floor" { return FLOOR; }
"ceil" { return CEIL; }
"abs" { return ABS; }
"pi"  { yylval.fval = M_PI ; return FLOAT; }
"linspace" { return LINSPACE; }

"[" { return LBRACKET; }
"]" { return RBRACKET; }

">>" { return GTGT; }
">"  { return GT; }
"<"  { return LT; }
">=" { return GE; }
"<=" { return LE; }
"==" { return EQ; }
"!=" { return NE; }

"psd_analytic"     { return PSD_ANALYTIC; }
"psd_gauss"        { return PSD_GAUSS; }
"psd_ensemble"     { return PSD_ENSEMBLE; }
"psd_rectangle"    { return PSD_RECTANGLE; }
"psd_test"         { return PSD_TEST; }
"psd_bump"         { return PSD_BUMP; }

"map_translate"    { return MAP_TRANSLATE; }
"map_rotate"       { return MAP_ROTATE; }
"map_cavity"       { return MAP_CAVITY; }
"map_kickl"        { return MAP_KICKL; }
"map_driftl"       { return MAP_DRIFTL; }
"map_kickpoly"     { return MAP_KICKPOLY; }
"map_driftpoly"    { return MAP_DRIFTPOLY; }
"map_kicksine"     { return MAP_KICKSINE; }
"map_driftsine"    { return MAP_DRIFTSINE; }
"map_hyperbolic"   { return MAP_HYPERBOLIC; }
"map_chicanec"     { return MAP_CHICANEC; }
"map_spacecharge"  { return MAP_SPACECHARGE; }
"map_csr"          { return MAP_CSR; }
"map_analytic"     { return MAP_ANALYTIC; }
"map_identity"     { return MAP_IDENTITY; }
"map_poisson"      { return MAP_POISSON; }
"map_poisson1d"    { return MAP_POISSON1D; }
"map_compose"      { return MAP_COMPOSE; }

"propagate" { return PROPAGATE; }

"write_projection"   { return WRITE_PROJECTION; }
"write_localmoments" { return WRITE_LOCALMOMENTS; }
"write_grid"         { return WRITE_GRID; }
"write_ensemble"     { return WRITE_ENSEMBLE; }
"integral"           { return INTEGRAL; }
"charge"             { return CHARGE; }
"multiply"           { return MULTIPLY; }
"modify"             { return MODIFY; }
"noise"              { return NOISE; }
"normalize"          { return NORMALIZE; }

"stats"     { return STATS; }
"centroid"  { return CENTROID; }
"variance"  { return VARIANCE; }
"maximum"   { return MAXIMUM; }
"reference" { return REFERENCE; }
"show"      { return SHOW; }
"plot"      { return PLOT; }
"save"      { return SAVE; }
"load"      { return LOAD; }
"copy"      { return COPY; }
"eval"      { return EVAL; }
"print"     { return PRINT; }
"include"   { return INCLUDE; }
"who"       { return WHO; }
"format"    { return FORMAT; }
"strcat"    { return STRCAT; }

"do"      { return DO; }
"for"     { return FOR; }
"in"      { return IN; }
"if"      { return IF; }
"else"    { return ELSE; }
"defined" { return DEFINED; }

"mkdir"  { return MKDIR; }
"chdir"  { return CHDIR; }
"getcwd" { return GETCWD; }
"getenv" { return GETENV; }
"strtod" { return STRTOD; }
"strcmp" { return STRCMP; }

"transfer4d" { return TRANSFER4D;  }
"chicanecoef" { return CHICANECOEF; }
"average"     { return AVERAGE; }

[a-zA-Z0-9_]+ { { yylval.sval = strdup(yytext); return SYMBOL; } }

. { return *yytext; }

%%

void parse_file__(const char* fname) {
	buffer_stack_push_file(fname);
	yyparse();
}

void parse_string__(const char* str) {
	buffer_stack_push_string(str);
	yyparse();
}

void print_buffer_position() {
	struct buf_t* buf = &buffer_stack[buffer_stack_ptr];
	printf("%s: line %d\n",buf->name,yylineno);
}
