/*
* selav1d - semi-lagrangian simulation of sparse phase-space densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

%{

#include <stdio.h>
#include <string>
#include <cstring>
#include <vector>
#include <error.h>
#include <map>
#include <set>
#include <unistd.h>
//#include <sys/types.h>
#include <sys/stat.h>


#include "selav.h"
#ifdef HAVE_X11
#include "X.h"
#endif
#include "var.h"


extern int yylex();
extern int yyparse();
extern FILE* yyin;
void yyerror(const char* s);

/* switch buffer; defined in .l */
extern void parse_file__(const char*);
extern void parse_string__(const char*);
extern void print_buffer_position();

environment ENV;
FILE* OUTFILE; // FIXME: If this declaration is before the ENV declaration
               //        we get segfaults on the first print.

/* ---------------- DEFAULTS -------------- */
namespace def {
	const int psd_depth = 7;
	const int psd_nexp = 2;
	const double psd_xmin = -1.0;
	const double psd_ymin = -1.0;
	const double psd_xmax =  1.0;
	const double psd_ymax =  1.0;

	const SLV::map_cavity_args map_cavity_args = {1e9, 0, 1e6, 0};
	const SLV::map_chicanec_args map_chicanec_args = {10, 0.5, 0.5, 0, 0};
	const double spacecharge_length = 1;
	const double spacecharge_beamsize = 1;
}


/* ---------------- CONSTANTS ------------- */
namespace cnst {
	static double c       = 299792458.0;      // Speed of light [m/s]
	static double eps0    = 8.8541878128e-12; // E-Field const [C/Vm]
	static double m_e     = 9.1093837015e-31; // Mass electron [kg]
	static double q_e     = -1.602176634e-19; // Charge electron [C]
	static double m_e_eV  = m_e*c*c / -q_e;   // Rest Energy electron [eV]
}


void accept_kwargs(environment* kwargs, std::set<std::string> accept) {
	for(auto& arg : *kwargs) {
		if (accept.find(arg.first) == accept.end()) {
			print_buffer_position();
			error(1,0,"Error: Keyword '%s' not recognized",arg.first.c_str());
		}
	}
}

/* ---------- PSD input ------------ */

SLV::psd<2>*  psd_from_kwargs(environment* kwargs, double* limits=NULL) {
	double xmin = def::psd_xmin;
	double ymin = def::psd_ymin;
	double xmax = def::psd_xmax;
	double ymax = def::psd_ymax;
	double weight = 1;
	int depth   = def::psd_depth;
	int nexp    = def::psd_nexp;
	int interpolation = 2;
	int topology = 0;
	for(auto& arg : *kwargs) {
		if      (arg.first == "depth")   depth = arg.second->fval[0];
		else if (arg.first == "nexp")    nexp = arg.second->fval[0];
		else if (arg.first == "weight")  weight = arg.second->fval[0];
		else if (arg.first == "limits") {
			assert(arg.second->size==4);
			limits = arg.second->fval;
		}
		else if (arg.first == "interpolation")  interpolation = arg.second->fval[0];
		else if (arg.first == "topology")  topology = arg.second->fval[0];
	}
	if(limits) {
		xmin = limits[0];
		ymin = limits[1];
		xmax = limits[2];
		ymax = limits[3];
	}
	auto psd = new SLV::psd<2>({(xmax+xmin)/2,(ymax+ymin)/2},{(xmax-xmin)/2,(ymax-ymin)/2});
	psd->depth = depth;
	psd->nexp = nexp;
	psd->weight = weight;
	psd->interpolation = interpolation;
	psd->set_topology(topology);

	return psd;
}

SLV::psd<2>*  psd_analytic__(char* fstr, environment* kwargs) {
	accept_kwargs(kwargs,{"depth","nexp","weight","limits","map","interpolation","topology"});
	SLV::psd<2>* psd = psd_from_kwargs(kwargs);
	mapwargs_t map = {NULL,NULL};
	for(auto& arg : *kwargs) {
		if (arg.first == "map")      map = *arg.second->mapval;
	}

	char b[16] = "q\0p\0";
	char* names[2] = {b,b+2};
	struct math_eval_args args = {fstr,names};
	SLV::fill(psd,eval_math<2>,&args,map.map,map.args,1);
	return psd;
}

SLV::psd<2>*  psd_gauss__(environment* kwargs) {
	accept_kwargs(kwargs,{"depth","nexp","weight","sig_q","sig_p","correlation","cutoff","limits","map","interpolation","topology"});
	struct gauss2D_args fargs = {0.2,0.2,0,3};
	mapwargs_t map = {NULL,NULL};
	for(auto& arg : *kwargs) {
		if      (arg.first == "sig_q") fargs.sig_q = arg.second->fval[0];
		else if (arg.first == "sig_p") fargs.sig_p = arg.second->fval[0];
		else if (arg.first == "correlation") fargs.correlation = arg.second->fval[0];
		else if (arg.first == "cutoff") fargs.cutoff = arg.second->fval[0];
		else if (arg.first == "map")      map = *arg.second->mapval;
	}
	double limits[4] = {
		-(fargs.cutoff+1)*fargs.sig_q,
		-(fargs.cutoff+1)*fargs.sig_p,
		 (fargs.cutoff+1)*fargs.sig_q,
		 (fargs.cutoff+1)*fargs.sig_p,
	};
	SLV::psd<2>* psd = psd_from_kwargs(kwargs,limits);
	SLV::fill(psd,gauss2D,&fargs,map.map,map.args,1);
	return psd;
}


void ensemble_bbox(const double* data, double* limits, size_t N, int dim) {
	// find limits
	for(int d=0; d<dim; d++) {
		limits[d+0*dim] =  INFINITY;
		limits[d+1*dim] = -INFINITY;
	}
	double* lim;
	for(size_t i=0; i<N*dim; i++) {
		int d = i%dim;
		lim = limits+d;
		*lim = data[i]<*lim ? data[i]: *lim;
		lim = limits+d+dim;
		*lim = data[i]>*lim ? data[i]: *lim;
	}
}

static inline double max(double a, double b) {
	return a>b?a:b;
}

SLV::psd<2>*  psd_ensemble__(char* fname, environment* kwargs) {
	accept_kwargs(kwargs,{"nslices","cutoff","weight","spread","depth","nexp","limits","map","interpolation","topology","type"});

	struct ensemble_t (*read_method)(FILE*) = read_ensemble;
	for(auto& arg : *kwargs) {
		if(arg.first == "type") {
		     	if(strcmp(arg.second->sval,"astra")==0) read_method = read_astra;
		     	if(strcmp(arg.second->sval,"plain")==0) read_method = read_ensemble;
		}
	}

	// Read ensemble
	FILE* fd = fopen(fname,"r");
	if(fd==NULL) { print_buffer_position(); perror(fname); exit(1); };
	struct ensemble_t ens = read_method(fd);
	fclose(fd);

	// Calculate bbox
	double limits[4];
	ensemble_bbox(ens.data,limits,ens.N,2);


	// Gather arguments
	psd_frommoments_args args;
	args.limits[0] = limits[0],
	args.limits[1] = limits[2];
	args.nslices = 32;
	args.cutoff  = 3;
	args.spread  = 0.0;
	mapwargs_t map = {NULL,NULL};
	double* weight = &ens.weight;
	bool limits_given = 0;
	for(auto& arg : *kwargs) {
		if      (arg.first == "nslices") args.nslices = arg.second->fval[0];
		else if (arg.first == "cutoff")  args.cutoff = arg.second->fval[0];
		else if (arg.first == "spread")  args.spread = arg.second->fval[0];
		else if (arg.first == "map")     map = *arg.second->mapval;
		else if (arg.first == "limits")  limits_given=true;
		else if (arg.first == "weight")  weight = arg.second->fval;
	}
	// Moments
	args.moments = calc_moments(ens.data,ens.N,args.nslices);

	SLV::psd<2>* psd = psd_from_kwargs(kwargs);
	psd->weight = *weight;
	psd->refpoint[1] = ens.reference[1];
	if(!limits_given) {
		psd->root->cpoint = {(limits[2]+limits[0])/2,(limits[3]+limits[1])/2};
		psd->root->cwidth = {(limits[2]-limits[0])/2,(limits[3]-limits[1])/2};
	}

	SLV::fill(psd,psd_frommoments,&args,map.map,map.args,1);

	free(ens.data);
	free(args.moments);
	return psd;
}

SLV::psd<2>*  psd_rectangle__(double* limits, environment* kwargs) {
	accept_kwargs(kwargs,{"depth","weight","nexp","limits","interpolation","topology"});
	SLV::psd<2>* psd = psd_from_kwargs(kwargs);
	SLV::fill(psd,rectangle,limits,NULL,NULL,1);
	return psd;
}

SLV::psd<2>*  psd_test__(environment* kwargs) {
	accept_kwargs(kwargs,{"depth","weight","nexp","limits","interpolation","topology"});
	double c=2;
	SLV::psd<2>* psd = psd_from_kwargs(kwargs);
	SLV::fill(psd,smiley2D,&c,NULL,NULL,1);
	return psd;
}

SLV::psd<2>*  psd_bump__(environment* kwargs) {
	accept_kwargs(kwargs,{"depth","weight","nexp","limits","interpolation","topology","sig_p","cutoff","width","dropoff","chirp"});
	SLV::psd<2>* psd = psd_from_kwargs(kwargs);
	double args[] = {0.1,5,0.5,0.1,0.0}; // sigma,cutoff,a,d,chirp
	for(auto& arg : *kwargs) {
		if      (arg.first == "sig_p")   args[0] = arg.second->fval[0];
		if      (arg.first == "cutoff")  args[1] = arg.second->fval[0];
		if      (arg.first == "width")   args[2] = arg.second->fval[0];
		if      (arg.first == "dropoff") args[3] = arg.second->fval[0];
		if      (arg.first == "chirp")   args[4] = arg.second->fval[0];
	}
	SLV::fill(psd,bump,&args,NULL,NULL,1);
	return psd;
}

/* ----------- MAPS ---------------- */

mapwargs_t* map_compose__(varlist* vl) {
	SLV::map_compose_args* args = new SLV::map_compose_args();
	args->n = vl->size();
	for(size_t i=0; i<vl->size(); i++) {
		args->maps[i] = (*vl)[i]->mapval->map;
		args->args[i] = (*vl)[i]->mapval->args;
	}
	return new mapwargs_t({SLV::map_compose,args});
}


mapwargs_t* map_translate__(const double* delta) {
	double* d = (double*) malloc(2*sizeof(double));
	memcpy(d,delta,2*sizeof(double));
	typedef SLV::map_translate_args arg_t;
	arg_t* margs = (arg_t*) malloc(sizeof(arg_t));
	*margs = {d,2};
	return new mapwargs_t({SLV::map_translate,margs});
}

mapwargs_t* map_rotate__(double a) {
	double* angle = (double*)  calloc(1,sizeof(double));
	*angle = a;
	return new mapwargs_t({SLV::map_rotate,angle});
}

mapwargs_t*  map_driftl__(double _d) {
	double* d = (double*)  calloc(1,sizeof(double));
	*d = _d;
	return new mapwargs_t({SLV::map_driftl,d});
}

mapwargs_t*  map_kickl__(double _k) {
	double* k = (double*)  calloc(1,sizeof(double));
	*k = _k;
	return new mapwargs_t({SLV::map_kickl,k});
}

mapwargs_t*  map_hyperbolic__(double _a) {
	double* a = (double*)  calloc(1,sizeof(double));
	*a = _a;
	return new mapwargs_t({SLV::map_hyperbolic,a});
}

mapwargs_t*  map_driftpoly__(int n, double* factors) {
	struct SLV::map_driftpoly_args* args = (struct SLV::map_driftpoly_args*) malloc(sizeof(struct SLV::map_driftpoly_args));
	args->n = n;
	memcpy(args->factors,factors,n*sizeof(double));
	return new mapwargs_t({SLV::map_driftpoly,args});
}

mapwargs_t*  map_kickpoly__(int n, double* factors) {
	struct SLV::map_kickpoly_args* args = (struct SLV::map_kickpoly_args*) malloc(sizeof(struct SLV::map_kickpoly_args));
	args->n = n;
	memcpy(args->factors,factors,n*sizeof(double));
	return new mapwargs_t({SLV::map_kickpoly,args});
}

mapwargs_t*  map_driftsine__(double a, double k, double phi) {
	double* args = (double*)  calloc(3,sizeof(double));
	args[0]=a; args[1]=k; args[2]=phi;
	return new mapwargs_t({SLV::map_driftsine,args});
}

mapwargs_t*  map_kicksine__(double a, double k, double phi) {
	double* args = (double*)  calloc(3,sizeof(double));
	args[0]=a; args[1]=k; args[2]=phi;
	return new mapwargs_t({SLV::map_kicksine,args});
}

mapwargs_t*  map_cavity__(environment* kwargs) {
	accept_kwargs(kwargs,{"freq","phase","ampl","mode"});
	typedef SLV::map_cavity_args arg_t;
	arg_t* margs = (arg_t*) malloc(sizeof(arg_t));
	*margs = def::map_cavity_args;
	for(auto& arg : *kwargs) {
		std::string kw = arg.first;
		var* v = arg.second;
		     if ( kw == "freq")  margs->freq  = v->fval[0];
		else if ( kw == "phase") margs->phase = v->fval[0];
		else if ( kw == "ampl")  margs->ampl  = v->fval[0];
		else if ( kw == "mode")  margs->mode  = v->fval[0];
	}
	//fprintf(stderr,"delta_E: %e\n",margs->ampl *  cos( margs->phase ));
	return new mapwargs_t({SLV::map_cavity,margs});
}

mapwargs_t*  map_chicanec__(environment* kwargs) {
	accept_kwargs(kwargs,{"alpha","lD","lB","energy","mode","verbose","name"});
	char* name = NULL;
	bool verbose = 0;
	typedef SLV::map_chicanec_args arg_t;
	arg_t* margs = (arg_t*) malloc(sizeof(arg_t));
	*margs = def::map_chicanec_args;
	for(auto& arg : *kwargs) {
		std::string kw = arg.first;
		var* v = arg.second;
		     if ( kw == "alpha")   margs->alpha  = v->fval[0];
		else if ( kw == "lD")      margs->lD     = v->fval[0];
		else if ( kw == "lB")      margs->lB     = v->fval[0];
		else if ( kw == "energy")  margs->energy = v->fval[0];
		else if ( kw == "mode")    margs->mode   = v->fval[0];
		else if ( kw == "verbose") verbose       = v->fval[0];
		else if ( kw == "name")    name          = v->sval;
	}

	char pref[256]={0};
	if(name) sprintf(pref,"%s_",name);
	if(verbose) {
		fprintf(OUTFILE,"%smode:   %d\n",pref,margs->mode);
		fprintf(OUTFILE,"%salpha:  %e\n",pref,margs->alpha);
		fprintf(OUTFILE,"%slD:     %e\n",pref,margs->lD);
		fprintf(OUTFILE,"%slB:     %e\n",pref,margs->lB);
		fprintf(OUTFILE,"%sR56:    %e\n",pref,SLV::chicanec_R56(margs->alpha,margs->lB,margs->lD));
		fprintf(OUTFILE,"%senergy: %e\n",pref,margs->energy);
	}
	return new mapwargs_t({SLV::map_chicanec,margs});
}

mapwargs_t*  map_csr__(const SLV::psd<2>* psd, environment* kwargs) {
	accept_kwargs(kwargs,{"angle","s","ds","length","file","energy",
		"npad","fudge","debug","verbose","mode","filter",
		"filterwidth","large_dist_cutoff","transient"});

	double angle = 1;
	double length = 1;
	double s  = -1;
	double ds = -1;
	double energy = cnst::m_e_eV;
	int npad = 2;
	char* fname = NULL;
	double fudge = 1.0;
	char* debug = NULL;
	bool verbose = 0;
	char mode = 5;
	char filter = 0;
	double filterwidth = 1./10.;
	double large_dist_cutoff = 0;
	int transient = 1;

	for(auto& arg : *kwargs) {
		std::string kw = arg.first;
		var* v = arg.second;
		     if ( kw == "angle")   angle   = v->fval[0];
		else if ( kw == "length")  length  = v->fval[0];
		else if ( kw == "s")       s       = v->fval[0];
		else if ( kw == "ds")      ds      = v->fval[0];
		else if ( kw == "energy")  energy  = v->fval[0];
		else if ( kw == "npad")    npad    = v->fval[0];
		else if ( kw == "file")    fname   = v->sval;
		else if ( kw == "fudge")   fudge   = v->fval[0];
		else if ( kw == "debug")   debug   = v->sval;
		else if ( kw == "verbose") verbose = v->fval[0];
		else if ( kw == "mode")    mode    = v->fval[0];
		else if ( kw == "filter")  filter  = v->fval[0];
		else if ( kw == "filterwidth") filterwidth = v->fval[0];
		else if ( kw == "large_dist_cutoff") large_dist_cutoff = v->fval[0];
		else if ( kw == "transient") transient = v->fval[0];
	}
	// set s, ds defaults
	if(s<0)   s = length;
	if(ds<0) ds = length;

	double gamma = energy / cnst::m_e_eV;

	struct SLV::grid_t p = SLV::psd_projection(psd,2);
	const size_t npts = 1u<<p.nexp;
	double* sol = (double*) calloc(npts,sizeof(double));

	//solve_poisson_args args = {
	//	.T=(p.limits[1]-p.limits[0]),
	//	.npad = npad,
	//	.green_s = NULL,
	//	.green_f = green_f_csr,
	//	.fargs= &radius
	//};
	//solve_poisson(p.data,sol,npts,(void*) &args, prefix);

	solve_csr_args args = {
		.T=(p.limits[1]-p.limits[0]),
		.npad = npad,
		.length = length,
		.angle = angle,
		.s = s,
		.gamma = gamma,
		.mode = mode,
		.filter = filter,
		.filterwidth = filterwidth,
		.transient = transient,
		.large_dist_cutoff = large_dist_cutoff
	};

	if(mode>=0 && mode<=5) {
		solve_csr(p.data,sol,npts,(void*) &args, debug);
	} else if(mode==99) {
		solve_csr_old(p.data,sol,npts,(void*) &args, debug);
	} else {
		fprintf(stderr,"csr mode %d unknown\n",mode); exit(1);
	}

	double factor =
			// conv(rho,G) // 1/m^2
			//length                   // m
			psd->weight            // C
			/// (4.0*M_PI*cnst::eps0) // Vm/C
			/ (4.0*M_PI*cnst::eps0) // Vm/C
			/// gamma                 // 1
			* fudge
			* ds
		;
	typedef SLV::map_datakick_args arg_t;
	arg_t* margs = (arg_t*) malloc(sizeof(arg_t));
	*margs = {
		.data=sol,
		.limits=p.limits,
		.dim=2,
		.dim_q=0,
		.dim_p=1,
		.npts=npts,
		.factor=factor
	};

	if(verbose) {
		fprintf(OUTFILE,"length: %e\n",    length);
		fprintf(OUTFILE,"angle:  %e\n",    angle);
		fprintf(OUTFILE,"s:      %e\n",    s);
		fprintf(OUTFILE,"ds:     %e\n",    ds);
		fprintf(OUTFILE,"energy: %e\n",    energy);
		fprintf(OUTFILE,"gamma:  %e\n",    gamma);
		fprintf(OUTFILE,"factor: %e\n",    factor);
		fprintf(OUTFILE,"weight: %e\n",    psd->weight);
		fprintf(OUTFILE,"T:      %e\n",    args.T);
		fprintf(OUTFILE,"limits: %e %e\n", p.limits[0],p.limits[1]);
		fprintf(OUTFILE,"mode:   %d\n",    args.mode);
	}

	if(fname!=NULL) {
		FILE* fd = fopen(fname,"w");
		if(fd==NULL) { print_buffer_position(); perror(fname); exit(1); }
		fprintf(fd,"q rho field\n");
		for(size_t i=0; i<margs->npts; i++) {
			double q = margs->limits[0] + (margs->limits[1]-margs->limits[0])/margs->npts * i;
			fprintf(fd,"%e %e %e\n",q,p.data[i],sol[i]*factor/ds);
		}
		fclose(fd);
	}
	free(p.data);
	return new mapwargs_t({SLV::map_datakick,margs});
}


mapwargs_t*  map_spacecharge__(const SLV::psd<2>* psd, environment* kwargs) {
	accept_kwargs(kwargs,{"beamsize","length","file","energy","npad","fudge","debug","verbose","beamsize_factor"});

	double beamsize = def::spacecharge_beamsize;
	double length = def::spacecharge_length;
	double energy = cnst::m_e_eV;
	int npad = 2;
	char* fname = NULL;
	double fudge = 1.0;
	double beamsize_factor = 1.747;
	char* debug = NULL;
	bool verbose = 0;

	for(auto& arg : *kwargs) {
		std::string kw = arg.first;
		var* v = arg.second;
		     if ( kw == "beamsize") beamsize  = v->fval[0];
		else if ( kw == "length")   length    = v->fval[0];
		else if ( kw == "energy")   energy    = v->fval[0];
		else if ( kw == "npad")     npad      = v->fval[0];
		else if ( kw == "file")     fname     = v->sval;
		else if ( kw == "fudge")    fudge     = v->fval[0];
		else if ( kw == "beamsize_factor") beamsize_factor = v->fval[0];
		else if ( kw == "debug")    debug     = v->sval;
		else if ( kw == "verbose")  verbose   = v->fval[0];
	}

	double gamma = energy / cnst::m_e_eV;
	double beamradius = beamsize * beamsize_factor;

	struct SLV::grid_t p = SLV::psd_projection(psd,2);
	solve_poisson_args args = {
		.T=(p.limits[1]-p.limits[0])*gamma,
		.npad = npad,
		.green_s = NULL,
		.green_f = green_f_adisk,
		.fargs = &beamradius
	};
	const size_t npts = 1u<<p.nexp;
	double* sol = (double*) calloc(npts,sizeof(double));
	solve_poisson(p.data,sol,npts,(void*) &args, debug);

	double factor =
			// conv(rho,G) // 1/m^2
			length                   // m
			* psd->weight            // C
			/// (4.0*M_PI*cnst::eps0) // Vm/C
			/ (2.0*M_PI*cnst::eps0) // Vm/C // see mthesis
			/ gamma                  // 1
			* fudge
		;
	typedef SLV::map_datakick_args arg_t;
	arg_t* margs = (arg_t*) malloc(sizeof(arg_t));
	*margs = {
		.data=sol,
		.limits=p.limits,
		.dim=2,
		.dim_q=0,
		.dim_p=1,
		.npts=npts,
		.factor=factor
	};

	if(verbose) {
		fprintf(OUTFILE,"length: %e\n",    length);
		fprintf(OUTFILE,"energy:  %e\n",   energy);
		fprintf(OUTFILE,"gamma:  %e\n",    gamma);
		fprintf(OUTFILE,"factor: %e\n",    factor);
		fprintf(OUTFILE,"weight: %e\n",    psd->weight);
		fprintf(OUTFILE,"T:      %e\n",    args.T);
		fprintf(OUTFILE,"limits: %e %e\n", p.limits[0],p.limits[1]);
		fprintf(OUTFILE,"beamsize: %e\n",        beamsize);
		fprintf(OUTFILE,"beamsize_factor: %e\n", beamsize_factor);
		fprintf(OUTFILE,"beamradius: %e\n",      beamradius);
	}

	if(fname!=NULL) {
		FILE* fd = fopen(fname,"w");
		if(fd==NULL) { print_buffer_position(); perror(fname); exit(1); }
		fprintf(fd,"q rho field\n");
		for(size_t i=0; i<margs->npts; i++) {
			double q = margs->limits[0] + (margs->limits[1]-margs->limits[0])/margs->npts * i;
			fprintf(fd,"%e %e %e\n",q,p.data[i],sol[i]*factor/length);
		}
		fclose(fd);
	}
	free(p.data);
	return new mapwargs_t({SLV::map_datakick,margs});
}

mapwargs_t*  map_poisson__(const SLV::psd<2>* psd, char* fnc, environment* kwargs) {
	accept_kwargs(kwargs,{"npad","file"});
	int npad=2;
	char* fname = NULL;
	for(auto& arg : *kwargs) {
		std::string kw = arg.first;
		var* v = arg.second;
		     if ( kw == "npad") npad = v->fval[0];
		     else if ( kw == "file") fname = v->sval;
	}
	
	static char b[] = "q";
	char* varname = b;
	struct math_eval_args mathevalargs {
		.fstr = fnc,
		.names = &varname
	};
	struct SLV::grid_t p = SLV::psd_projection(psd,2);
	solve_poisson_args args = {
		.T=(p.limits[1]-p.limits[0]),
		.npad = npad,
		.green_s = eval_math<1>,
		.green_f = NULL,
		.fargs = &mathevalargs
	};
	const size_t npts = 1u<<p.nexp;
	double* sol = (double*) calloc(npts,sizeof(double));
	solve_poisson(p.data,sol,npts,(void*) &args);

	typedef SLV::map_datakick_args arg_t;
	arg_t* margs = (arg_t*) malloc(sizeof(arg_t));
	*margs = {
		.data=sol,
		.limits=p.limits,
		.dim=2,
		.dim_q=0,
		.dim_p=1,
		.npts=npts,
		.factor=1
	};

	if(fname!=NULL) {
		FILE* fd = fopen(fname,"w");
		if(fd==NULL) { print_buffer_position(); perror(fname); exit(1); }
		for(size_t i=0; i<margs->npts; i++) {
			double q = margs->limits[0] + (margs->limits[1]-margs->limits[0])/margs->npts * i;
			fprintf(fd,"%e %e %e\n",q,p.data[i],sol[i]);
		}
		fclose(fd);
	}
	free(p.data);
	return new mapwargs_t({SLV::map_datakick,margs});
}

mapwargs_t*  map_poisson1d__(const SLV::psd<2>* psd, environment* kwargs) {
	accept_kwargs(kwargs,{"factor","file"});
	char* fname = NULL;
	double factor = 1;
	for(auto& arg : *kwargs) {
		std::string kw = arg.first;
		var* v = arg.second;
		     if ( kw == "factor") factor = v->fval[0];
		     else if ( kw == "file") fname = v->sval;
	}

	struct SLV::grid_t p = SLV::psd_projection(psd,2);
	const size_t npts = 1u<<p.nexp;
	solve_poisson1d_args args = {
		.T=(p.limits[1]-p.limits[0]),
		.npad = 1,
	};
	double* sol = (double*) calloc(npts,sizeof(double));
	solve_poisson1d(p.data,sol,npts,(void*) &args);
	typedef SLV::map_datakick_args arg_t;
	arg_t* margs = (arg_t*) malloc(sizeof(arg_t));
	*margs = {
		.data=sol,
		.limits=p.limits,
		.dim=2,
		.dim_q=0,
		.dim_p=1,
		.npts=npts,
		.factor=factor
	};
	if(fname!=NULL) {
		FILE* fd = fopen(fname,"w");
		if(fd==NULL) { print_buffer_position(); perror(fname); exit(1); }
		fprintf(fd,"q rho field\n");
		for(size_t i=0; i<margs->npts; i++) {
			double q = margs->limits[0] + (margs->limits[1]-margs->limits[0])/margs->npts * i;
			fprintf(fd,"%e %e %e\n",q,p.data[i],sol[i]*factor);
		}
		fclose(fd);
	}
	free(p.data);
	return new mapwargs_t({SLV::map_datakick,margs});
}

mapwargs_t*  map_analytic__(char* fq, char* fp) {
	static char b[16] = "t\0q\0p\0";
	typedef map_analytic_args arg_t;
	arg_t* margs = (arg_t*) malloc(sizeof(arg_t));
	margs->dim = 2;
	margs->fstrs[0] = strdup(fq); // TODO: This never gets freed
	margs->fstrs[1] = strdup(fp);
	margs->names[0] = b;
	margs->names[1] = b+2;
	margs->names[2] = b+4;
	return new mapwargs_t({map_analytic,margs});
}


mapwargs_t*  map_identity__() {
	return new mapwargs_t({SLV::map_identity,NULL});
}


/* ----------- Perron-Frobenius Step -------- */

SLV::psd<2>*  propagate__(const SLV::psd<2>* psd, mapwargs_t* map, environment* kwargs) {
	accept_kwargs(kwargs,{"center","box","nexp","depth","limits","pre","t","epsilon"});
	double t = 1;
	//SLV::psd_step_args args = {0,0,SLV::AUTO,SLV::AUTO,NULL,2,1e-12};
	SLV::psd_step_args args = {
		.nexp = 0,
		.depth = 0,
		.center_strat = SLV::AUTO,
		.box_strat = SLV::AUTO,
		.limits = NULL,
		.pre = 2,
		.epsilon = 1e-12
	};
	for(auto& arg : *kwargs) {
		std::string kw = arg.first;
		var* v = arg.second;
		     if ( kw == "center") {
		     	     if(strcmp(v->sval,"KEEP")==0) args.center_strat = SLV::KEEP;
		     	else if(strcmp(v->sval,"AUTO")==0) args.center_strat = SLV::AUTO;
			else {
				print_buffer_position();
				error(1,0,"%s unknown",v->sval);
			}
		     }
		     if ( kw == "box") {
		     	     if(strcmp(v->sval,"KEEP" )==0) args.box_strat = SLV::KEEP;
		     	else if(strcmp(v->sval,"EQUAL")==0) args.box_strat = SLV::EQUAL;
		     	else if(strcmp(v->sval,"AUTO" )==0) args.box_strat = SLV::AUTO;
		     	else if(strcmp(v->sval,"AREA" )==0) args.box_strat = SLV::AREA;
			else {
				print_buffer_position();
				error(1,0,"%s unknown",v->sval);
			}
		     }
		     if ( kw == "nexp") {
		     	     args.nexp = v->fval[0];
		     }
		     if ( kw == "depth") {
		     	     args.depth = v->fval[0];
		     }
		     if ( kw == "limits") {
		     	     args.limits = v->fval;
		     }
		     if ( kw == "pre") {
		     	     args.pre = v->fval[0];
		     }
		     if ( kw == "epsilon") {
		     	     args.epsilon = v->fval[0];
		     }
		     if ( kw == "t") {
		     	     t = v->fval[0];
		     }
	}
	return SLV::psd_step(psd,map->map,map->args,t,args);
}


/* ------------- Other Functions ------------ */

void show__(SLV::psd<2>* psd, environment* kwargs) {
	accept_kwargs(kwargs,{"file","cells","negative","limits","grid","contours","rectangle","projection","points","title"});
	#ifdef HAVE_X11
	xwindow vis;
	vis.fun = SLV::psd_eval<2>;
	vis.fargs=psd;
	vis.psd = psd;
	vis.root=psd->root;
	vis.center();
	char* fname=NULL;
	for(auto& arg : *kwargs) {
		std::string kw = arg.first;
		var* v = arg.second;
		if ( kw == "file") fname = v->sval;
		if ( kw == "title") vis.set_name(v->sval);
		if ( kw == "cells") {
			if(v->fval[0] > 0) {
				vis.drawmask |= vis.DRAWCELLS;
			} else {
				vis.drawmask &= ~vis.DRAWCELLS;
			}
		}
		if ( kw == "negative") {
			if(v->fval[0] > 0) {
				vis.drawmask |= vis.DRAWNEGATIVE;
			} else {
				vis.drawmask &= ~vis.DRAWNEGATIVE;
			}
		}
		if ( kw == "grid") {
			if(v->fval[0] > 0) {
				vis.drawmask |= vis.DRAWGRID;
			} else {
				vis.drawmask &= ~vis.DRAWGRID;
			}
		}
		if ( kw == "contours") {
			if(v->fval[0] > 0) {
				vis.drawmask |= vis.DRAWCONTOURS;
			} else {
				vis.drawmask &= ~vis.DRAWCONTOURS;
			}
		}
		if ( kw == "limits") {
			memcpy(vis.limits,v->fval,4*sizeof(double));
		}
		if ( kw == "rectangle") {
			memcpy(vis.rectangle,v->fval,4*sizeof(double));
			vis.drawmask |= vis.DRAWRECTANGLE;
		}
		if ( kw == "projection") {
			if(v->fval[0] > 0) {
				vis.drawmask |= vis.DRAWPROJECTION;
			} else {
				vis.drawmask &= ~vis.DRAWPROJECTION;
			}
		}
		if ( kw == "points") {
			if(v->fval[0] > 0) {
				vis.drawmask |= vis.DRAWPOINTS;
			} else {
				vis.drawmask &= ~vis.DRAWPOINTS;
			}
		}
	}
	if(fname!=NULL) {
		FILE* fd = fopen(fname,"w");
		if(fd==NULL) { print_buffer_position(); perror(fname); exit(1); }
		vis.update();
		vis.screenshot(fd);
		fclose(fd);
	} else {
		//if(fork()) return;
		vis.loop();
		//_exit(0);
	}
	#else
	fprintf(stderr,"Visualization not available. Recompile with X11 support.\n");
	#endif
}

void plot__(SLV::psd<2>* psd, const char* fname, environment* kwargs) {
	FILE* fd = fopen(fname,"w");
	psd_plot(psd,fd);
	fclose(fd);
}


void write_projection__(SLV::psd<2>* psd, const char* fname, unsigned int dims, environment* kwargs) {
	accept_kwargs(kwargs,{"method"});
	int method = 0;
	for(auto& arg : *kwargs) {
		std::string kw = arg.first;
		var* v = arg.second;
		     if ( kw == "method") method = v->fval[0];
	}
	SLV::grid_t p = SLV::psd_projection(psd, dims, method);
	FILE* fd = fopen(fname,"w");
	if(fd==NULL) { print_buffer_position(); perror(fname); exit(1); }
	int Nd = 1;
	int npts = 1<<(p.dim*p.nexp);

	double coords[p.dim * 1<<(p.nexp*p.dim)];
	for(int d=0; d<p.dim; d++) {
		for(int n=0; n < npts; n++) {
			int i = n*p.dim+d;
			coords[i] = ( (n/Nd) % (1<<p.nexp) ) * (1.0 / (1<<p.nexp));
			coords[i] = p.limits[d] + coords[i] * (p.limits[d+p.dim]-p.limits[d]);
		}
		Nd *= 1<<p.nexp;
	}

	// write header
	for(int d=0; d<p.dim; d++) {
		fprintf(fd,"z%d ",d);
	}
	fprintf(fd,"proj\n");

	// write data
	for(int i=0; i<npts; i++) {
		for(int d=0; d<p.dim; d++) {
			fprintf(fd,"%e ",coords[i*p.dim+d]);
		}
		fprintf(fd,"%0.16e\n",p.data[i]);
	}
	fclose(fd);
	free(p.data);
	free(p.limits);
}

void write_localmoments__(SLV::psd<2>* psd, const char* fname, unsigned int axis) {
	const int N=3;
	SLV::grid_t moments[N];
	double exps[2] = {0,0};
	struct monomial_args args = {2,exps};
	for(int n=0; n<N; n++) {
		exps[axis] = n;
		moments[n] = SLV::psd_projection(psd, 1<<axis, 0, monomial, &args);
	}

	FILE* fd = fopen(fname,"w");
	if(fd==NULL) { print_buffer_position(); perror(fname); exit(1); }
	const size_t npts = 1<<(moments[N-1].nexp);

	fprintf(fd,"z rho mu std\n"); // write header
	for(size_t i=0; i<npts; i++) {
		double q = moments[0].limits[0] + i*(moments[0].limits[1]-moments[0].limits[0])/npts;
		double   u = moments[0].data[i];
		double  mu = u<=0 ? 0.0 : moments[1].data[i]/u;
		double var = u<=0 ? 0.0 : moments[2].data[i]/u - mu*mu;
		fprintf(fd,"%0.16e ",q);
		fprintf(fd,"%0.16e ",u);
		fprintf(fd,"%0.16e ",mu);
		fprintf(fd,"%0.16e",sqrt(var));
		fprintf(fd,"\n");
	}
	fclose(fd);

	for(int n=0; n<N; n++) {
		free(moments[n].data);
		free(moments[n].limits);
	}
}

void write_grid__(SLV::psd<2>* psd, const char* fname , environment* kwargs) {
	accept_kwargs(kwargs,{"npts","limits"});
	int npts[2] = {128,128};
	double* limits = NULL;
	for(auto& arg : *kwargs) {
		std::string kw = arg.first;
		var* v = arg.second;
		if( kw == "npts") {
			npts[0] = v->fval[0];
			npts[1] = v->fval[1];
		}
		if( kw == "limits") limits = v->fval;
	}
	FILE* fd = fopen(fname,"w");
	if(fd==NULL) { print_buffer_position(); perror(fname); exit(1); };
	psd_write_grid(psd,fd,limits,npts);
	fclose(fd);
}

void write_ensemble__(SLV::psd<2>* psd, const char* fname , environment* kwargs) {
	accept_kwargs(kwargs,{"n","form","seed","ref"});
	int n = 10*1000;
	int form = 0;
	int seed = 1337;
	double ref[2] = {0};
	for(auto& arg : *kwargs) {
		std::string kw = arg.first;
		var* v = arg.second;
		if( kw == "n")       n = v->fval[0];
		if( kw == "form") form = v->fval[0];
		if( kw == "seed") seed = v->fval[0];
		if( kw == "ref") { // reference
			if(v->size!=2) { print_buffer_position(); fprintf(stderr,"Error: size of 'ref' need to be 2\n"); exit(1); };
			ref[0] = v->fval[0];
			ref[1] = v->fval[1];
		}
	}
	FILE* fd = fopen(fname,"w");
	if(fd==NULL) { print_buffer_position(); perror(fname); exit(1); };
	switch(form) {
		case 0:
			SLV::psd_ensemble(psd,fd,n,seed,ref);
			break;
		case 1:
			SLV::psd_ensemble_astra(psd,fd,n,seed,ref);
			break;
	}
	fclose(fd);
}

void modify__(SLV::psd<2>* psd, char* fstr) {
	char b[16] = "psi\0q\0p\0";
	char* names[3] = {b,b+4,b+6};
	struct math_eval_args args = {fstr,names};
	psd->modify(eval_mathf<2>,&args);
}

void noise__(SLV::psd<2>* psd, double a, environment* kwargs) {
	accept_kwargs(kwargs,{"type","seed"});
	int type = 0;
	int seed = 0;
	for(auto& arg : *kwargs) {
		std::string kw = arg.first;
		var* v = arg.second;
		if( kw == "seed") seed = v->fval[0];
		if( kw == "type") type = v->fval[0];
	}
	switch(type) {
	case 0:
		SLV::noise_uniform(psd->root, a);
		break;
	case 1:
		SLV::noise_poisson(psd->root, a, seed);
		break;
	default:
		print_buffer_position();
		error(1, 0, "Error: Noise type %d unknown (possible values: 0,1)\n",type);
		break;
	}
}

double integral_function__(SLV::psd<2>* psd, char* fstr) {
	char b[16] = "psi\0q\0p\0";
	char* names[3] = {b,b+4,b+6};
	struct math_eval_args args = {fstr,names};
	return psd->integral(eval_mathf<2>,&args);
}

FILE* fdstack[128];
FILE** fdstack_ptr = fdstack;

void fdstack_push(FILE* fd) {
	fdstack_ptr++;
	*fdstack_ptr = fd;
	OUTFILE = *fdstack_ptr;
}
void fdstack_pop() {
	fclose(OUTFILE);
	fdstack_ptr--;
	OUTFILE = *fdstack_ptr;
}


%}

%code requires {
	#include <map>
	#include <string>
	#include <var.h>
	#include <psd.h>
}

%union {
	char* sval;
	double fval;
	environment* envval;
	std::vector<double>* fvals;
	varlist* varvals;
	var* varval;
}

/* TOKENS */
%token WHO FORMAT STRCAT
%nonassoc GTGT GT LT GE LE EQ NE;
%left PLUS MINUS;
%left TIMES DIVIDE COLONEQUAL;
%left POW;
%left UMINUS UPLUS;
%right SIN COS ASIN ACOS TAN ATAN2 SQRT FLOOR CEIL ABS;
%left RBRACKET LBRACKET
%token<fval> FLOAT;
%token<sval> STRING SYMBOL BLOCK;
%type<envval> kwargs;
%type<varval> var var_psd var_map var_flt var_str kwarg exp map_call psd_call array string;
%type<fvals> vars_flt;
%type<varvals> vars vars_map;

%token SHOW PLOT EVAL INCLUDE PRINT SAVE LOAD COPY STATS CENTROID VARIANCE MAXIMUM REFERENCE;
%token MAP_COMPOSE MAP_TRANSLATE MAP_ROTATE MAP_KICKL MAP_IDENTITY MAP_HYPERBOLIC;
%token MAP_CAVITY MAP_DRIFTL MAP_CHICANEC MAP_SPACECHARGE MAP_CSR MAP_ANALYTIC MAP_DRIFTPOLY MAP_KICKPOLY MAP_DRIFTSINE MAP_KICKSINE;
%token MAP_POISSON MAP_POISSON1D;
%token PSD_ANALYTIC PSD_GAUSS PSD_ENSEMBLE PSD_RECTANGLE PSD_TEST PSD_BUMP PROPAGATE MULTIPLY MODIFY NOISE NORMALIZE;
%token WRITE_PROJECTION WRITE_LOCALMOMENTS WRITE_GRID WRITE_ENSEMBLE INTEGRAL CHARGE;
%token DO FOR IN IF ELSE DEFINED;
%token MKDIR CHDIR GETCWD GETENV STRTOD STRCMP BASENAME;
%token LINSPACE AVERAGE;
%token TRANSFER4D CHICANECOEF;

%token ACCEPT

%%

/* Input loop */
input:
	| input command
;

/* Accepted commands */
command: assignment ';'
	| ACCEPT { YYACCEPT; }
	//| var ';' {}
	| var ';' { $1->print(OUTFILE); fprintf(OUTFILE,"\n");  $1->unref(); }
	//| BLOCK ';' { parse_string__($1); free($1); }
	| BLOCK GT var_str ';' { // >
		FILE* fd = fopen($3->sval,"w");
		if(fd==NULL) { print_buffer_position(); perror($3->sval); exit(1); };
		fdstack_push(fd);
		parse_string__($1);
		fdstack_pop();
		free($1);
		$3->unref();
	}
	| BLOCK GTGT var_str ';' { // >>
		FILE* fd = fopen($3->sval,"a");
		if(fd==NULL) { print_buffer_position(); perror($3->sval); exit(1); };
		fdstack_push(fd);
		parse_string__($1);
		fdstack_pop();
		free($1);
		$3->unref();
	}
	| DO '(' var_flt ')' BLOCK ';' {
		for(int n=0; n<floor($3->fval[0]); n++) {
			parse_string__($5);
		}
	}
	| FOR SYMBOL IN var_flt BLOCK ';' {
		if(ENV.count($2)!=0) { /* Variable exists */
			ENV[$2]->unref();
		}
		ENV[$2] = new var(0.0,1);
		for(size_t n=0; n<$4->size; n++) {
			ENV[$2]->fval[0] = $4->fval[n];
			parse_string__($5);
		}
		free($2);
		free($5);
		$4->unref();
	}
	| IF '(' var_flt ')' BLOCK ';' { if($3->fval[0]>0) parse_string__($5); }
	| IF '(' var_flt ')' BLOCK ELSE BLOCK ';' { if($3->fval[0]>0) parse_string__($5); else parse_string__($7);}
	|    SHOW '(' var_psd kwargs ')' ';' { show__($3->psdval,$4); $3->unref(); delete $4; }
	|    PLOT '(' var_psd ',' var_str kwargs ')' ';' { plot__($3->psdval,$5->sval,$6); $3->unref(); $5->unref(); delete $6; }
	|    EVAL '(' var_str ')' ';' { parse_string__($3->sval); $3->unref(); }
	| INCLUDE '(' var_str ')' ';' { parse_file__($3->sval); $3->unref(); }
	| WHO '(' ')' ';' {
		for(auto v : ENV) {
			switch(v.second->type) {
			case(STR):
				//fprintf(OUTFILE,"%s (STR) = ", v.first.c_str());
				fprintf(OUTFILE,"%s = ", v.first.c_str());
				break;
			case(FLT):
				//fprintf(OUTFILE,"%s (FLT[%ld]) = ", v.first.c_str(),v.second->size);
				fprintf(OUTFILE,"%s = ", v.first.c_str());
				break;
			case(MAP):
				//fprintf(OUTFILE,"%s (MAP)",    v.first.c_str());
				fprintf(OUTFILE,"%s ",    v.first.c_str());
				break;
			case(PSD):
				//fprintf(OUTFILE,"%s (PSD) = ", v.first.c_str());
				fprintf(OUTFILE,"%s = ", v.first.c_str());
				break;
			}
			v.second->print(OUTFILE);
			fprintf(OUTFILE,";\n");
		}
	}
	| WRITE_PROJECTION '(' var_psd ',' var_str ',' var_flt kwargs ')' ';' {
		write_projection__($3->psdval,$5->sval,*($7->fval),$8);
		$3->unref(); $5->unref(); $7->unref();
		delete $8;
	}
	| WRITE_LOCALMOMENTS '(' var_psd ',' var_str ',' var_flt ')' ';' {
		write_localmoments__($3->psdval,$5->sval,*($7->fval));
		$3->unref(); $5->unref(); $7->unref();
	}
	| WRITE_GRID '(' var_psd ',' var_str kwargs')' ';'  {
		write_grid__($3->psdval,$5->sval,$6);
		$3->unref(); $5->unref();
		delete $6;
	}
	| WRITE_ENSEMBLE '(' var_psd ',' var_str kwargs ')' ';'  {
		write_ensemble__($3->psdval, $5->sval, $6);
		$3->unref(); $5->unref();
		delete $6;
	}
	| MULTIPLY '(' var_psd ',' var_flt ')' ';'{
		$3->psdval->multiply($5->fval[0]);
		$3->unref(); $5->unref();
	}
	| MODIFY '(' var_psd ',' var_str ')' ';'{
		modify__($3->psdval, $5->sval);
		$3->unref(); $5->unref();
	}
	| NOISE '(' var_psd ',' var_flt kwargs ')' ';'{
		noise__($3->psdval, $5->fval[0], $6);
		$3->unref(); $5->unref();
		delete $6;
	}
	| NORMALIZE '(' var_psd ')' ';' {
		$3->psdval->normalize();
		$3->unref();
	}
	| SAVE '(' var_psd ',' var_str ')' ';' {
		const char* fname = $5->sval;
		FILE* fd = fopen(fname,"w");
		if(fd==NULL) { print_buffer_position(); perror(fname); exit(1); };
		SLV::save($3->psdval,fd);
		fclose(fd);
		$3->unref(); $5->unref();
	}
	| STATS '(' var_psd ')' ';' {
		SLV::psd_stats($3->psdval,OUTFILE);
		$3->unref();
	}
	| PRINT '(' vars ')' ';'    {
		for(auto v : *$3) {
			if(v->type==STR) {
				fprintf(OUTFILE,"%s",v->sval);
			}
			else if(v->type==FLT) {
				for(size_t i=0; i<v->size; i++) {
					fprintf(OUTFILE,"%.16e",v->fval[i]);
					if(i<v->size-1) fprintf(OUTFILE," ");
				}
			} else {
				v->print(OUTFILE);
			}
			if(v != $3->back()) {
				fprintf(OUTFILE," ");
			}
			v->unref();
		}
		fprintf(OUTFILE,"\n");
		delete $3;
	}
	| MKDIR '(' var_str ')' ';' {
		int status;
		status = mkdir($3->sval,0755);
		if(status && errno != EEXIST) { print_buffer_position(); perror($3->sval); exit(1); };
		$3->unref();
	}
	| CHDIR '(' var_str ')' ';' {
		int status;
		status = chdir($3->sval);
		if(status) { print_buffer_position(); perror($3->sval); exit(1); };
		$3->unref();
	}
;

/* Functions returning PSDs */
psd_call: PSD_ANALYTIC '(' var_str kwargs ')' {
		$$ = new var(psd_analytic__($3->sval,$4));
		$3->unref(); delete $4;
		}
	| PSD_GAUSS '(' kwargs ')' {
		$$ = new var(psd_gauss__($3));
		delete $3;
		}
	| PSD_ENSEMBLE '(' var_str  kwargs ')' {
		$$ = new var(psd_ensemble__($3->sval,$4));
		$3->unref();
		delete $4;
		}
	| PSD_RECTANGLE '(' var_flt  kwargs ')' {
		$$ = new var(psd_rectangle__($3->fval,$4));
		//$3->unref();
		delete $4;
		}
	| PSD_TEST '(' kwargs ')' {
		$$ = new var(psd_test__($3));
		delete $3;
		}
	| PSD_BUMP '(' kwargs ')' {
		$$ = new var(psd_bump__($3));
		delete $3;
		}
	| PROPAGATE '(' var_psd ',' var_map kwargs ')' {
		$$ = new var(propagate__($3->psdval,$5->mapval,$6));
		$3->unref(); $5->unref();
		delete $6;
		}
	| LOAD '(' var_str ')' {
		const char* fname = $3->sval;
		FILE* fd = fopen(fname,"r");
		if(fd==NULL) { print_buffer_position(); perror(fname); exit(1); };
		$$ = new var(SLV::load<2>(fd));
		fclose(fd);
		$3->unref();
		}
	| COPY '(' var_psd ')' {
		$$ = new var($3->psdval->copy());
		$3->unref();
	}
;

/* Functions returning Maps */
map_call: MAP_TRANSLATE    '(' var_flt ')' { $$ = new var(map_translate__($3->fval)); $3->unref(); }
	| MAP_ROTATE       '(' var_flt ')' { $$ = new var(map_rotate__($3->fval[0])); delete $3; }
	| MAP_KICKL        '(' var_flt ')' { $$ = new var(map_kickl__($3->fval[0])); $3->unref(); }
	| MAP_DRIFTL       '(' var_flt ')' { $$ = new var(map_driftl__($3->fval[0]));$3->unref(); }
	| MAP_HYPERBOLIC   '(' var_flt ')' { $$ = new var(map_hyperbolic__($3->fval[0]));$3->unref(); }
	| MAP_KICKPOLY     '(' var_flt ')' { $$ = new var(map_kickpoly__($3->size,$3->fval)); $3->unref(); }
	| MAP_DRIFTPOLY    '(' var_flt ')' { $$ = new var(map_driftpoly__($3->size,$3->fval));$3->unref(); }
	| MAP_KICKSINE     '(' var_flt ',' var_flt ',' var_flt ')' {
		$$ = new var(map_kicksine__($3->fval[0],$5->fval[0],$7->fval[0]));
		$3->unref(); $5->unref(); $7->unref();
		}
	| MAP_DRIFTSINE    '(' var_flt ',' var_flt ',' var_flt ')' {
		$$ = new var(map_driftsine__($3->fval[0],$5->fval[0],$7->fval[0]));
		$3->unref(); $5->unref(); $7->unref();
		}
	| MAP_CAVITY       '(' kwargs ')'  { $$ = new var(map_cavity__($3)); delete $3; }
	| MAP_CHICANEC     '(' kwargs ')'  { $$ = new var(map_chicanec__($3)); delete $3; }
	| MAP_SPACECHARGE  '(' var_psd kwargs ')' { $$ = new var(map_spacecharge__($3->psdval,$4)); $3->unref(); delete $4; }
	| MAP_CSR          '(' var_psd kwargs ')' { $$ = new var(map_csr__($3->psdval,$4)); $3->unref(); delete $4; }
	| MAP_POISSON      '(' var_psd ',' var_str kwargs ')' { $$ = new var(map_poisson__($3->psdval,$5->sval,$6)); $3->unref(); $5->unref(); delete $6; }
	| MAP_POISSON1D    '(' var_psd kwargs ')' { $$ = new var(map_poisson1d__($3->psdval,$4)); $3->unref(); delete $4; }
	| MAP_COMPOSE      '(' vars_map ')' { $$ = new var(map_compose__($3)); delete $3; }
	| MAP_ANALYTIC     '(' var_str ',' var_str ')' { $$ = new var(map_analytic__($3->sval,$5->sval)); $3->unref(); $5->unref(); }
	| MAP_IDENTITY     '(' ')' { $$ = new var(map_identity__()); }
;

/* Variable assignment */
assignment: SYMBOL '=' var {
		if(ENV.count($1)!=0) { /* Variable exists */
			ENV[$1]->unref();
		}
		ENV[$1] = $3;
		free($1);
	}
	| SYMBOL COLONEQUAL var { // Default assignment
		if(ENV.count($1)==0) { /* Variable does not exists */
			ENV[$1] = $3;
		}
		free($1);
	}
;

/* Type-safe variable access */
var_flt: var { if( true != $1->assert_type(FLT)) { print_buffer_position(); error(1,0,"Error: variable used wrong context\n"); }; }
var_str: var { if( true != $1->assert_type(STR)) { print_buffer_position(); error(1,0,"Error: variable used wrong context\n"); }; }
var_psd: var { if( true != $1->assert_type(PSD)) { print_buffer_position(); error(1,0,"Error: variable used wrong context\n"); }; }
var_map: var { if( true != $1->assert_type(MAP)) { print_buffer_position(); error(1,0,"Error: variable used wrong context\n"); }; }


//var: var_foo { printf("%p %d\n",$1->ptr,$1->refcount) ;}

/* General variables */
var: FLOAT { $$ = new var($1); }
	| STRING { $$ = new var($1); free($1); }
	| BLOCK  { $$ = new var($1); free($1); }
	| SYMBOL {
		if(ENV.count($1)<=0) {
			print_buffer_position();
			error(1,0,"Error: Symbol '%s' not defined\n",$1);
		}
		$$ = ENV.at($1)->ref();
		free($1);
		}
	| DEFINED '(' SYMBOL ')' { $$ = new var(ENV.count($3)>0 ? 1 : 0); free($3); }
	| exp
	| string
	| psd_call
	| map_call
	| array
	| INTEGRAL  '(' var_psd ')' { $$ = new var($3->psdval->integral()); $3->unref(); }
	| INTEGRAL  '(' var_psd ',' var_str ')' { $$ = new var(integral_function__($3->psdval,$5->sval)); $3->unref(); $5->unref(); }
	| CHARGE    '(' var_psd ')' { $$ = new var($3->psdval->weight);     $3->unref(); }
	| CENTROID  '(' var_psd ')' { $$ = new var(0.0,2);  SLV::centroid($3->psdval,$$->fval); $3->unref(); }
	| VARIANCE  '(' var_psd ')' { $$ = new var(0.0,4);  SLV::variance($3->psdval,$$->fval); $3->unref(); }
	| MAXIMUM   '(' var_psd ')' { $$ = new var($3->psdval->max()); $3->unref(); }
	| REFERENCE '(' var_psd ')' { $$ = new var(&$3->psdval->refpoint[0],2); $3->unref(); }
	| LINSPACE  '(' var_flt ',' var_flt ',' var_flt ')' {
		double a = *$3->fval;
		double b = *$5->fval;
		int n = *$7->fval;
		$$ = new var(0.0,n);
		for(int i=0; i<n; i++) {
			$$->fval[i] = a + (b-a)/(n-1) * i;
		}
		$3->unref(); $5->unref(); $7->unref();
	}
	| CHICANECOEF '(' var_flt ',' var_flt ',' var_flt ',' var_flt ')' {
		int    n     = $3->fval[0];
		double angle = $5->fval[0];
		double lB    = $7->fval[0];
		double lD    = $9->fval[0];
		$$ = new var(chicanecoef(n,angle,lB,lD));
		$3->unref(); $5->unref(); $7->unref(); $9->unref();
	}
	| TRANSFER4D '(' var_flt ',' var_flt ',' var_flt ',' var_flt ',' var_flt ')' {
		double* R0    = $3->fval;
		double l      = $5->fval[0];
		double fangle = $7->fval[0];
		double k0     = $9->fval[0];
		double k1     = $11->fval[0];
		$$ = new var(0.0, 16);
		transfer4d($$->fval, R0, l, fangle, k0, k1);
		$3->unref(); $5->unref(); $7->unref(); $9->unref(); $11->unref();
	}
	| AVERAGE '(' var_flt ',' var_flt ',' var_flt ',' var_flt ')' {
		double a   = $3->fval[0];
		double b   = $5->fval[0];
		double* xp = $7->fval;
		double* yp = $9->fval;
		$$ = new var(average(a, b, xp, yp, $7->size));
		$3->unref(); $5->unref(); $7->unref(); $9->unref();
	}
	| STRCMP '(' var_str ',' var_str ')' {
		$$ = new var(1.0*strcmp($3->sval, $5->sval));
		$3->unref(); $5->unref();
	}
;

string:  FORMAT '(' var_str ',' vars_flt ')' {
		char buf[512];
		double args[8];
		if($5->size()>8) {
			print_buffer_position();
			error(1,0,"Error: format: Too many arguments (>8).\n");
		}
		for(size_t i=0; i<$5->size(); i++) {
			args[i]=(*$5)[i];
		}
		sprintf(buf,$3->sval,
			args[0], args[1], args[2], args[3],
			args[4], args[5], args[6], args[7]
			);
		$$ = new var(buf);
		$3->unref(); delete $5;
	}
	| STRCAT '(' var_str ',' var_str ')' {
		int len=strlen($3->sval) + strlen($5->sval) + 1;
		char buf[len];
		strcpy(buf,$3->sval);
		strcat(buf,$5->sval);
		$$ = new var(buf);
		$3->unref();
		$5->unref();
	} 
	| GETCWD '(' ')' {
		char* dir =  get_current_dir_name();
		if(NULL==dir) { print_buffer_position(); perror("getcwd"); exit(1); };
		$$ = new var(dir);
	}
	| GETENV '(' var_str ')' {
		const char* s = getenv($3->sval);
		if(NULL==s) {
			print_buffer_position();
			fprintf(stderr,"getenv: Environment variable %s not defined\n",$3->sval);
			exit(1);
		};
		$$ = new var(s);
		$3->unref();
	}
	| STRTOD '(' var_str ')' {
		errno = 0;
		double x = strtod($3->sval,NULL);
		if(errno!=0) {
			print_buffer_position();
			fprintf(stderr,"strtod: %s conversion failed\n",$3->sval);
			exit(1);
		}
		$$ = new var(x);
		$3->unref();
	}
;
/* Double array */
array: LBRACKET vars_flt RBRACKET { $$ = new var($2); delete $2; }
;
vars_flt: var_flt {
		$$ = new std::vector<double>;
		$$->push_back($1->fval[0]);
		$1->unref();
	}
	| vars_flt ',' var_flt {
		$1->push_back($3->fval[0]);
		$$ = $1;
		$3->unref();
	}
;

/* List of general variables */
vars: var { $$ = new varlist; $$->push_back($1); }
	| vars ',' var { $1->push_back($3); $$ = $1;}

/* List of map variables */
vars_map: var_map { $$ = new varlist; $$->push_back($1); }
	| vars_map ',' var_map { $1->push_back($3); $$ = $1;}


/* Double arithmetic */
exp:'(' var_flt ')' {$$ = $2;}
	| var_flt PLUS   var_flt     { $$ = new var(*$1 + *$3);          $1->unref(); $3->unref();}
	| var_flt MINUS  var_flt     { $$ = new var(*$1 - *$3);          $1->unref(); $3->unref();}
	| MINUS var_flt %prec UMINUS { $$ = new var(-1.0 * *$2);         $2->unref(); }
	| PLUS  var_flt %prec UPLUS  { $$ = new var(*$2);                $2->unref(); }
	| var_flt TIMES  var_flt     { $$ = new var(*$1 * *$3);          $1->unref(); $3->unref();}
	| var_flt POW    var_flt     { $$ = new var(pow($1->fval[0],$3->fval[0]));
	                                                                 $1->unref(); $3->unref();}
	| var_flt DIVIDE var_flt     { $$ = new var(*$1 / *$3);          $1->unref(); $3->unref();}
	| SIN  '(' var_flt ')'       { $$ = new var(sin($3->fval[0]));   $3->unref(); }
	| COS  '(' var_flt ')'       { $$ = new var(cos($3->fval[0]));   $3->unref(); }
	| ASIN  '(' var_flt ')'      { $$ = new var(asin($3->fval[0]));  $3->unref(); }
	| ACOS  '(' var_flt ')'      { $$ = new var(acos($3->fval[0]));  $3->unref(); }
	| TAN  '(' var_flt ')'       { $$ = new var(tan($3->fval[0]));   $3->unref(); }
	| ATAN2 '(' var_flt ',' var_flt ')'
	                             { $$ = new var(atan2($3->fval[0],$5->fval[0]));
				                                         $3->unref(); $5->unref();}
	| SQRT '(' var_flt ')'       { $$ = new var(sqrt($3->fval[0]));  $3->unref(); }
	| FLOOR '(' var_flt ')'      { $$ = new var(floor($3->fval[0])); $3->unref(); }
	| CEIL '(' var_flt ')'       { $$ = new var(ceil($3->fval[0]));  $3->unref(); }
	| ABS  '(' var_flt ')'       { $$ = new var(fabs($3->fval[0]));  $3->unref(); }
	| var_flt LBRACKET var_flt RBRACKET { 
		size_t i = $3->fval[0];
		if(i<0 || i>=$1->size) { print_buffer_position(); error(1,0,"Error: Index out of range.\n"); }
		$$ = new var($1->fval[i]);
		$1->unref(); $3->unref();
		}
	// FIXME: GT causes a shift/reduce conflict with the ``output redirection'' GT
	//| var_flt GT var_flt        { $$ = new var($1->fval[0] >  $3->fval[0] ?1:0); $1->unref(); $3->unref();}
	| var_flt LT var_flt        { $$ = new var($1->fval[0] <  $3->fval[0] ?1:0); $1->unref(); $3->unref();}
	| var_flt LE var_flt        { $$ = new var($1->fval[0] <= $3->fval[0] ?1:0); $1->unref(); $3->unref();}
	| var_flt GE var_flt        { $$ = new var($1->fval[0] >= $3->fval[0] ?1:0); $1->unref(); $3->unref();}
	| var_flt EQ var_flt        { $$ = new var($1->fval[0] == $3->fval[0] ?1:0); $1->unref(); $3->unref();}
	| var_flt NE var_flt        { $$ = new var($1->fval[0] != $3->fval[0] ?1:0); $1->unref(); $3->unref();}
;

/* Keyword arguments */
kwarg: SYMBOL '=' var { $$ = $3; $$->name = strdup($1); free($1); }
;
kwargs: { $$ = new environment; }
	| kwarg {
		auto a = new environment;
		(*a)[$1->name] = $1;
		$$ = a;
	}
	| kwargs ',' kwarg {
		(*$1)[$3->name] = $3;
	}
	| kwargs ',' /* allow trailing comma */
;


%%

int main(int argc, char* argv[]) {
	if(isatty(fileno(stdin))) fprintf(stdout,
		"selav1d  Copyright (C) 2020-2022  Philipp Amstutz (DESY)\n"
		"This program comes with ABSOLUTELY NO WARRANTY.\n"
		"This is free software, and you are welcome to redistribute it\n"
		"under certain conditions; see the file COPYING for details.\n"
		);
	fdstack_push(stdout);
	parse_file__("/dev/stdin");
	//yyparse();
	fdstack_pop();
	return 0;
}


extern int yylineno;
void yyerror(const char* s) {
	print_buffer_position();
	fprintf(stderr,"Error in line %d: %s\n",yylineno,s);
	exit(1);
}
