/*
* selav1d - semi-lagrangian simulation of sparse phase-space densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <cstddef>
#include <cstdlib>
#include <cstring>
#include <cassert>
#include <cstdio>
#include <error.h>

#include <var.h>
#include <psd.h>



var::var(double x, int n) : type(FLT), size(n) {
	fval = (double*) calloc(size,sizeof(double));
	for(int i=0; i<n; i++) {
		fval[i] = x;
	}
}

var::var(double* x, int n) : type(FLT), size(n) {
	fval = (double*) calloc(size,sizeof(double));
	memcpy(fval,x,n*sizeof(double));
}

var::var(std::vector<double>* v) : type(FLT), size(v->size()) {
	fval = (double*) calloc(size,sizeof(double));
	for(size_t i=0; i<size; i++) fval[i] = (*v)[i];
}

var::var(const char* s) : type(STR) {
	sval = strdup(s);
}

var::var(SLV::psd<2>* psd) : type(PSD) {
	psdval = psd;
}
var::var(mapwargs_t* map) : type(MAP) {
	mapval = map;
}

var::var(const var& v) : type(v.type), size(v.size), ptr(v.ptr){
	if(v.name!=NULL) name = strdup(v.name);
	switch(v.type) {
	case STR:
		sval = strdup(v.sval);
		break;
	case FLT:
		fval = (double*) calloc(v.size,sizeof(double));
		memcpy(fval,v.fval,size*sizeof(double));
		break;
	default:
		break;
	}
}


void var::print(FILE* fd) const {
	//if(name!=NULL) fprintf(fd,"%s",name);
	switch(type) {
	case FLT:
		if(size>1) {
			fprintf(fd,"[");
			for(size_t i=0; i<size; i++) {
				fprintf(fd,"%e",fval[i]);
				if(i<size-1) fprintf(fd,", ");
			}
			fprintf(fd,"]");
		} else {
			fprintf(fd,"%0.16e",fval[0]);
		}
		break;
	case STR:
		fprintf(fd,"'%s'",sval);
		break;
	case PSD:
		fprintf(fd,"\n");
		psdval->print(fd);
		break;
	default:
		break;
	}
}

bool var::assert_type(vartype type_) const {
	return type_==type;
}

var* var::ref() {
	refcount++;
	return this;
}

void var::unref() {
	refcount--;
	if(refcount<=0) {
		if(type==FLT) free(fval);
		if(type==STR) free(sval);
		if(type==PSD) delete psdval;
		if(type==MAP) delete mapval;
		free(name);
		delete this;
	}
}

var::~var() {
}

var operator+(const var& a, const var& b) {
	assert(a.type == FLT && b.type == FLT); assert(a.size == b.size);
	var res(a);
	for(size_t i=0; i<res.size; i++) res.fval[i] += b.fval[i];
	return res;
}
var operator-(const var& a, const var& b) {
	assert(a.type == FLT && b.type == FLT); assert(a.size == b.size);
	var res(a);
	for(size_t i=0; i<res.size; i++) res.fval[i] -= b.fval[i];
	return res;
}
var operator*(const var& a, const var& b) {
	assert(a.type == FLT && b.type == FLT); assert(a.size == b.size);
	var res(a);
	for(size_t i=0; i<res.size; i++) res.fval[i] *= b.fval[i];
	return res;
}
var operator/(const var& a, const var& b) {
	assert(a.type == FLT && b.type == FLT); assert(a.size == b.size);
	var res(a);
	for(size_t i=0; i<res.size; i++) res.fval[i] /= b.fval[i];
	return res;
}
