/*
* selav1d - semi-lagrangian simulation of sparse phase-space densities
* Copyright (C) 2020-2022  Philipp Amstutz (DESY)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef VAR_H
#define VAR_H

#include <cstddef>
#include <cstdio>
#include <vector>
#include <map>

#include <psd.h>
#include <maps.h>

struct mapwargs_t { // Map with arguments
	SLV::map_t map;
	void* args;
	~mapwargs_t() {
		SLV::free_map_args(map,args);
		free(args);
	}
};

enum vartype {FLT,STR,PSD,MAP};

class var {
public:
	const vartype type;
	const size_t size = 0;
	char* name = NULL;
	union {
		void* ptr;
		char* sval;
		double* fval;
		SLV::psd<2>* psdval;
		mapwargs_t* mapval;
	};
	int refcount = 1;

	var(double x, int n=1);
	var(double* x, int n);
	var(std::vector<double>*);
	var(const char*);
	var(const var&);
	var(SLV::psd<2>*);
	var(mapwargs_t*);
	bool assert_type(vartype) const;
	void print(FILE* fd) const;
	var* ref();
	void unref();
	~var();
};

var operator+(const var&, const var&);
var operator-(const var&, const var&);
var operator*(const var&, const var&);
var operator/(const var&, const var&);

class environment : public std::map<std::string,var*> {
public:
	~environment() {
		for(auto v : *this) {
			v.second->unref();
		}
	}
};

typedef std::vector<var*> varlist;

#endif
